(* $Id: autolad_lp.ml 9 2005-12-14 12:54:07Z lemaire $ *)
(** ========================================================================
    autolad_lp: part of autolad, a software for Logical Analysis of Data
    Copyright (C) Pierre Lemaire, 2004-2005.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ======================================================================== *)

(* ===========================================================================
  author: Pierre Lemaire
  ========================================================================== *)

open Printf;;
open List;;

open Messages;;
open Misc;;
open Instance;;
open Pandect;;
open Model;;

(** Defines a LP program to select patterns from a pandect. *)
class lp_model
  (inst: instance) (** the instance *)
  (var: int list)  (** the list of the (indices of the) variables *)
  (model: pandect) (** the pandect *)
  (c: float)       (** acceptable cost for an uncovered observation *)
  (k: float) =     (** desired coverage of each observation *)

object(self)
  val pat = Array.of_list (model#all_patterns())
  val mutable nbpat = 0
  val mutable nbobs = inst#m
  val mutable nbvar = length var
  val mutable x = Array.make 1 0.0 (* value xi, "is pattern i used?" *)
  val mutable y = Array.make 1 0.0 (* value yj, "is observation j covered?" *)
  val mutable z = Array.make 1 0.0 (* value zk, "is variable k used?" *)
  val mutable fname = ""
  val mutable file  = stdout

  initializer
    begin let x,y = Filename.open_temp_file "model" ".lp"
    in fname <- x; file <- y;
    end;

    nbpat <- Array.length pat;
    x <- Array.make nbpat 0.0;
    y <- Array.make nbobs 0.0;
    z <- Array.make nbvar 0.0;
    trace TRACE_MOST
      (sprintf "... LP model creation [%d %d %d | %.2f %.2f]%s"
	 nbpat nbobs nbvar c k !eol);

    fprintf file "set I := 0..%d;\n" (nbpat-1);
    fprintf file "set J := 0..%d;\n" (nbobs-1);
    fprintf file "set K := 0..%d;\n" (nbvar-1);
    fprintf file "\n";
    fprintf file "param a{i in I, j in J};\n";
    fprintf file "param v{i in I, k in K};\n";
    fprintf file "\n";
    fprintf file "var x{I} binary;\n";
    fprintf file "var y{J} >= 0, <= 1;\n";
    fprintf file "var z{K} >= 0;\n";
    fprintf file "\n";
    fprintf file "minimize cost: (sum{k in K}z[k])-%f*(sum{j in J} y[j]);\n" c;
    fprintf file "\n";
    fprintf file "s.t. coverage{j in J}: sum{i in I}a[i,j]*x[i]>=%f*y[j];\n" k;
    fprintf file "s.t. usage{i in I, k in K}: z[k] >= v[i,k]*x[i];\n";
    fprintf file "\n";
    fprintf file "solve;\n";
    fprintf file "\n";
    fprintf file "data;\n";
    fprintf file "param a : \n";

    (* print spatobs *)
    fprintf file "     ";
    for i=0 to nbobs - 1 do fprintf file " %5d" i; done;
    fprintf file " :=\n";
    Array.iteri
      (fun k p -> 
	 fprintf file "%5d" k;
	 for o=0 to nbobs - 1 do let o = inst#observation o in
	   fprintf file (* " %5d" *) " %5f"
	     (if p#covers o then if p#kind != o#kind then -0.5 else 1.
	      else 0.)
	 done;
	 fprintf file "\n";
      )
      pat;
    fprintf file ";\n";
    fprintf file "\n";
    (* print patvar *)
    fprintf file "param v : \n";
    fprintf file "     ";
    for i=0 to nbvar - 1 do fprintf file " %5d" i; done;
    fprintf file " :=\n";
    Array.iteri
      (fun k p -> 
	 fprintf file "%5d" k;
	 iter
	   (fun v ->  fprintf file " %5d" (if p#contains v then 1 else 0))
	   var;
	 fprintf file "\n";
      )
      pat;
    fprintf file ";\n";

    fprintf file "\n";
    close_out file;

  (** Select patterns/solves the LP using GPLK; if the argument is
      true, solve it as a MIP. *)
  method solve (mip:bool) =
    let solname = fname^".sol"
    in
      (* Solve LP model *)
      trace TRACE_MOST ("... solve LP model"^(!eol));
      begin
	let e = Unix.system
	  (sprintf "glpsol %s --math %s -o %s >/dev/null"
	     (if mip then "" else "--nomip") fname solname)
	in
	  match e with
	    | Unix.WEXITED x when x=0 -> ()
	    | _ ->
		trace TRACE_SOME ("GLPK problem!"^(!eol));
		raise Sys.Break;
      end;
      trace TRACE_MOST ("... retrieve solution"^(!eol));
      let file = Pervasives.open_in solname
      in begin
	  try while (true) do
	    let ws = Str.split (Str.regexp "[][*\tBUNL ]+") (input_line file)
	      (* ws is normally: num::name::index::value::we-dont-care *)
	    in
	      try
		let v = hd (tl ws)
		and i = int_of_string (hd (tl (tl ws)))
		and t = float_of_string (hd (tl (tl (tl ws))))
		in match v with
		  | "x" -> x.(i) <- t
		  | "y" -> y.(i) <- t
		  | "z" -> z.(i) <- t
		  | _ -> ()
	      with _ -> ()
	  done;
	  with
	    | End_of_file -> close_in file
	end;
	Sys.remove fname;
	Sys.remove solname;

  (** Returns the coverage achieved by the selected patterns. *)
  method coverage () =
    let a = ref 0 
    in
      Array.iter (fun x -> if x > 0. then incr a) y;
      (float_of_int !a) /. (float_of_int nbobs)

  (** Returns a list with all used variables together with their 'score'. *)
  method score_var () =
    let a = ref []
    in 
      Array.iteri (fun i x -> if x > 0. then a := (nth var i,x)::(!a)) z;
      List.rev (!a)

  (** Returns the pandect of the actually used patterns. *)
  method get_sol () =
    let ps = ref []
    in
      Array.iteri (fun i x -> if x > 0. then ps := pat.(i)::!ps) x;
      new model (model#k) !ps

  (** Returns the accuracy on the given tes set of the selected
      pandect. *)
  method accuracy tes =
    let m = self#get_sol() in m#compute_for tes; m#datascope_accuracy

end
