(* $Id: classifier.ml 14 2007-06-29 08:14:54Z pierre $ *)
(** ========================================================================
    classifier: a companion for ladoscope.
    Copyright (C) Pierre Lemaire, 2005.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ======================================================================== *)

open Messages;;
open Misc;;
open Instance;;
open Pandect;;
open Model;;
open Io;;

open Printf;;

type vote_type = AVE_VOTE | SUM_VOTE | MAX_VOTE;;

(* ========================================================================= *)
let classifier arg =
  try
    let name = "classifier"
    and author = "Pierre Lemaire"
    and version = "2007-06-29"
    and copyright = "2005"
    and description = "make predictions using one or several models"
    and usage_msg =
      "usage is: "^Sys.argv.(0)^" [options] m1[=w1] [m2[=w2] [...]]"
    and models  = ref []
    and weights = ref []
    and vote    = ref AVE_VOTE
    and undefval = ref 0.5
    and instfile = ref ""
    and sort = ref true
    and alloutputs = ref false
    in let add s =
	let i = try String.rindex s '=' with | Not_found -> -1
	in let m,w =
	    if i < 0 then s, 1.0
	    else
	      (String.sub s 0 i),
	    float_of_string (String.sub s (i+1) (String.length s - i-1))
	in
	  models := m::!models;
	  weights := w::!weights;

       and arglist =
	(common_options (name,version,description,author,copyright))
	@[

	  ("-sort",
           Arg.Set sort,
           " sorts the output by reading order");
	  ("-nosort",
           Arg.Clear sort,
           " does not sort the output.");

	  (* --- parameters ------------------------------------------------ *)
	  ("-undef",
           Arg.Float (fun x -> undefval := x),
           "(float) value of an undefined vote (default: 0.5).");

	  ("-inst",
           Arg.String (fun s -> instfile := s),
           "(file) set the instance file (default: stdin).");

	  ("-alloutputs",
	   Arg.Set alloutputs,
	   " display, as extra columns, the complete vector of outputs. ");

	  (* --- voting choices -------------------------------------------- *)
	  ("-mean",
           Arg.Unit (fun () -> vote := AVE_VOTE),
           " mean of the votes.");
	  ("-sum",
           Arg.Unit (fun () -> vote := SUM_VOTE),
           " sum of the votes.");
	  ("-max",
           Arg.Unit (fun () -> vote := MAX_VOTE),
           " max of the votes.");
	]
    in
      (* initiate program *)
      if (Array.length arg = 0) then
	Arg.parse (Arg.align arglist) (fun s -> add s) usage_msg
      else begin
	let x = ref 0
	in
	  Arg.parse_argv
	    ~current:x
	    arg
	    (Arg.align arglist)
	    (fun s -> add s) usage_msg
      end;
      flush_all();
      let inst = read_instance (open_in !instfile)
      and model  = Array.of_list
	(List.map
	   (fun x -> read_model (open_in x))
	   !models
	)
      and weight = Array.of_list (!weights)
      and sumw = List.fold_left (fun a b -> a +. b) 0.0 !weights
      in
      let f s c w = (* current score, current vote, weight *)
	match !vote with
	  | AVE_VOTE ->
	      let x = if (c < 0) then !undefval else float c
	      in x*.w/.sumw +. s
	  | SUM_VOTE ->
	      let x = if (c < 0) then !undefval else float c
	      in w*.x +. s
	  | MAX_VOTE ->
              let x = if (c < 0) then !undefval else float c
	      in max s (w*.x)
      in
	Observation.set_compare_function Observation.compare_index;
	if !sort then inst#sort();
	for o=0 to inst#m -1 do
	  let o = inst#observation o;
	  and sco = ref 0.0
	  in
	    for m=0 to (Array.length model -1) do
	      let x = model.(m)#classification o
	      in sco := f !sco x weight.(m) 
	    done;
	    if (names()) then fprintf (out()) "%s;" (o#name);
	    fprintf (out()) "%d;%d;%.3f" o#index o#kind !sco;
	    if (!alloutputs) then
	      for m=0 to (Array.length model -1) do
		fprintf (out()) ";%d" (model.(m)#classification o)
	      done;
	    fprintf (out()) "%s" !eol;
	      
	done;
	
	(* terminates program *)
	0;
  with
    | _ as ex -> terminal_catch ex    
;;

