(* $Id: cutpointset.ml 11 2007-02-28 09:38:48Z pierre $ *)
(* ============================================================================
author: Pierre Lemaire
2004-11-03             : creation
============================================================================ *)

open Printf;;
open Messages;;
open Instance;;
open Monome;;

(* =========================================================================
   class cutpointset
   ========================================================================= *)
class cutpointset (_var: int) _val = (* variable, init values *)
object (self)
  (* =======================================================================
     fields
     ======================================================================= *)
  val mutable value = _val       (* the values *)
  val mutable next = _val        (* the current position *)
  
  (* =======================================================================
     properties
     ======================================================================= *)
     
  method variable = _var;
  
  (* ===================================================================== *)
  method count = List.length value;
       
  (* ===================================================================== *)
  method first = next <- value; self#next

  (* ===================================================================== *)
  method next =
    match next with
    | []    -> nan
    | c::cs -> next <- cs; c

  method values = value
  
  (* =======================================================================
     manipulation
     ======================================================================= *)

  (* ===================================================================== *)
  method add x = value <- x::value

  (* ===================================================================== *)
  method merge (x: cutpointset) =
    value <- value @ x#values;

  (* ===================================================================== *)
  method del x = 
    let rec delrec x ls = 
      match ls with
      | [] -> []
      | l::ls -> if l=x then ls else l::(delrec x ls)
    in value <- delrec x value
      
  (* ===================================================================== *)
  method exists x =
    let p a = (Pervasives.compare x a)=0
    in List.exists p value

  (* =======================================================================
     sorting
     ======================================================================= *)
     
  method sort () = value <- List.sort Pervasives.compare value
  
  method sort_decreasing () =
    let f x y = - Pervasives.compare x y in value <- List.sort f value

  (* =======================================================================
     conversion
     ======================================================================= *)
     
  method to_simple_monomes () =
    let rec parse vs =
      match vs with
      | []    -> []
      | v::vs -> (new monome _var (LT v))::(new monome _var (GT v))::(parse vs)
    in parse value

end;;

let empty_cutpointset i = new cutpointset i [];;

(* =========================================================================
   construction methods
   ========================================================================= *)

(* ========================================================================= *)
(* ========================================================================= *)

(* very inefficient (and only 2-class) version *)
let simplify_for (inst: instance) (cps: cutpointset) =
  let v = cps#variable
  and keep = ref []
  in
    cps#sort();
    let pred = ref neg_infinity
    and curr = ref cps#first
    in
      while (Pervasives.compare !curr nan != 0) do
	let bpos, bneg = ref false, ref false
	and apos, aneg = ref false, ref false
	in
	  for o = 0 to inst#m - 1 do
	    let x = (inst#observation o)#value v
	    in let k = (inst#observation o)#kind
	    in
	      if (!pred <= x && x <= !curr) then begin
		if (k=0) then bneg := true else bpos := true;
	      end else if (!curr < x) then begin
		if (k=0) then aneg := true else apos := true;
	      end
 	  done;
	  if (!apos && !bneg) || (!aneg && !bpos)
	  then keep := (!curr)::(!keep);
	  pred := !curr; curr := cps#next
      done;
      new cutpointset v !keep

(* ========================================================================= *)
let of_monomes ml = 
  let rec sort cpsl =
    match cpsl with
    | [] -> []
    | c::cpsl -> c#sort(); c::(sort cpsl);
  in
  let rec addrec (v,x) cpsl =
    match cpsl with
    | []       -> [new cutpointset v [x]]
    | (cp::cpsl) as all
    -> if cp#variable = v then begin
         if not (cp#exists x) then cp#add x;
         all
       end else cp::(addrec (v,x) cpsl)
  in
  let add (v,x) cpsl =
    match classify_float x with
    | FP_normal | FP_zero -> addrec (v,x) cpsl
    | FP_subnormal | FP_infinite | FP_nan -> cpsl
  in  
  let rec parse ml =
    match ml with
    | []    -> []
    | m::ml -> let (v,ub,lb) = m#variable, m#ub, m#lb
               in add (v,ub) (add (v,lb) (parse ml))
  in
    sort (parse ml)
;;

(* ========================================================================= *)
let rec add c cpsl =
  match cpsl with
  | []       -> [c]
  | cp::cpsl
  ->
    if cp#variable = c#variable then
      let x = ref (c#first)
      in while Pervasives.compare !x nan <> 0 do
        if not (cp#exists !x) then cp#add !x;
        x := c#next
      done;
      cp#sort();
      cp::cpsl
    else cp::(add c cpsl)

(* ========================================================================= *)
let all_candidates inst i = 
  let rec add (v,k) ls =
  match ls with
  | []        -> [(v,k)]
  | (a,b)::ls as xx ->
    if v<a then (v,k)::xx
    else if v>a then (a,b)::(add (v,k) ls)
    else if b=k then xx
    else (a,-1)::ls
  in let rec parse ls =
    match ls with
    | [] -> []
    | (_,_)::[] -> []
    | (a,b)::(c,d)::ls ->
      if (b = -1 or b<>d) then ((a+.c)/.2.0)::(parse ((c,d)::ls))
      else parse ((c,d)::ls)
  and all = ref []    
  in
    for o=0 to inst#m - 1 do
      let obs = inst#observation o
      in let v = obs#value i
      in 
        if Pervasives.compare v nan != 0
        then all := add (obs#value i, obs#kind) (!all);
    done;
    parse !all
;;
    
let all inst i =
      new cutpointset i (all_candidates inst i)
;;

(* ========================================================================= *)
let rand x inst i =
  let rec clean zs =
    match zs with
    | [] -> []
    | z::zs -> if (Random.float 1.0 < x) then z::(clean zs) else clean zs
  in
    new cutpointset i (clean (all_candidates inst i))

(* ========================================================================= *)
let bin n inst i =
  let rec keep k x zs =
    match zs with
    | [] -> []
    | z::zs -> if (x mod k = 0)
      then z::(keep k (x+1) zs)
      else keep k (x+1) zs
  and all = all_candidates inst i 
  in let l = List.length all
  in let k = int_of_float (ceil ((float_of_int l) /. (float_of_int n)))
  in let k = max k 1
  in let offset = (l mod k) + 1 
  in new cutpointset i (keep k offset all)

(* ========================================================================= *)
let greed n inst i =
  let apart = Array.make_matrix inst#m inst#m 0
  in 
  let value m = 
    let k = ref 0
    and m = new monome i (LT m)
    in for o1= 0 to inst#m - 1 do for o2=o1+1 to inst#m - 1 do
      if apart.(o1).(o2) = 0 then
        let o1, o2 = (inst#observation o1), (inst#observation o2)
        in
          try
            if (o1#kind != o2#kind) && (m#covers o1 != m#covers o2)
            then k := !k + 1;
          with
          | Failure "undefined" -> ()
      done; done;
      !k
  in
  let rec findbest ms =
    match ms with
    | [] -> 0.0, -1, []
    | m::ms -> let v = value m
               and (bm,bv,a) = findbest ms
               in if v > bv then (m,v,bm::a) else (bm,bv,m::a)
  and  all = ref (all_candidates inst i)
  and  keep = ref []
  in 
    for k=1 to n do
      let (bm,bv,a) = findbest !all
      in if bv > 0 then begin
           keep := bm::!keep;
           let bm = new monome i (LT bm)
           in for o1=0 to inst#m-1 do for o2=o1+1 to inst#m-1 do
             try
                if bm#covers (inst#observation o1) != bm#covers (inst#observation o2)
                then apart.(o1).(o2) <- 1;
             with
             | Failure "undefined" -> ()
           done; done;
         end;
         all := a
    done;  
    new cutpointset i !keep

(* ========================================================================= *)
let interval n inst i =
  let lb,ub,n =
    let lb, ub, vs = ref infinity, ref neg_infinity, ref []
    in 
      for o = 0 to inst#m - 1 do
	let o = (inst#observation o)
	in let v = o#value i
	in
	  if Pervasives.compare v nan != 0 then begin
	    lb := min !lb v;
	    ub := max !ub v;
	    if not (List.exists (fun x -> Pervasives.compare x v == 0) !vs)
	    then vs := v::(!vs);
	  end
      done;
      !lb, !ub, min n ((List.length !vs)-1)
  in
  let keep = ref []
  and fn = float_of_int n
  in for k = n downto 1 do
      keep := (lb +. ((float_of_int k) *. (ub-.lb) /. (fn+.1.)))::(!keep);
    done;
    new cutpointset i !keep

(* ========================================================================= *)
let sequence dx inst i =
  if (dx > 0.0) then begin
    let e = (* epsilon_float *) 0.00001
    in
    let lb,ub =
      let lb, ub = ref infinity, ref neg_infinity
      in 
	for o = 0 to inst#m - 1 do
	  let v = (inst#observation o)#value i
	  in
	    if Pervasives.compare v nan != 0 then begin
	      lb := min !lb v;
	      ub := max !ub v;
	    end
	done;
      (ceil ((!lb+.dx-.e)/.dx))*.dx,
      (floor((!ub-.dx+.e)/.dx))*.dx
    in
    let x = ref lb
    and keep = ref []
    in
      while (!x <= ub) do keep := !x :: !keep; x := !x +. dx; done;
      new cutpointset i (List.rev !keep)
  end
  else
    new cutpointset i []


(* ========================================================================= *)
let hybrid ns n inst i =
  let ngreed, nbin = ns
  in
  let cp = empty_cutpointset i
  in
    cp#merge (greed (min ngreed n) inst i);
    if (ngreed < n) then begin
      cp#merge (bin (min nbin (n-ngreed)) inst i);
      if (ngreed+nbin < n) then begin
	cp#merge (interval (n-ngreed-nbin) inst i);
      end
    end;
    cp#sort();
    simplify_for inst cp

