(* $Id: cutpointset.mli 8 2005-12-13 11:23:16Z lemaire $ *)
(** 
 The class for a set of cut-points for a variable.

 @version November 08, 2004
 @author Pierre Lemaire
*)

(* ============================================================================
author: Pierre Lemaire
2004-11-03             : creation
2004-11-03..2004-11-08 : basics
============================================================================ *)

open Instance;;
open Monome;;

(* ======================================================================= *)
(** {6 the class [cutpointset]}*)

(** The class [cutpointset] defines a set of cut-points on which [monomes]
    and then [patterns] can be built. This set of values are possible
    cut-point values for a given variable.  
  *)
class cutpointset:
  int ->          (* the index of the variable *)
  float list ->   (* the initial set of values *)
object 
  (* ===================================================================== *)
  (** {6 fields} *)
  
  (** the values of the cut-points for this [cutpointset]. *)
  val mutable value: float list
  
  (** the next value (see first and next methods). *)
  val mutable next: float list 
  
  (* ===================================================================== *)
  (** {6 properties} *)
    
  (** returns the index of the variable this [cutpointset] is made for. *)
  method variable: int

  (** returns the number of cut-points (values) this [cutpointset] is
      made of. *)
  method count: int
    
  (** returns the first value in this [cutpointset]. *)
  method first: float
  
  (** returns the next value in this [cutpointset].
      Returns [nan] if there is no more value. *)
  method next: float

  method values: float list
  
  (* ===================================================================== *)
  (** {6 manipulation}*)

  (** adds the given value to the list of values. *)
  method add: float -> unit
  
  (** adds all the value of the given [cutpointset]. *)
  method merge: cutpointset -> unit

  (** deletes the given value (one occurence) from the list of values. *)
  method del: float -> unit

  (** returns whether the given value is in the list of values. *)
  method exists: float -> bool

  (* ===================================================================== *)
  (** {6 sorting}*)
  
  (** sorts the list of values by non-decreasing order. *)
  method sort: unit -> unit
  
  (** sorts the list of values by non-increasing order. *)
  method sort_decreasing: unit -> unit

  (* ===================================================================== *)
  (** {6 conversions}*)
  
  (** returns the list of all simple [monomes] (i.e. LT and GT types) on this
      [cutpointset]. *)
  method to_simple_monomes: unit -> monome list

end

(** An empty cut-point set for the given variable. *)
val empty_cutpointset: int -> cutpointset
  
(* ======================================================================= *)
(** {6 construction procedures}*)

(** returns all the cutpointset corresponding to a list of [monomes]. *)
val of_monomes: monome list -> cutpointset list

val add: cutpointset -> cutpointset list -> cutpointset list
  
(** finds a complete set of values for the given [instance] and the given
    variable, that is a set that contains a value in each interval between
    two [observations] of different kinds/classes (undefined is not consider
    equal to itself). The set is sorted. *)
val all: instance -> int -> cutpointset

val rand: float -> instance -> int -> cutpointset

val bin: int -> instance -> int -> cutpointset

val greed: int -> instance -> int -> cutpointset

val interval: int -> instance -> int -> cutpointset

val sequence: float -> instance -> int -> cutpointset

val hybrid: (int*int) -> int -> instance -> int -> cutpointset

(** removes cut-points redundant for the given instance. Between two
equivalent cp, the smallest one is kept. *)
val simplify_for: instance -> cutpointset -> cutpointset

