(* $Id: instance.mli 1 2005-07-15 19:39:24Z lemaire $ *)
(** 
 The class for an instance.

 @version November 08, 2004
 @author Pierre Lemaire
*)

(* ============================================================================
author: Pierre Lemaire
2004-10-29             : creation
2004-10-29..2004-11-08 : basics
============================================================================ *)

open Observation;;

(* ======================================================================= *)
(** {6 the class [instance]}*)

(** The class [instance] defines an instance for LAD,
    that is a collection of [observations].
    
    [instance k obs] creates a new instance using [obs] as observations;
    the [observations] must have kind/class in \[1..[k]-1\].
    
    All the [observations] are supposed to have the same number of variables,
    but this is not checked (hence, one can use [observations] with different
    number of variables, as long as the extra-variables are never used).
     
    The number of [observations] cannot change (but the [observations] 
    themselves can).
*)
class instance:
  int -> 
  observation array ->
object

  (* ===================================================================== *)
  (** {6 fields} *)
  
  (** the [observations] contained in this [instance]. *)
  val mutable obs: observation array

  (** the number of [observations] of each kind/class. *)
  val kind: int array  
  
  (** the number of [observations] without kind/class. *)
  val mutable undef: int

  (* ===================================================================== *)
  (** {6 counters} *)

  (** returns the number of classes/kinds (except undefined). *)
  method k: int
  method count_kinds: int (** same as [method k: int] *)

  (** returns the number of [observations] in this [instance]. *)
  method m: int
  method count_observations: int  (** same as [method m: int] *)
    
  (** returns the number of variables in this [instance]. *)
  method n: int
  method count_variables: int  (** same as [method n: int] *)
  
  (** [count_of_kind k] returns the number of [observations] of
      class/kind [k].
      
      Raises [Invalid_argument "Instance.count_of_kind: no such  kind"]
      if the [instance] does not have a kind [k].
      *)
  method count_of_kind: int -> int
  
  (** returns the number of [observations] with an undefined kind/class. *)
  method count_undefined: int
  
  (* ===================================================================== *)
  (** {6 properties} *)

  (** [observation i] returns the [i]th [observation] of this [instance].
      
      Raises [Invalid_argument "Instance.observation: no such observation"]
      if the [instance] does not have a [i]th [observation]. *)
  method observation: int -> observation

  (** sets the [i]th [observation] of this [instance].
      
      Raises [Invalid_argument "Instance.set_observation: no such observation"]
      if the [instance] does not have a [i]th [observation].
      
      Raises [Invalid_argument "Instance.set_observation: no such kind"]
      if the [instance] is not intended for the kind/class of the
      [observation]. *)
  method set_observation: int -> observation -> unit

  (* ===================================================================== *)
  (** {6 methods} *)
  
  (** Normalizes this [instance]. That is, the value xi of each variable x
      is replaced by (xi-average(x))/(stddev(x) ; stdev is the standard
      deviation based on the unbiased variance. *)
  method normalize: unit -> unit
  
  (* ===================================================================== *)
  (** {6 sorting} *)
  
  (** sorts the [observations] according to the current comparison
      function for [observations]. The sorting is stable. *)
  method sort: unit -> unit
  
  (** mixes all the observations up. *)
  method mixup: unit -> unit
  
end
