(* $Id: io.ml 8 2005-12-13 11:23:16Z lemaire $ *)
(* ============================================================================
author: Pierre Lemaire
2004-11-05             : creation
2004-11-05..2004-11-08 : observation, instance
============================================================================ *)

open Messages;;
open Misc;;
open Observation;;
open Instance;;
open Monome;;
open Pattern;;
open Pandect;;
open Model;;
open Cutpointset;;

open Lexing;;
open Printf;;

(* =========================================================================
   i/o for observations
   ========================================================================= *)

let newline_buf buf =
  buf.lex_curr_p
  <- {pos_fname = buf.lex_curr_p.pos_fname;
      pos_lnum = buf.lex_curr_p.pos_lnum + 1;
      pos_bol = buf.lex_curr_p.pos_cnum;
      pos_cnum = buf.lex_curr_p.pos_cnum;}
;;

(* ========================================================================= *)
let read_observation buf = 
  try
    let (n, k, vs) = Observation_gra.read Observation_lex.token buf
    in
      newline_buf buf;
      if (String.length n > 0)
      then
	new observation ~_name:n (-1) k (Array.of_list vs)
      else
	new observation (-1) k (Array.of_list vs)
  with
    | Parsing.Parse_error
      -> let s = lexeme buf
      in
	if String.length s > 0 then raise (IO_error (Parse buf))
	else raise (IO_error (Eof buf))
    | Failure "lexing: empty token"
      -> raise (IO_error (Parse buf))
;;

(* ========================================================================= *)
let observation_of_string s = read_observation (Lexing.from_string s)
;;
  
(* ===================================================================== *)
let string_of_observation (o:observation) =
  let vs  = Array.make o#n 0.0
  and name = names()
  in let pre = (if name then (o#name)^";" else "")^(string_of_int o#kind)
  in
    for i=0 to o#n-1 do vs.(i) <- o#value i done;
    String.concat ";"
      (pre
       ::(Array.to_list
            (Array.map (sprintf "%f") vs)))
;;

(* ========================================================================= *)
let print_observation out o =
  output_string out (string_of_observation o)
;;

(* =========================================================================
   i/o for instances
   ========================================================================= *)

(* ========================================================================= *)
let read_instance buf = 
  let i = ref []
  and k = ref (-1)
  and n = ref 0
  in
(*     trace TRACE_MOST "Read instance\n"; *)
    begin try
      while true do
	let o = read_observation buf
	in o#set_index !n; n := !n + 1;
	  i := o::(!i);
          if !k < o#kind then k := o#kind;
(* 	  trace TRACE_ALL ("...read: "^(string_of_observation o)^"\n"); *)
      done;
    with
      | IO_error (Eof _) -> ()
    end;
    new instance !k (Array.of_list !i);
;;

(* ========================================================================= *)
let instance_of_string s = read_instance (Lexing.from_string s)
;;
  
(* ===================================================================== *)
let string_of_instance (inst:instance) =
  let m = inst#m
  and l = ref []
  in for i=0 to m - 1 do
    l := (string_of_observation (inst#observation i))::(!l)
  done;
  String.concat "\n" !l
;;

(* ========================================================================= *)
let print_instance out inst =
  let m = inst#m
  in for i=0 to m - 1 do
    print_observation out (inst#observation i);
    output_char out '\n'
  done
;;

(* =========================================================================
   i/o for monomes and patterns
   ========================================================================= *)

(* ========================================================================= *)
let read_pattern buf = 
  let rec parse ms =
    match ms with
    | []    -> []
    | (v,lb,ub)::ms
    -> (if Pervasives.compare lb neg_infinity = 0 then
          new monome v (LT ub)
        else if Pervasives.compare ub infinity = 0 then
          new monome v (GT lb)
        else new monome v (BOX (lb,ub)))
       ::(parse ms)
  in 
  try
    let ((k,hom,pre,hr), ms) = Pattern_gra.read Pattern_lex.token buf
    in let p = new pattern (-1) (parse ms)
    in if (k>=0) then begin
	p#set_kind k;
	p#set_homogeneity hom;
	p#set_max_prevalence pre;
	p#set_hazardratio hr;
      end;
      newline_buf buf;
      p
  with 
    | Parsing.Parse_error ->
	let s = lexeme buf
	in if String.length s > 0 then raise (IO_error (Parse buf))
	  else raise (IO_error (Eof buf))
    | Failure "lexing: empty token" -> raise (IO_error (Parse buf))
    | End_of_file -> raise (IO_error (Eof buf))
;;

(* ========================================================================= *)
let read_pats noer buf = 
  let i = ref []
  and k = ref (-1)
  and n = ref 0
  in
(*     trace TRACE_MOST "Read pandect/model\n"; *)
    begin try
      while true do
	let p = read_pattern buf
	in
	  p#set_index !n; n := !n + 1;
	  i := p::(!i);
          if !k < p#kind then k := p#kind; 
      done;
    with
      | IO_error (Eof _) -> ()
      | _ as e -> if noer then () else raise e
    end;
    !k,!i

let read_pandect ?(noer=false) buf = 
  let k,p = read_pats noer buf in new pandect k p;;

let read_model ?(noer=false) buf =
  let k,p = read_pats noer buf in new model k p;;

(* ========================================================================= *)
let string_of_monome mon = 
  match mon#inequality with
  | LT x    -> (sprintf "x%d < %f" mon#variable x)
  | GT x    -> (sprintf "x%d > %f" mon#variable x)
  | BOX(x,y)-> (sprintf "%f < x%d < %f" x mon#variable y)
;;

(* ========================================================================= *)
let print_monome out mon = 
  match mon#inequality with
  | LT x    -> output_string out (sprintf "x%d < %f" mon#variable x)
  | GT x    -> output_string out (sprintf "x%d > %f" mon#variable x)
  | BOX(x,y)-> output_string out (sprintf "%f < x%d < %f" x mon#variable y)
;;
 
(* ========================================================================= *)
let print_pattern out pat =
  let rec print ml =
    match ml with
    | [] -> ()
    | m::[] -> print_monome out m
    | m::ml -> print_monome out m; output_string out " & "; print ml
  in output_string out (sprintf "(%d,%.4f,%.4f,%.4f) " (pat#kind) (pat#homogeneity) (pat#max_prevalence) (pat#hazardratio));
     print (pat#monomes())
;;

(* =========================================================================
   i/o for pandect
   ========================================================================= *)

let print_pandect out pan =
  let rec print pl =
    match pl with
    | [] -> ()
    | p::pl -> print_pattern out p; output_string out "\n"; print pl
  in for k=0 to pan#k - 1 do
(*       output_string out (sprintf "--- class %d ---\n" k); *)
      print (pan#patterns_of_kind k)
     done
;;

(* =========================================================================
   i/o for ...
   ========================================================================= *)

let print_cutpointset out cps = 
  output_string out (sprintf "%d:" cps#variable);
  let x = ref (cps#first)
  in while Pervasives.compare !x nan != 0 do
       output_string out (sprintf " %.3f" !x);
       x := cps#next
     done;
;;

let print_cutpointset_list out cpsl = 
  let rec print cpsl =
    match cpsl with
    | [] -> ()
    | cps::cpsl -> print_cutpointset out cps; output_string out "\n"; print cpsl
  in print cpsl
;;
    

let read_cuts buf = 
  let cuts = ref []
  in begin try
    while true do
      let (i,v) = Cutpointset_gra.read Cutpointset_lex.token buf
      in
	newline_buf buf;
	cuts := (new cutpointset i v)::(!cuts);
    done;
    with
      | Parsing.Parse_error
	-> let s = lexeme buf
	in if String.length s > 0 then raise (IO_error (Parse buf))
        else ()
      | Failure "lexing: empty token"
	-> raise (IO_error (Parse buf))
    end;
    !cuts
;;
