(* $Id: ladoscope.ml 14 2007-06-29 08:14:54Z pierre $ *)
(** ========================================================================
    ladoscope: a software for Logical Analysis of Data
    Copyright (C) Pierre Lemaire, 2004-2005.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ======================================================================== *)

(* ============================================================================
author: Pierre Lemaire
2004-10-25             : creation
2004-10-26..2004-10-29 : basics 
============================================================================ *)

open Messages;;
open Misc;;
open Observation;;
open Instance;;
open Cutpointset;;
open Monome;;
open Pattern;;
open Pandect;;
open Model;;
open Io;;
open Autolad_lp;;

open Printf;;

type ladoscope_action_type =
    NONE
  | NORMALIZE
  | PATTERNS of bool | UPDATEPAT | CLEANCOV | CLEANINCL
  | CUTS
  | PATPROD
  | GREEDYMOD of int
  | GREEDYCOVMOD of ((pattern -> pattern -> int) * int * int)
  | SELECTCOVMOD of ((pattern -> pattern -> int) * int * int)
  | LPSELECT of (float * float * bool) 
  | ALLNANOMODS of float | BESTNANOMODS of char | BOXMOD
  | ACCURACY  | CLASSIFICATION | DISCRIMINANT | DATASCOPEACCURACY
  | BATCH
  | COVERAGE  | INCIDENCES | VARINC | VARIABLES | BINARY
  | SCORE_SUMCARD | SCORE_SUMRATIO | SCORE_SUMPREV
  | SCORE_BESTPREV | SCORE_BESTHR
  | SCORE_HRVALUE
  | SCORE_CLOSESTOPPONENT of int | SCORE_CLOSESTGRIDOPPONENT
  | SCORE_DISTANCE  | SCORE_PREVDISTANCE  | SCORE_HRDISTANCE
;;

type cut_type =
    ALLCUTS
  | RANDCUTS of float
  | BINCUTS of int
  | GREEDYCUTS of int
  | INTERVCUTS of int
  | SEQCUTS of (float list)
  | HYBRIDCUTS of (int*int*int)
  | READCUTS
;;  

(* ========================================================================= *)
let ladoscope arg =
  try
    let name = "ladoscope"
    and author = "Pierre Lemaire"
    and version = "2007-06-29"
    and copyright = "2004-2005"
    and description = "a software for Logical Analysis of Data"
    and usage_msg = "usage is: "^Sys.argv.(0)^" [options] [inputfile]"
    and seed = ref (-1)
    and deg = ref 2
    and pre = [| 0.6; 0.6 |]
    and hom = [| 1.0; 1.0 |]
    and hr  = [| 0.0; 0.0 |]
    and finst  = ref ""
    and fmodel = ref ""
    and fcuts = ref ""
    and what = ref NONE
    and cut  = ref ALLCUTS
    and sort = ref false
    and i = ref 0
    and countout = ref false
    and cutnumbers = ref []
    and cleancuts = ref false

    and sdumb = ref ""
    and idumb = ref 0
    and idumb2 = ref 0
    and fdumb = ref 0.
    in let arglist =
	(common_options (name,version,description,author,copyright))
	@[
	  ("-seed",
           Arg.Set_int seed,
           "(int) sets the random number seed (default: auto-init)");

	  ("-sort",
           Arg.Set sort,
           " sorts the output by reading order");
	  ("-nosort",
           Arg.Clear sort,
           " does not sort the output.");

	  (* --- pattern parameters ---------------------------------------- *)
	  ("-d",
           Arg.Set_int deg,
           "(int) sets the degree of the patterns");
	  ("-h",
           Arg.Float (fun x -> hom.(0) <- x; hom.(1) <- x; ),
           "(float) sets the homogeneity of the patterns");
	  ("-hpos",
           Arg.Float (fun x -> hom.(1) <- x;),
           "(float) sets the homogeneity of the positive patterns");
	  ("-hneg",
           Arg.Float (fun x -> hom.(0) <- x;),
           "(float) sets the homogeneity of the negative patterns");     
	  ("-p",
           Arg.Float (fun x -> pre.(0) <- x; pre.(1) <- x; ),
           "(float) sets the prevalence of the patterns");
	  ("-ppos",
           Arg.Float (fun x -> pre.(1) <- x;),
           "(float) sets the prevalence of the positive patterns");
	  ("-pneg",
           Arg.Float (fun x -> pre.(0) <- x;),
           "(float) sets the prevalence of the negative patterns");
	  ("-hr",
           Arg.Float (fun x -> hr.(0) <- x; hr.(1) <- x; ),
           "(float) sets the hazard ratio of the patterns");
	  ("-hrpos",
           Arg.Float (fun x -> hr.(1) <- x;),
           "(float) sets the hazard ratio of the positive patterns");
	  ("-hrneg",
           Arg.Float (fun x -> hr.(0) <- x;),
           "(float) sets the hazard ratio of the negative patterns");
	  ("-mis",
	   Arg.Float (fun x -> Pattern.missing_threshold.(0) <-x;
		        Pattern.missing_threshold.(1) <-x; ),
	   " sets the missing values max ratio (draft).");
	  ("-mispos",
	   Arg.Float (fun x -> Pattern.missing_threshold.(1) <-x; ),
	   " sets the missing values max ratio for positive (draft).");
	  ("-misneg",
	   Arg.Float (fun x -> Pattern.missing_threshold.(0) <-x; ),
	   " sets the missing values max ratio for negative (draft).");
	  ("-misforall",
	   Arg.Set Pattern.missing_for_all,
	   " missing values affect patterns of every kind (draft).");       
	  
	  (* --- cutpoints parameters -------------------------------------- *)

	  ("-allcuts",
	   Arg.Unit (fun () -> cut := ALLCUTS),
	   " uses all relevant cutpoints.");
	  ("-randcuts",
	   Arg.Float (fun z -> cut := RANDCUTS z;),
	   "(float) uses all relevant cutpoints, each one kept with the given "
	   ^"probability.");
	  ("-bincuts",
	   Arg.Int (fun z -> cut := BINCUTS z;),
	   "(int) uses bin cutpoints.");
	  ("-greedycuts",
	   Arg.Int (fun z -> cut := GREEDYCUTS z;),
	   "(int) produces the given number of greedy cutpoints per variable.");
	  ("-intervalcuts",
	   Arg.Int (fun z -> cut := INTERVCUTS z;),
	   "(int) produces the given number of interval cutpoints per variable.");
	  ("-cutnumbers",
	   Arg.String
	     (fun s ->
		try
		  cutnumbers :=
		    List.map
		      int_of_string
		      (Str.split (Str.regexp "[ \t\n\r]+") s)
		with Failure _ ->
		  raise (IO_error (Misc "parse error reading indices"))),
	   "(int_list) list of numbers of cp for each variable.");

	  ("-seqcuts",
	   Arg.String
	     (fun s ->
		try
		  let z = 
		    List.map
		      float_of_string
		      (Str.split (Str.regexp "[ \t\n\r]+") s)
		  in cut := SEQCUTS z
		with Failure _ ->
		  raise (IO_error (Misc "parse error reading dx"))),
	   "(float_list) produces cp as a multiple of their associated dx.");

	  ("-hybridcuts",
	   Arg.Tuple
	     [Arg.Set_int idumb;
	      Arg.Set_int idumb2;
	      Arg.Int
		(fun i -> cut := HYBRIDCUTS (!idumb,!idumb2,i))],
	   "(i*i*i) produces hybrid (greed,bin,total) cp.");



	  ("-readcuts",
	   Arg.Unit (fun () -> cut := READCUTS),
	   " reads cutpoints from a file (or stdin).");
	  ("-fcuts",
	   Arg.String (fun s -> cut := READCUTS; fcuts := s),
	   "(file) reads cutpoints from the given file.");

	  ("-cleancuts",
	   Arg.Unit (fun () -> cleancuts := true),
	   " removes redundant cut-points.");

	  (* --- model ----------------------------------------------------- *)

	  ("-mod",
	   Arg.Set_string fmodel,
           "(string) sets the model/patterns file");

	  (* --- actions --------------------------------------------------- *)

	  ("-normalize",
	   Arg.Unit (fun () -> what := NORMALIZE),
           " normalizes an instance.");

	  ("-binary",
	   Arg.Unit (fun () -> what := BINARY),
           " binarizes the instance using a the cut-points defined in a given "
           ^"cutpointset (or model).");
	  ("-patterns",
	   Arg.Unit (fun () -> what := PATTERNS false),
           " computes all patterns (according to other parameters).");
	  ("-pseudoprimes",
	   Arg.Unit (fun () -> what := PATTERNS true),
           " computes primes patterns (or almost...).");
	  ("-updatepat",
	   Arg.Unit (fun () -> what := UPDATEPAT),
           " updates the characteristics of the patterns for the given inst.");
	  ("-cleancov",
	   Arg.Unit (fun () -> what := CLEANCOV),
           " cleans pandect according to coverage-inclusion.");
	  ("-cleanincl",
	   Arg.Unit (fun () -> what := CLEANINCL),
           " cleans pandect according to definition-inclusion.");
	  ("-patprod",
	   Arg.Unit (fun () -> what := PATPROD),
           " makes the product of all the patterns (draft version).");

	  ("-cuts",
	   Arg.Unit (fun () -> what := CUTS),
           " outputs the computed cutpoinset.");

	  ("-allnanomods",
	   Arg.Float (fun z -> what := ALLNANOMODS z),
	   "(float) computes all nanomodels with at least the given accuracy.");
	  ("-bestnanomods",
	   Arg.String (fun s -> what := BESTNANOMODS( 
			 try s.[0]
			 with _ ->
			   warning "No nanomodel evaluation (use accuracy)";
			   'a')
		      ),
	   "(string) computes the best nano-models for a-ccuracy or b-alance.");
	  ("-greedymod",
	   Arg.Int (fun z -> what := GREEDYMOD z),
           "(int) selects greedily patterns to cover each point x times "
           ^"(if possible).");

	  ("-greedycovmod",
	   Arg.Tuple
	     [Arg.Set_string sdumb;
	      Arg.Set_int idumb;
	      Arg.Int
		(fun i -> 
		   what := GREEDYCOVMOD 
		     ((try match (!sdumb).[0] with
			 | 'I' -> Pattern.compare_index
			 | 'i' -> fun p q -> - Pattern.compare_index p q
			   | 'D' -> Pattern.compare_degree
			   | 'd' -> fun p q -> - Pattern.compare_degree p q
			     | 'P' -> Pattern.compare_prevalence
			     | 'p' -> fun p q -> - Pattern.compare_prevalence p q
			       | 'H' -> Pattern.compare_homogeneity
			       | 'h' -> fun p q -> - Pattern.compare_homogeneity p q
				 | _ ->
				     error 7 "-greedymod: unknown method";
				     Pattern.compare_index
		       with
			 | _ -> error 7 "-greedymod: unknown method";
			     Pattern.compare_index
		      ), !idumb, i))],
           "(char*int*int) greedy selection of patterns.");

	  ("-selectcovmod",
	   Arg.Tuple
	     [Arg.Set_string sdumb;
	      Arg.Set_int idumb;
	      Arg.Int
		(fun i -> 
		   what := SELECTCOVMOD 
		     ((try match (!sdumb).[0] with
			 | 'i' -> Pattern.compare_index
			 | 'I' -> (fun p q -> - Pattern.compare_index p q)
			 | 'd' -> Pattern.compare_degree
			 | 'D' -> (fun p q -> - Pattern.compare_degree p q)
			 | 'p' -> Pattern.compare_prevalence
			 | 'P' -> (fun p q -> - Pattern.compare_prevalence p q)
			 | 'h' -> Pattern.compare_homogeneity
			 | 'H' -> (fun p q -> - Pattern.compare_homogeneity p q)
			 | _ ->
			     error 7 "egreedymod: unknown method";
			     Pattern.compare_index
		       with
			 | _ -> error 7 "egreedymod: unknown method";
			     Pattern.compare_index
		      ), !idumb, i))],
           "(char*int*int) iterative selection of patterns.");

	  ("-lpselect",
	   Arg.Tuple
	     [Arg.Set_string sdumb;
	      Arg.Set_float fdumb;
	      Arg.Float
		(fun f ->
		   what := LPSELECT
		     ((!fdumb),
		      f,
		      (!sdumb).[0] != 'l' ))],
	   "(char*float*float) (MI)LP selection of model.");

	  ("-boxmod",
	   Arg.Unit (fun () -> what := BOXMOD),
           " builds greedily boxes, one for each observation.");

	  ("-accuracy",
	   Arg.Unit (fun () -> what := ACCURACY),
           " displays the accuracy of the model on the instance.");
	  ("-datascopeaccuracy",
	   Arg.Unit (fun () -> what := DATASCOPEACCURACY),
           " displays the accuracy (as DataScope) of the model on the instance.");
	  ("-classification",
	   Arg.Unit (fun () -> what := CLASSIFICATION),
           " outputs the classification of the given observations for the "
           ^"given model.");

	  ("-batch",
	   Arg.Unit (fun () -> what := BATCH),
           " displays acc/dsacc/spe/sen for models read from stdin.");

	  ("-discriminant",
	   Arg.Unit (fun () -> what := DISCRIMINANT),
           " outputs the discriminant value of the given observations for the given "
           ^"model.");

	  ("-coverage",
	   Arg.Unit (fun () -> what := COVERAGE),
           " outputs the coverage of the patterns of the given model.");
	  ("-incidences",
	   Arg.Unit (fun () -> what := INCIDENCES),
           " outputs the observations-patterns incidence matrix.");
	  ("-varinc",
	   Arg.Unit (fun () -> what := VARINC),
           " outputs the patterns-variable incidence matrix.");
	  ("-plain", 
	   Arg.Unit (fun () -> i := 0),
           " uses plain format for incidence matrices.");
	  ("-csv", 
	   Arg.Unit (fun () -> i := 1),
           " uses CSV format for incidence matrices.");
	  ("-ampl", 
	   Arg.Unit (fun () -> i := 2),
           " uses AMPL format for incidence matrices.");
	  ("-variables",
	   Arg.Unit (fun () -> what := VARIABLES),
           " outputs the 'monotonicity' of the variables in the given model.");

	  ("-sumcardscore",
	   Arg.Unit (fun () -> what := SCORE_SUMCARD),
           " sum-of-card-score [#POS-#NEG] of an instance for a model.");
	  ("-sumratioscore",
	   Arg.Unit (fun () -> what := SCORE_SUMRATIO),
           " sum-of-ratio-score [%POS-%NEG] of an instance for a model.");
	  ("-sumprevscore",
	   Arg.Unit (fun () -> what := SCORE_SUMPREV),
           " sum-of-prev-score [pre(POS)-pre(NEG)] of an instance for a model.");
	  ("-bestprevscore",
	   Arg.Unit (fun () -> what := SCORE_BESTPREV),
           " best-prev-score [max(pre)] of an instance for a model.");
	  ("-besthrscore",
	   Arg.Unit (fun () -> what := SCORE_BESTHR),
           " best-hr-score [max(hr)] of an instance for a model.");

	  ("-sumhr",
	   Arg.Unit (fun () -> what := SCORE_HRVALUE),
           " sum-of-hr [hr(POS)+hr(NEG)] of an instance for a model.");

	  ("-closestgridopponentscore",
	   Arg.Unit (fun () -> what := SCORE_CLOSESTGRIDOPPONENT),
           " displays closest-opponant score, computed on the grid."); 
	  ("-closestopponentscore",
	   Arg.Int (fun x -> what := SCORE_CLOSESTOPPONENT x),
	   "(int) displays closest-opponant score, computed for the given depth "
	   ^"(draft version).");

	  ("-dscore",
	   Arg.Unit (fun () -> what := SCORE_DISTANCE),
	   " displays sum-of-distance to patterns score.");
	  ("-dprevscore",
	   Arg.Unit (fun () -> what := SCORE_PREVDISTANCE),
	   " displays sum-of-distance to patterns, weighted by prev, score.");
	  ("-dhrscore",
	   Arg.Unit (fun () -> what := SCORE_HRDISTANCE),
	   " displays sum-of-distance to patterns, weighted by hr, score."); 
	  ("-countout",
	   Arg.Set countout,
	   " counts also patterns that are not triggered.");
	  ("-countinonly",
	   Arg.Clear countout,
	   " counts only patterns that are triggered.");

	]
    in
      (* initiate program *)
      if (Array.length arg = 0) then
	Arg.parse (Arg.align arglist) (set_ref finst) usage_msg
      else begin
	let x = ref 0
	in
	  Arg.parse_argv
	    ~current:x
	    arg
	    (Arg.align arglist)
	    (set_ref finst) usage_msg
      end;

      if (!seed < 0) then Random.self_init() else Random.init(!seed);
      Observation.set_compare_function Observation.compare_index;
      Pattern.set_compare_function Pattern.compare_index;

      let f_cuts =
	let good_i n i = try List.nth !cutnumbers n with Failure "nth" -> i
	in
	  match !cut with
	    | ALLCUTS      -> Cutpointset.all
	    | RANDCUTS x   -> Cutpointset.rand x
	    | BINCUTS i    ->
		fun inst n -> Cutpointset.bin (good_i n i) inst n
		| GREEDYCUTS i ->
		    fun inst n -> Cutpointset.greed (good_i n i) inst n
		    | INTERVCUTS i ->
			fun inst n -> Cutpointset.interval (good_i n i) inst n
 			| SEQCUTS dxs  -> 
			    fun inst n -> Cutpointset.sequence
			      (try List.nth dxs n with Failure "nth" ->
				 List.hd (List.rev dxs))
			      inst n
   			    | HYBRIDCUTS ns -> let ng,nb,n = ns in Cutpointset.hybrid (ng,nb) n
			    | READCUTS     -> Cutpointset.all (* fake *)
      in
      let get_cpsl inst = 
	match !cut with
	  | READCUTS -> read_cuts (open_in !fcuts)
	  | ALLCUTS
	  | RANDCUTS _
	  | BINCUTS _
	  | INTERVCUTS _ 
	  | GREEDYCUTS _ 
	  | HYBRIDCUTS _
	  | SEQCUTS _ 
	    ->
	      let tmp = ref []
	      in
		for n=0 to inst#n - 1 do
		  tmp :=
		    (if (!cleancuts) then simplify_for inst (f_cuts inst n)
		     else f_cuts inst n
		    )::(!tmp)
		done;
		!tmp
      in begin match !what with

	(* ================================================================= *)
	| NONE -> ()

	(* ================================================================= *)
	| NORMALIZE ->
	    let inst = read_instance (open_in !finst)
	    in
	      if !sort then inst#sort();
	      inst#normalize();
	      print_instance (out()) inst;

	(* ================================================================= *)
	| BINARY ->
	    let inst = read_instance (open_in !finst)
	    in
	      (* binary representation of obs o with monomes ms *)
	    let rec binary o ms =
	      match ms with
		| [] -> "\n"
		| m::[] -> (if m#covers o then "1" else "0")^"\n"
		| m::ms -> (if m#covers o then "1;" else "0;")^(binary o ms)
	    in
	    let mons =
	      if String.compare !fcuts "" == 0
	      then
		(* reads monomes from model *)
		let model = read_model (open_in !fmodel)
		in model#get_monomes()
	      else
		(* reads monomes from cutpointset *)
		let cut = Array.of_list (read_cuts (open_in !fcuts))
		and tmp = ref []
		in for k=0 to Array.length cut - 1 do
		    let v=cut.(k)#variable
		    and x=ref cut.(k)#first
		    in while(Pervasives.compare !x nan <> 0) do
			tmp := add_gt_monome !tmp (new monome v (GT !x));
			x := cut.(k)#next
		      done;
		  done;
		  !tmp
	    in
	      if (names()) then output_string (out()) "name;";
	      output_string (out()) "class;";
	      output_string
		(out())
		((String.concat ";" (List.map string_of_monome mons))^"\n");
	      if !sort then inst#sort();
	      for o=0 to inst#m - 1 do
		let o=inst#observation o
		in
		  if (names()) then fprintf (out()) "%s;" o#name;
		  fprintf (out()) "%d;" o#kind;
		  output_string (out()) (binary o mons)
	      done;

	(* ================================================================= *)
	| PATTERNS b ->
	    let inst = read_instance (open_in !finst)
	    in
	      Pandect.all_with_action
		~pseudoprimes:b
		(fun p -> print_pattern (out()) p; output_char (out()) '\n')
		!deg hom pre hr (get_cpsl inst) inst;

	(* ================================================================= *)
	| UPDATEPAT ->
	    let pan = read_pandect (open_in !fmodel)
	    and inst = read_instance (open_in !finst)
	    in
	    let rec parse ps =
	      match ps with
		| [] -> ()
		| p::ps -> p#compute_for inst; parse ps;
	    in     
	      parse (pan#all_patterns());
	      if !sort then
		List.iter
		  (fun p -> print_pattern (out()) p;output_string (out()) "\n")
		  (List.sort Pattern.compare_index (pan#all_patterns()))
	      else
		print_pandect (out()) pan;
	      
	(* ================================================================= *)
	| CLEANCOV ->
	    let model = read_model (open_in !fmodel)
	    and inst = read_instance (open_in !finst)
	    in
	      model#clean_coverage inst;
	      if !sort then
		List.iter
		  (fun p -> print_pattern (out()) p;output_string (out()) "\n")
		  (List.sort Pattern.compare_index (model#all_patterns()))
	      else 
		print_pandect (out()) (model:>pandect);

 	(* ================================================================= *)
	| CLEANINCL ->
	    let model = read_model (open_in_alt !fmodel !finst)
	    in model#clean_inclusion();
	      if !sort then
		List.iter
		  (fun p -> print_pattern (out()) p; output_string (out()) "\n")
		  (List.sort Pattern.compare_index (model#all_patterns()))
	      else 
		print_pandect (out()) (model:>pandect);

	(* ================================================================= *)
	| PATPROD ->
	    let pan = read_pandect (open_in_alt !fmodel !finst)
	    in let pats = Array.of_list (pan#all_patterns())
	    in let n    = Array.length pats
	    in for i=0 to n - 2 do
		for j=i+1 to n - 1 do
		  let p = pats.(i)#intersection pats.(j)
		  in  
		    print_pattern (out()) p;
		    output_char (out()) '\n';
		done
	      done

	(* ================================================================= *)
	| CUTS ->
	    let inst = read_instance (open_in !finst)
	    in
	      Io.print_cutpointset_list (out()) (get_cpsl inst);

	(* ================================================================= *)
	| ALLNANOMODS min_acc ->
	    let inst = read_instance (open_in !finst)
	    in let parsecps cps =
		let v = cps#variable
		and f = ref cps#first
		in while (Pervasives.compare !f nan != 0) do
		    let p1 = new pattern 1 [new monome v (LT !f)]
		    and p2 = new pattern 2 [new monome v (GT !f)]
		    in p1#compute_for inst;
		      p2#compute_for inst;
		      let m = new model 1 [p1;p2]
		      in m#compute_for inst;
			if m#accuracy >= min_acc then begin
			  print_pandect (out()) (m:>pandect);
			  fprintf (out()) "acc = %f\n" m#accuracy;
			  fprintf (out()) "spe = %f\n" m#specificity;
			  fprintf (out()) "sen = %f\n" m#sensitivity;
			end;
			f := cps#next
		  done;
	    in let rec parsecpsl cpsl =
		match cpsl with
		  | [] -> ()
		  | cps::cpsl -> parsecps cps; parsecpsl cpsl;
	    in parsecpsl (get_cpsl inst)

	(* ================================================================= *)
	| BESTNANOMODS c ->
	    let inst = read_instance (open_in !finst)
	    in let value m = 
		match c with
		  | 'a' -> 
		      m#accuracy +. 0.0005*.(m#specificity +. m#sensitivity)
		  | 'b' ->
		      m#specificity +. m#sensitivity +. 0.001*.m#accuracy
		  | _ ->
		      warning "Unknown nanomodel evaluation (use accuracy)";
		      m#accuracy +. 0.0005*.(m#specificity +. m#sensitivity)

	    in let parsecps cps =
		let v = cps#variable
		and bmodel = ref (new model 1 [])
		and bvalue = ref 0.0
		and f = ref cps#first
		in while (Pervasives.compare !f nan != 0) do
		    let p1 = new pattern 1 [new monome v (LT !f)]
		    and p2 = new pattern 2 [new monome v (GT !f)]
		    in p1#compute_for inst;
		      p2#compute_for inst;
		      let m = new model 1 [p1;p2]
		      in
			m#compute_for inst;
			let v = value m
			in if v > !bvalue then begin
			    bmodel := m;
			    bvalue := v;
			  end;
			  f := cps#next
		  done;
		  print_pandect (out()) (!bmodel:>pandect);
		  fprintf (out()) "acc = %f\n" !bmodel#accuracy;
		  fprintf (out()) "spe = %f\n" !bmodel#specificity;
		  fprintf (out()) "sen = %f\n" !bmodel#sensitivity;
	    in let rec parsecpsl cpsl =
		match cpsl with
		  | [] -> ()
		  | cps::cpsl -> parsecps cps; parsecpsl cpsl;
	    in parsecpsl (get_cpsl inst)

	(* ================================================================= *)
	| GREEDYMOD cov ->
	    let inst = read_instance (open_in !finst)
	    and pand = read_model (open_in !fmodel)
	    in let model = Model.greedy_old cov inst (pand:>pandect)
	    in
	      if !sort then
		List.iter
		  (fun p -> print_pattern (out()) p;output_string (out()) "\n")
		  (List.sort Pattern.compare_index (model#all_patterns()))
	      else 
		print_pandect (out()) (model:>pandect);

	(* ================================================================= *)
	| GREEDYCOVMOD x ->
	    let f,ncov,pcov = x 
	    and inst = read_instance (open_in !finst)
	    and pand = read_model (open_in !fmodel)
	    in let model = Model.greedy f [|ncov; pcov|] inst (pand:>pandect)
	    in
	      if !sort then
		List.iter
		  (fun p -> print_pattern (out()) p;output_string (out()) "\n")
		  (List.sort Pattern.compare_index (model#all_patterns()))
	      else 
		print_pandect (out()) (model:>pandect);

	(* ================================================================= *)
	| SELECTCOVMOD x ->
	    let f,ncov,pcov = x 
	    and inst = read_instance (open_in !finst)
	    and pand = read_model (open_in !fmodel)
	    in let model = Model.select f [|ncov; pcov|] inst (pand:>pandect)
	    in
	      if !sort then
		List.iter
		  (fun p -> print_pattern (out()) p;output_string (out()) "\n")
		  (List.sort Pattern.compare_index (model#all_patterns()))
	      else 
		print_pandect (out()) (model:>pandect);

	(* ================================================================= *)
	| LPSELECT x ->
	    let c,k,mip = x
	    and inst = read_instance (open_in !finst)
	    and pand = read_model (open_in !fmodel)
	    in let l = 
		let l = ref []
		in for i=0 to inst#n-1 do l := (i)::(!l) done; !l
	    in
	    let lpm = new lp_model inst l (pand:>pandect) c k
	    in
	      lpm#solve(mip);
	      print_pandect (out()) ((lpm#get_sol()):>pandect);

	(* ================================================================= *)
	| BOXMOD ->
	    let inst = read_instance (open_in !finst)
	    in let model = Model.boxes (get_cpsl inst) inst 
	    in
	      if !sort then
		List.iter
		  (fun p -> print_pattern (out()) p;output_string (out()) "\n")
		  (List.sort Pattern.compare_index (model#all_patterns()))
	      else 
		print_pandect (out()) (model:>pandect);

	(* ================================================================= *)
	| ACCURACY -> 
	    let model = read_model (open_in !fmodel)
	    and inst  = read_instance (open_in !finst)
	    in model#compute_for inst;
	      fprintf (out()) "acc = %f\n" model#accuracy;
	      fprintf (out()) "spe = %f\n" model#specificity;
	      fprintf (out()) "sen = %f\n" model#sensitivity;

	(* ================================================================= *)
	| DATASCOPEACCURACY -> 
	    let model = read_model (open_in !fmodel)
	    and inst  = read_instance (open_in !finst)
	    in model#compute_for inst;
	      fprintf (out()) "acc = %f\n" (model#datascope_accuracy);
	      fprintf (out()) "spe = %f\n" model#specificity;
	      fprintf (out()) "sen = %f\n" model#sensitivity;

	(* ================================================================= *)
	| BATCH ->
	    let inst = read_instance (open_in !finst)
	    and input = open_in ""
	    in
	      begin try
		while (true)
		do
		  let model = read_model ~noer:false input
		  in 
		    if model#is_empty () then failwith "done";
		    model#compute_for inst;
		    fprintf (out()) "%f %f %f %f\n"
		      model#accuracy
		      model#datascope_accuracy
		      model#specificity
		      model#sensitivity;
		    flush (out());
		done;
	      with Failure "done" -> ()
	      end

	(* ================================================================= *)
	| CLASSIFICATION ->
	    let model = read_model (open_in !fmodel)
	    and inst = read_instance (open_in !finst)
	    in 
	      if !sort then inst#sort();
	      for o=0 to inst#m - 1 do
		let o = inst#observation o
		in
		  if (names()) then fprintf (out()) "%s;" (o#name);
		  fprintf (out()) "%d;%d;%d\n"
		    o#index
		    o#kind
		    (model#classification o)
	      done;

	(* ================================================================= *)
	| DISCRIMINANT ->
	    let model = read_model (open_in !fmodel)
	    and inst  = read_instance (open_in !finst)
	    in
	    let n = [|float_of_int (model#count_of_kind 0);
		      float_of_int (model#count_of_kind 1)|]
	    and s = [| -1.; 1.|]
	    and pat = Array.of_list (model#all_patterns())
	    in
	      if !sort then inst#sort();
	      for o=0 to inst#m - 1 do
		let sco = ref 0.0
		and o = inst#observation o
		in for k=0 to Array.length pat -1 do
		    let p = pat.(k)#kind
		    in if (p==0 || p==1) then
			if (pat.(k)#covers o) then
			  sco := !sco +. s.(p)/.n.(p);
		  done;
		  let sck =
		    if Pervasives.compare !sco 0.0 == 0 then -1
		    else if !sco > 0.0 then 1 else 0
		  in 
		    if (names()) then fprintf (out()) "%s;" (o#name);
		    fprintf (out()) "%d;%d;%d;%.4f\n" o#index o#kind sck !sco;
	      done;

	(* ================================================================= *)
	| COVERAGE ->
	    let model = read_model (open_in !fmodel)
	    and inst = read_instance (open_in !finst)
	    in
	    let pat = Array.of_list 
	      (if !sort then
		 List.sort (Pattern.compare_index) (model#all_patterns())
	       else (model#all_patterns()))
	    in
	      for k=0 to Array.length pat - 1 do
		pat.(k)#compute_for inst;
		print_pattern (out()) pat.(k);
		fprintf (out()) " : %2d %2d\n"
		  (int_of_float
		     ((pat.(k)#prevalence 0)
		      *. (float_of_int (inst#count_of_kind 0))))
		  (int_of_float
		     ((pat.(k)#prevalence 1)
		      *. (float_of_int (inst#count_of_kind 1))))
	      done;

	(* ================================================================= *)
	| INCIDENCES ->
	    warning ("incidences is deprecated in ladoscope, "
		     ^"you should use the 'matrices' program");
	    let model = read_model (open_in !fmodel)
	    and inst = read_instance (open_in !finst)
	    in let pat = Array.of_list (model#all_patterns())
	    in
	      if (!i=0) then begin (* plain *)
		(* no headings *)
		(* data *)
		for k=0 to Array.length pat - 1 do
		  for o=0 to inst#m - 1 do
		    let obs = inst#observation o
		    in fprintf (out()) "%c " (if pat.(k)#covers obs then '1' else '0')
		  done;
		  fprintf (out()) "\n";
		done
	      end
	      else if (!i=1) then begin (* CSV *)
		(* headings *)
		for o=0 to inst#m - 1 do
		  let obs = inst#observation o
		  in fprintf (out()) ";%c%d" (if obs#kind=0 then '-' else '+') o
		done;
		fprintf (out()) "\n";
		(* data *)
		for k=0 to Array.length pat - 1 do
		  fprintf (out()) "%c%d" (if pat.(k)#kind = 0 then 'N' else 'P') k;
		  for o=0 to inst#m - 1 do
		    let obs = inst#observation o
		    in fprintf (out()) ";%c" (if pat.(k)#covers obs then '1' else '0')
		  done;
		  fprintf (out()) "\n";
		done
	      end
	      else if (!i=2) then begin (* AMPL *)
		(* headings *)
		fprintf (out()) "param:   ";
		for o=0 to inst#m - 1 do fprintf (out()) "%6d" o done;
		fprintf (out()) " := \n";
		(* data *)
		for k=0 to Array.length pat - 1 do
		  fprintf (out()) " %6d  " k;
		  for o=0 to inst#m - 1 do
		    let obs = inst#observation o
		    in fprintf (out()) "%6d" (if pat.(k)#covers obs then 1 else 0)
		  done;
		  fprintf (out()) "\n";
		done;
		fprintf (out()) ";\n";
	      end

	(* ================================================================= *)
	| VARINC ->
	    warning ("incidences is deprecated in ladoscope, "
		     ^"you should use the 'matrices' program");
	    let model = read_model (open_in_alt !fmodel !finst)
	    in 
	    let rec maxvarms ms =
	      match ms with
		| []    -> 0
		| m::ms -> max (m#variable) (maxvarms ms)
	    in
	    let rec maxvarps ps = 
	      match ps with
		| []    -> 0
		| p::ps -> max (maxvarms (p#monomes())) (maxvarps ps)
	    in let pat = Array.of_list (model#all_patterns())
	       and var = maxvarps (model#all_patterns())
	    in 
	      if (!i=0) then begin (* plain *)
		(* no headings *)
		(* data *)
		for i=0 to (Array.length pat) -1 do
		  for k=0 to var do
		    fprintf (out()) "%d " (if pat.(i)#contains k then 1 else 0) done;
		  fprintf (out()) "\n";
		done
	      end
	      else if (!i=1) then begin (* CSV *)
		(* headings *)
		for k=0 to var do fprintf (out()) ";%d" k done; fprintf (out()) "\n";
		(* data *)
		for i=0 to (Array.length pat) -1 do
		  fprintf (out())  "%d" i;
		  for k=0 to var do
		    fprintf (out()) ";%d" (if pat.(i)#contains k then 1 else 0) done;
		  fprintf (out()) "\n";
		done
	      end
	      else if (!i=2) then begin (* AMPL *)
		(* headings *)
		fprintf (out()) "param: ";
		for k=0 to var do fprintf (out()) "%6d" k done;
		fprintf (out()) " := \n";
		(* data *)
		for i=0 to (Array.length pat) -1 do
		  fprintf (out()) " %6d  " i;
		  for k=0 to var do
		    fprintf (out()) "%6d" (if pat.(i)#contains k then 1 else 0) done;
		  fprintf (out()) "\n";
		done;
		fprintf (out()) ";\n";
	      end

	(* ================================================================= *)
	| VARIABLES ->
	    let model = read_model (open_in_alt !fmodel !finst)
	    and x = 5000
	    in
	    let pat = Array.of_list (model#all_patterns())
	    and neg_lt = Array.make x 0
	    and neg_gt = Array.make x 0
	    and pos_lt = Array.make x 0
	    and pos_gt = Array.make x 0
	    in for k=0 to Array.length pat - 1 do
		let ms = Array.of_list (pat.(k)#monomes())
		in for m=0 to Array.length ms - 1 do
		    match ms.(m)#inequality with
		      | LT _  -> if  pat.(k)#kind = 0 then
			  neg_lt.(ms.(m)#variable)<-neg_lt.(ms.(m)#variable)+1
			else
			  pos_lt.(ms.(m)#variable)<-pos_lt.(ms.(m)#variable)+1
		      | GT _  -> if  pat.(k)#kind = 0 then
			  neg_gt.(ms.(m)#variable)<-neg_gt.(ms.(m)#variable)+1
			else
			  pos_gt.(ms.(m)#variable)<-pos_gt.(ms.(m)#variable)+1
		      | BOX _ -> if  pat.(k)#kind = 0 then begin
			  neg_lt.(ms.(m)#variable)<-neg_lt.(ms.(m)#variable)+1;
			  neg_gt.(ms.(m)#variable)<-neg_gt.(ms.(m)#variable)+1
			end else begin
			  pos_lt.(ms.(m)#variable)<-pos_lt.(ms.(m)#variable)+1;
			  pos_gt.(ms.(m)#variable)<- pos_gt.(ms.(m)#variable)+1
			end
		  done
	      done;
	      for i=0 to x-1 do
		if (neg_lt.(i)+neg_gt.(i)+pos_lt.(i)+pos_gt.(i)) > 0 then
		  fprintf (out())
		    "%d : %2d %2d %2d %2d\n"
		    i neg_lt.(i) neg_gt.(i) pos_lt.(i) pos_gt.(i)
	      done;

	(* ================================================================= *)
	| SCORE_SUMCARD
	| SCORE_SUMRATIO
	| SCORE_SUMPREV
	| SCORE_HRVALUE ->
	    let model = read_model (open_in !fmodel)
	    and inst  = read_instance (open_in !finst)
	    in
	    let f p = match !what with
	      | SCORE_SUMCARD  -> 1.0
	      | SCORE_SUMRATIO -> 1.0
		  /. (float_of_int (model#count_of_kind (p#kind)))
	      | SCORE_SUMPREV  -> p#max_prevalence
	      | SCORE_HRVALUE  -> (if (p#kind=0) then -1.0 else 1.0)
		  *. p#hazardratio
	      | _ -> failwith "bug"
	    in let pat = Array.of_list (model#all_patterns())
	    in
	      if !sort then inst#sort();
	      for o=0 to inst#m - 1 do
		let scopos = ref 0.0
		and sconeg = ref 0.0
		and o = inst#observation o
		in for k=0 to Array.length pat -1 do
		    if (pat.(k)#kind = 0) then begin
		      if pat.(k)#covers o
		      then sconeg := !sconeg +. (f pat.(k))
		    end else if (pat.(k)#kind > 0) then begin
		      if pat.(k)#covers o
		      then scopos := !scopos +. (f pat.(k))
		    end done;
		  let sck =
		    if !scopos = !sconeg
		    then -1 else if !scopos > !sconeg then 1 else 0
		  and sco = !scopos -. !sconeg
		  in
		    if (names()) then fprintf (out()) "%s;" (o#name);
		    fprintf (out()) "%d;%d;%d;%.4f\n"
		      o#index o#kind sck sco;
	      done;

	(* ================================================================= *)
	| SCORE_BESTPREV
	| SCORE_BESTHR ->
	    let model = read_model (open_in !fmodel)
	    and inst  = read_instance (open_in !finst)
	    in let f p = match !what with
	      | SCORE_BESTPREV -> p#max_prevalence
	      | SCORE_BESTHR   -> p#hazardratio
	      | _ -> failwith "bug"
	    in let pat = Array.of_list (model#all_patterns())
	    in
	      if !sort then inst#sort();
	      for o=0 to inst#m - 1 do
		let bestkin = ref (-1)
		and bestsco = ref 0.0
		and o = inst#observation o
		in for k=0 to Array.length pat -1 do
		    if (f pat.(k)) > !bestsco && pat.(k)#covers o
		    then begin
		      bestkin := pat.(k)#kind;
		      bestsco := f pat.(k)
		    end done;
		  let sck = !bestkin
		  and sco = if !bestkin = -1 then 0.0 else
		    if !bestkin > 0 then !bestsco else (-. !bestsco)
		  in
		    if (names()) then fprintf (out()) "%s;" (o#name);
		    fprintf (out()) "%d;%d;%d;%.4f\n"
		      o#index (model#classification o) sck sco;
	      done;

	(* ================================================================= *)
	| SCORE_CLOSESTGRIDOPPONENT ->
	    let model = read_model (open_in !fmodel)
	    and inst  = read_instance (open_in !finst)
	    in
	    let closest_grid_opponant model inst =
	      let cp = Array.of_list (model#cutpointsets())
	      in
		if !sort then inst#sort();
		for o=0 to inst#m-1 do
		  let  o = inst#observation o
		  in let p = o#clone()
		     and k = model#classification o
		     and d = ref infinity
		  in for c = 0 to Array.length cp - 1 do
		      let x = ref cp.(c)#first
		      and v = o#value (cp.(c)#variable)
		      in while Pervasives.compare !x nan <> 0 do
			  p#set_value cp.(c)#variable !x;
			  if (model#classification p) = (1-k) then begin
			    if (!x < v) && (v -. !x < !d) then d := v -. !x;
			    if (!x > v) && (!x -. v < !d) then d := !x -. v;
			  end;
			  x := cp.(c)#next
			done;
			p#set_value (cp.(c)#variable) v
		    done;
		    let sck,sco =
		      if      Pervasives.compare !d nan == 0 then -1,  0.0
		      else if Pervasives.compare !d 0.0 == 0 then -1,  0.0
		      else if Pervasives.compare  k 0   >  0 then  1,  !d
		      else                                        -1, -.(!d)
		    in 
		      if (names()) then fprintf (out()) "%s;" (o#name);
		      fprintf (out()) "%d;%d;%d;%.4f\n" o#index k sck sco;
		done
	    in
	      closest_grid_opponant model inst;
	      
	(* ================================================================= *)
	| SCORE_CLOSESTOPPONENT haltmax ->
	    let model = read_model (open_in !fmodel)
	    and ins = read_instance (open_in !finst)
	    in
	    let rec opponant_distance ps o c =
	      trace TRACE_DEBUG
		(sprintf "  opponant_distance [%d]\n" (List.length ps));
	      let rec parse_proj ys =
		trace TRACE_DEBUG
		  (sprintf "    parse_proj [%d]\n" (List.length ys));
		match ys with
		  | [] -> infinity
		  | y::ys -> let ds = parse_proj ys
			     and c1 = model#classification_with_epsilon 0.0001 y
		    in
		      if (c1 >= 0 && c1 <> c) then min ds (o#distance y)
		      else ds
	      in
		match ps with
		  | [] -> infinity
		  | p::ps
		    -> let ys = p#projections o
		    in let d1,d2 = parse_proj ys, opponant_distance ps o c 
		    in min d1 d2           
	    in
	    let closest_opponant model inst =
	      let p0 = trace TRACE_SOME "Get patterns...\n";
		model#all_patterns()
	      in
		if !sort then inst#sort();
		for o=0 to inst#m-1 do
		  trace TRACE_SOME (sprintf "Observation %d...\n" o);
		  let  o = inst#observation o
		  in let k = model#classification o
		  in let d,p1  = if k < 0 then ref 0.0, ref []
		    else ref (opponant_distance p0 o k), ref p0
		  in
		  let halt = ref haltmax in
		    while (List.length !p1 > 0) && (!halt > 0) do
		      p1 := Pandect.intersection p0 !p1;
		      d  := opponant_distance !p1 o k;
		      halt := !halt - 1;
		    done;
		    let sck,sco =
		      if      Pervasives.compare !d nan == 0 then -1,  0.0
		      else if Pervasives.compare !d 0.0 == 0 then -1,  0.0
		      else if Pervasives.compare  k 0   >  0 then  1,  !d
		      else                                        -1, -.(!d)
		    in
		      if (names()) then fprintf (out()) "%s;" (o#name);
		      fprintf (out()) "%d;%d;%d;%.4f\n" o#index k sck sco;
		done
	    in closest_opponant model ins;

	(* ================================================================= *)
	| SCORE_DISTANCE
	| SCORE_PREVDISTANCE
	| SCORE_HRDISTANCE ->
	    let model = read_model (open_in !fmodel)
	    and inst = read_instance (open_in !finst)
	    and countout = !countout
	    in let w p = match !what with
		SCORE_DISTANCE -> 1.0
	      | SCORE_PREVDISTANCE -> p#max_prevalence
	      | SCORE_HRDISTANCE -> p#hazardratio
	      | _ -> failwith "bug"
	    in
	    let pats = Array.of_list (model#all_patterns())
	    in
	      if !sort then inst#sort();
	      for o=0 to inst#m - 1 do
		let o = inst#observation o
		and d = ref 0.0
		in for p=0 to Array.length pats -1 do
		    d := !d
		    +.
		    (if (pats.(p)#kind=0) then -1.0 else 1.0)
		    *. (pats.(p)#distance_of_observation o countout)
		    *. (w pats.(p));
		  done;
		  let k = o#kind
		  and sck = if !d < 0.0 then 0 else if !d > 0.0 then 1 else -1
		  and sco = !d
		  in
		    if (names()) then fprintf (out()) "%s;" (o#name);
		    fprintf (out()) "%d;%d;%d;%.4f\n"
		      o#index k sck sco;
	      done

	(* ================================================================= *)
	end;
	(* terminates program *)
	0;
  with
    | _ as ex -> terminal_catch ex
;;
