(* $Id: ladoscript.ml 14 2007-06-29 08:14:54Z pierre $ *)
(** ========================================================================
    ladoscript: scripting language for LAD
    Copyright (C) Pierre Lemaire, 2005.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ======================================================================== *)

open Messages;;
open Misc;;

open Ladoscope;;
open Classifier;;
open Matrices;;
open Stat_instance;;
open Stat_model;;
open Sampler;;

open String;;
open Printf;;
open Hashtbl;;

let prompt = "lad > ";;
let interactive_name = "<stdin>";;

(* ========================================================================= *)
exception Command_exit of int;;
exception Command_failure of string;;
exception Command_failed of string;;

let failure s = raise (Command_failure s);;
let failed  s = raise (Command_failed s);;

(* ========================================================================= *)
(* === script_type class =================================================== *)
(* ========================================================================= *)
class script_type name =
object(self)
  (* ----------------------------------------------------------------------- *)
  (* channel from which the commands are read *)
  val file = 
    if (compare name interactive_name = 0) then stdin
    else Pervasives.open_in name

  (* current line number *)
  val mutable line = 0

  (* variables *)
  val variables  = Hashtbl.create 10 

  (* ----------------------------------------------------------------------- *)
  (* file name *)
  method name = name

  (* current line number *)
  method line = line

  (* close the file associated to this script*)
  method close () = close_in file

  (* read and return the next line (not command!) *)
  method next_line () =
    line <- line + 1;
    Str.replace_first (Str.regexp "^[ \t]+") "" (input_line file);
    
  (* read and return the next command *)
  method command ?(split=true) () =
    try
      let x = ref (self#next_line())
      in if  !x <> "" then begin
	  (* read following line if this one ends with \ *)
	  while(!x.[String.length !x - 1] = '\\') do
	    !x.[String.length !x - 1] <- ' ';
	    try x := (!x)^(self#next_line());
	    with End_of_file -> failure "unexpected end of command"
	  done;
	  
	  (* expand or not variables *)
	  if !x.[0] = '\\' then !x.[0] <- ' ' else  x := self#expand_var !x;
	  
	  (* split, removing white spaces *)
	  if (split) then
	    (Array.of_list (Str.split (Str.regexp "[ \t]+") !x))
	  else
	    [| !x |]
	end
	else [||]
    with End_of_file -> raise (Command_exit 1)

  (* export the variables, NOT replacing existing ones *)
  method export_var (s: script_type) = 
    Hashtbl.iter
      (fun var _ ->
	 if not (s#exist_var var)
	 then s#add_var var (self#get_var var)
      )
      variables

  (* add a variable (hide but not replace previous occurences) *)
  method add_var (v:string) (x:string) = Hashtbl.add variables v x
    
  (* set the value of a variable (created if needed) *)
  method set_var v x = Hashtbl.replace variables v x
    
  (* unset the value of a variable *)
  method unset_var v = Hashtbl.remove variables v

  (* get the value of a variable (raise Not_found if needed) *)
  method get_var v = Hashtbl.find variables v

  (* if the variable exists *)
  method exist_var v = 
    try ignore (self#get_var v); true
    with Not_found -> false

  (* deletes all variables *)
  method clear_var = Hashtbl.clear variables

  (* prints all variables *)
  method print_var x = Hashtbl.iter (fprintf x "%s = %s\n") variables

  (* change every $x by the value of x*)
  method expand_var s = 
    let r = ref ""
    in begin
	let i = ref 0 
	and n = String.length s
	in try while (true) do 
	    let j,k = ref 0, ref 0
	    in
	      j := index_from s !i '$';
	      r := !r^(String.sub s !i (!j - !i));
	      i := !j+1;
	      if !i < n then begin
		if s.[!i] = '(' then begin (* long name *)
		  i := !i+1;
		  (try k := index_from s !i ')';
		   with Not_found -> failure "bad variable name");
		  j := !k - 1;
		end else begin (* short name *)
		  j := !i;
		  k := !i;
		end;
		(try
		   r := !r^(self#get_var (String.sub s !i (!j - !i + 1)));
		 with Not_found -> ());
 		i := !k + 1;
	      end
	  done;
	  with Not_found -> r := !r^(String.sub s !i (n - !i));
      end;
      !r

end;;

(* ========================================================================= *)
(* === commands ============================================================ *)
(* ========================================================================= *)
let rec load_cmd ?import arg =
  if (Array.length arg < 2) then failure "load: nothing to load";
  let interactive =  (compare arg.(1) interactive_name = 0)
  and script =
    try new script_type arg.(1)
    with Sys_error s -> failure (sprintf "load: %s" s)
  in
    begin match import with
	None -> ()
      | Some s -> s#export_var script;
    end;
    while (true) do
      try
	if (interactive) then begin printf "%s" prompt; flush_all(); end;
	exec_command (script#command()) script;
      with
	| Command_failed s ->
	    if interactive then error (-1) s
	    else raise (Command_failed s)
	| Command_failure s ->
	    if interactive then error (-1) s
	    else raise (Command_failed
	      (sprintf "%s, line %d of %s" s script#line script#name))
	| Command_exit i ->
	    if i=0 then ()
	    else begin
	      script#close();
	      raise (Command_exit (i-1))
	    end
	| _ as ex -> 
	    error (-1) (sprintf "unexpected error (%s) for %s, line %d"
			  (Printexc.to_string ex)
			  arg.(0) script#line);
	    raise ex
    done;

(* ========================================================================= *)
and exec_command arg script = 
  (* ----------------------------------------------------------------------- *)
  let rec comment_cmd _ = ()

  (* ----------------------------------------------------------------------- *)
  and read_block endblock =
    let blockscript = Filename.temp_file "block" "script"
    in let file =
	try Pervasives.open_out blockscript
	with Sys_error _ -> failure "(block): unexpected i/o error"
       and ok = ref false
    in
      while (not !ok)
      do
	let l = script#next_line()
	in
	  if Pervasives.compare l endblock = 0 then ok := true
	  else begin
	    output_string file l;
	    output_string file "\n"
	  end
      done;
      close_out file;
      blockscript

  (* ----------------------------------------------------------------------- *)
  and amnesia_cmd _ = script#clear_var

  (* ----------------------------------------------------------------------- *)
  and echo_cmd xs = 
    fprintf (out()) "%s\n" (concat " " (List.tl (Array.to_list xs)));
    
  (* ----------------------------------------------------------------------- *)
  and exec cmd xs = 
    if cmd xs > 0 then failure "unknown cause"

  (* ----------------------------------------------------------------------- *)
  and exit_cmd xs = 
    let i = try  int_of_string (xs.(1)) with _ ->  1
    in raise (Command_exit i);

  (* ----------------------------------------------------------------------- *)
  and expr_cmd xs = 
    try
      let a,op,b =
	int_of_string xs.(1),
	xs.(2),
	int_of_string xs.(3)
      in match op with
	  "+" -> fprintf (out()) "%d" (a + b)
	| "-" -> fprintf (out()) "%d" (a - b)
	| "*" -> fprintf (out()) "%d" (a * b)
	| "/" -> fprintf (out()) "%d" (a / b)
	| "%" -> fprintf (out()) "%d" (a mod b)
	| _ -> failure "expr: invalid operator"
    with
	_ -> failure "expr: invalid arguments"

  (* ----------------------------------------------------------------------- *)
  and extern_cmd xs = 
    if Sys.command (concat " " (List.tl (Array.to_list xs))) > 0
    then failure "unknown cause"

  (* ----------------------------------------------------------------------- *)
  and for_cmd xs = 
    let var    = xs.(1)
    and values = Array.to_list (Array.sub xs 2 (Array.length xs - 2))
    and subscript = read_block "endfor"
    in
      List.iter
	 (fun v ->
	    script#add_var var v;
	    try load_cmd ~import:script [| "load"; subscript |];
	    with Command_exit _ -> ();
	    script#unset_var var;
	 )
	 values
      
  (* ----------------------------------------------------------------------- *)
(*   and let_cmd xs =  *)
(*     if (Array.length xs > 1) then *)
(*       script#set_var xs.(1) (concat " " (List.tl (List.tl (Array.to_list xs)))) *)
(*     else *)
(*       script#print_var (out()) *)

  (* ----------------------------------------------------------------------- *)
  and rand_cmd xs = 
    try
      let a =
	(try int_of_string xs.(1)
	 with Invalid_argument("index out of bounds")  -> 100)
      in fprintf (out()) "%d " (Random.int a);
    with
	_ -> failure "rand: invalid arguments"
	  
  (* ----------------------------------------------------------------------- *)
  and randseed_cmd xs = 
    try
      Random.init (int_of_string xs.(1))
    with _ -> Random.self_init()

  (* ----------------------------------------------------------------------- *)
  and seq_cmd xs = 
    try
      let a,b,c =
	int_of_string xs.(1),
	int_of_string xs.(2),
	(try int_of_string xs.(3)
	 with Invalid_argument("index out of bounds")  -> 1)
      in let i = ref a 
      in while (if c>0 then !i <= b else !i >= b) do
	  fprintf (out()) "%d " !i;
	  i := !i + c;
	done
    with
	_ -> failure "seq: invalid arguments"
    
  (* ----------------------------------------------------------------------- *)
  and set_cmd xs = 
    if (Array.length xs > 1) then
      script#set_var xs.(1) (concat " " (List.tl (List.tl (Array.to_list xs))))
    else
      script#print_var (out())

  (* ----------------------------------------------------------------------- *)
  and setexec_cmd (xs: string array) = 
    if (Array.length xs > 2) then
      let var = xs.(1)
      and value =
	let save_outputs = get_current_outputs()
	and tmpfile = Filename.temp_file "aze" "rty"
	and ys = Array.sub xs 2 (Array.length xs - 2)
	in
	  ys.(0) <- xs.(2);
	  set_output_file tmpfile;
	  exec_command ys script;
	  flush_all();
	  close_out (out());
	  restore_outputs save_outputs;
	  let file = try
	    Pervasives.open_in tmpfile
	  with Sys_error _ -> failure "setexec: unexpected i/o error"
	  in let x = ref (input_line file)
	  in begin try while (true) do x := !x^"\n"^(input_line file); done;
	    with End_of_file -> close_in file
	    end;
	    !x
      in
	script#set_var var value
    else failure "setexec: no command"

  (* ----------------------------------------------------------------------- *)
  and substitute_cmd xs =
    if (Array.length xs > 1) then begin
      try
	let file = new script_type xs.(1)
	in
	  script#export_var file;
	  try while (true) do
	    echo_cmd (Array.append [|"echo"|] (file#command ~split:false ()));
	  done
	  with
	    | Command_exit _ -> ()
      with Sys_error s -> failure (sprintf "substitute: %s" s)
    end
    else failure "substitute: no file"

  (* ----------------------------------------------------------------------- *)
  and unset_cmd xs =
    if (Array.length xs > 1) then script#unset_var xs.(1)
    else failure "undef: no variable"

  (* ----------------------------------------------------------------------- *)
  and whoami_cmd _ = fprintf (out()) "%s\n" script#name;
      
 in
  (* ----------------------------------------------------------------------- *)
  (* ----------------------------------------------------------------------- *)
  (* list of commands with name, function, and help message *)
  let commands = [
    ("#",             (fun xs -> comment_cmd xs),
       "introduce a comment");
    ("amnesia",       (fun xs -> amnesia_cmd xs),
       "remove all variables");
    ("classifier",    (fun xs -> exec classifier xs),
       "run classifier");
    ("echo",          (fun xs -> echo_cmd xs),
       "print its arguments");
    ("exit",          (fun xs -> exit_cmd xs),
       "exit normaly");
    ("expr",          (fun xs -> expr_cmd xs),
       "evaluate integer expressions");
    ("extern",        (fun xs -> extern_cmd xs),
       "run its arguments as an extern program");
    ("for",           (fun xs -> for_cmd xs),
       "for loop (for [var] [values] .. endfor)");
    ("ladoscope",     (fun xs -> exec ladoscope xs),
       "run ladoscope");
    ("load",          (fun xs -> load_cmd xs),
       "load and run a script");
    ("matrices",      (fun xs -> exec matrices xs),
       "run matrices");
    ("rand",          (fun xs -> rand_cmd xs),
       "print a randomly generated random number");
    ("randseed",      (fun xs -> randseed_cmd xs),
       "initialise the random number generator");
    ("sampler",       (fun xs -> exec sampler xs),
       "run sampler");
    ("seq",           (fun xs -> seq_cmd xs),
       "create a sequence of integers");
    ("set",           (fun xs -> set_cmd xs),
       "set the value of a variable / print all variables");
    ("setexec",       (fun xs -> setexec_cmd xs),
       "set the value of a variable to be the stdout of the command");
    ("stat_instance", (fun xs -> exec stat_instance xs),
       "run stat_instance");
    ("stat_model",    (fun xs -> exec stat_model xs),
       "run stat_model");
    ("substitute",    (fun xs -> substitute_cmd xs),
       "print a file, expanding the variables in it");
    ("unset",         (fun xs -> unset_cmd xs),
       "remove a variable");
    ("whoami",        (fun xs -> whoami_cmd xs),
       "print the current script name");
  ]
  in
  (* ----------------------------------------------------------------------- *)
  (* ----------------------------------------------------------------------- *)
  let rec help cs x =
    match cs with
      | [] -> begin
	  try ignore(Str.search_forward x "help" 0);
	    fprintf stderr "%20s: %s\n" "help" "print this help"
	  with Not_found -> ();
	  try ignore(Str.search_forward x "h" 0);
	    fprintf stderr "%20s: %s\n" "h" "print this help"
	  with Not_found -> ();
	end
      | c::cs ->
	  let (n,_,h) = c
	  in begin
	    try 
	      ignore(Str.search_forward x n 0);
	      fprintf stderr "%20s: %s\n" n h;
	    with Not_found -> ()
	    end;
	    help cs x;
  in
  let rec run cs =
    match cs with
      | [] ->
	  warning (sprintf "unknow command %s, line %d [ignored]\n%s"
		     arg.(0)
		     script#line
		     "type 'help' for the list of all commands"
		  );
      | c::cs ->
	  let (n,f,_) = c
	  in
	    if (compare arg.(0) n = 0) then
	      try f arg
	      with Arg.Bad _ ->
		(Array.iter (printf "%s ") arg);
		raise (Command_failure (sprintf "%s: invalid arguments" n))
	    else run cs
  in
    if (Array.length arg > 0) then
      let save_outputs = get_current_outputs()
      in
	if (compare arg.(0) "help" = 0) || (compare arg.(0) "h" = 0)
	then
	  help
	    commands
	    (Str.regexp (if Array.length arg >= 2 then arg.(1) else ""))
	else run commands;
	flush_all();
	restore_outputs save_outputs;
;;

(* ========================================================================= *)
(* === main program ======================================================== *)
(* ========================================================================= *)
let ladoscript () =
try
  let name = "ladoscript"
  and author = "Pierre Lemaire"
  and version = "2007-06-29"
  and copyright = "2005"
  and description = "an interpreter for the ladascope suite"
  and usage_msg = "usage is: "^Sys.argv.(0)^" [options] [script]"
  and scriptfile = ref interactive_name
  in let arglist =
    (common_options (name,version,description,author,copyright))
    @[
(* --- xxx ----------------------------------------------------------------- *)
    ]
  in
    (* initiate program *)
    Arg.parse (Arg.align arglist) (fun s -> scriptfile := s) usage_msg;
    flush_all();
    load_cmd [| "load"; !scriptfile |];
    (* terminates program *)
    close_messages();
    0
with
  | Command_failed s  -> error 1 s; 0
  | Command_failure s -> error 1 s; 0
  | Command_exit _ -> 0
  | End_of_file -> 0
  | _ as ex -> terminal_catch ex
;;
(* ========================================================================= *)
if !Sys.interactive then () else exit (ladoscript());;
