(* $Id: matrices.ml 14 2007-06-29 08:14:54Z pierre $ *)
(** ========================================================================
    matrices: produces matrices for LAD
    Copyright (C) Pierre Lemaire, 2005.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ======================================================================== *)

open Messages;;
open Misc;;
open Instance;;
open Pandect;;
open Model;;
open Io;;

open Printf;;
open Lexing;;

type matrix_type =
    NONE 
  | CLASSES
  | VARPAT of bool | PATVAR of bool
  | OBSPAT of bool | PATOBS of bool;;

(* ========================================================================= *)
let matrices arg =
try
  let name = "matrices"
  and author = "Pierre Lemaire"
  and version = "2007-06-29"
  and copyright = "2005"
  and description = "computes LAD-related matrices"
  and usage_msg =
    "usage is: "^Sys.argv.(0)^" [options]"
  and instfile = ref ""
  and modelfile = ref ""
  and defaultfile = ref ""
  and forcenames = ref false
  and sep = ref " "
  and eol = ref "\n"
  and matrix = ref NONE
  and align = ref true
  and alignchar = ref ' '
  and sort = ref true
  in let arglist =
    (common_options (name,version,description,author,copyright))
    @(eol_options)
    @[

     ("-sort",
        Arg.Set sort,
        " sorts the output by reading order");
     ("-nosort",
        Arg.Clear sort,
        " does not sort the output.");

(* --- i/o ----------------------------------------------------------------- *)
     ("-inst",
        Arg.String (fun s -> instfile := s),
        "(file) the instance.");
     ("-mod",
        Arg.String (fun s -> modelfile := s),
        "(file) the model.");

     ("-sep",
        Arg.String (fun s -> sep := s),
        "(string) the column separtor to use.");
     
     ("-forcenames",
       Arg.Set forcenames,
       " force names to be displayed.");

     ("-align",
        Arg.Set align,
        " align display in column.");
     ("-noalign",
        Arg.Clear align,
        " do not align display in column.");
     ("-alignchar",
        Arg.String (fun s -> alignchar := s.[0]),
        "(char) character use to fill columns if align.");

     ("-csv",
        Arg.Unit (fun () -> sep := ","; align := false;),
        " use CSV format");

(* --- matrices ------------------------------------------------------------ *)
     ("-classes",
        Arg.Unit (fun () -> matrix := CLASSES),
        " classes.");

     ("-varpat",
        Arg.Unit (fun () -> matrix := VARPAT false),
        " variables x patterns.");

     ("-patvar",
        Arg.Unit (fun () -> matrix := PATVAR false),
        " patterns x variables.");

     ("-svarpat",
        Arg.Unit (fun () -> matrix := VARPAT true),
        " signed variables x patterns (blockers/promoters).");

     ("-spatvar",
        Arg.Unit (fun () -> matrix := PATVAR true),
        " signed patterns x variables (blockers/promoters).");
     
     ("-obspat",
        Arg.Unit (fun () -> matrix := OBSPAT false),
        " observations x patterns.");

     ("-patobs",
        Arg.Unit (fun () -> matrix := PATOBS false),
        " patterns x observations.");

     ("-sobspat",
        Arg.Unit (fun () -> matrix := OBSPAT true),
        " signed observations x patterns.");

     ("-spatobs",
        Arg.Unit (fun () -> matrix := PATOBS true),
        " signed patterns x observations.");

    ]
  in
    (* initiate program *)

    if (Array.length arg = 0) then
      Arg.parse (Arg.align arglist) (fun s ->defaultfile := s) usage_msg
    else begin
      let x = ref 0
      in
	Arg.parse_argv
	  ~current:x
	  arg
	  (Arg.align arglist)
	  (fun s -> defaultfile := s) usage_msg
    end;
    flush_all();
    Observation.set_compare_function Observation.compare_index;
    Pattern.set_compare_function Pattern.compare_index;

    let width_int x = String.length (string_of_int x)
    in
    let fill_int w x =
      let s = string_of_int x
      in let n = w - String.length s
      in let r = if (n > 0) then String.make n !alignchar else ""
      in r^s
    in
    let rec print_index w k xs = match xs with
      | [] -> fprintf (out()) "%s" !eol
      | _::xs -> fprintf  (out()) "%s%s" !sep (fill_int w k);
          print_index w (k+1) xs
    in
    begin match !matrix with

    (* =================================================================== *)
    | NONE -> ()

    (* =================================================================== *)
    | CLASSES ->
	let inst = read_instance (open_in_alt !instfile !defaultfile)
	in let m = inst#m - 1 
	   and k = inst#k - 1
	in let w1, w2 = if (!align)
	   then (width_int m, width_int k)
	   else (0,0)
	in
	  if !sort then inst#sort();
	  for o=0 to m do
	    if (!forcenames) then
	      fprintf (out()) "%s%s" (inst#observation o)#name !sep;
	    fprintf (out()) "%s%s%s"
	      (if names() then sprintf "%s%s" (fill_int w1 o) !sep else "")
	      (fill_int w2 (inst#observation o)#kind)
	      !eol
	  done

    (* =================================================================== *)
    | VARPAT sign ->
	let model = read_model (open_in_alt !modelfile !defaultfile)
	in let pat =
	    if !sort then
	      List.sort Pattern.compare_index (model#all_patterns())
	    else (model#all_patterns())
	in let var = model#get_variables()
	in let vmax = List.fold_left max (-1) var
	in let w0,w1 = if (!align)
	  then (width_int vmax,
	        width_int (List.length pat -1))
	  else (0,0)
	in
	let rec print_pat v ps = match ps with
	  | []    -> fprintf (out()) "%s" !eol
	  | p::[] -> fprintf (out()) "%s%s"
	      (fill_int w1 (if p#contains v then 
			      if sign then
				if (p#lb v != neg_infinity) then 
				  if p#kind = 1 then 1 else -1
				else
				  if p#kind = 1 then -1 else 1
			      else 1
			    else 0))
		!eol
	  | p::ps -> fprintf (out()) "%s%s"
	      (fill_int w1 (if p#contains v then
			      if sign then
				if (p#lb v != neg_infinity) then 
				  if p#kind = 1 then 1 else -1
				else
				  if p#kind = 1 then -1 else 1
			      else 1
			    else 0))
		(!sep);
	      (print_pat v ps)
	in
	  if (names()) then begin
	    fprintf (out()) "%s" (String.make w0 !alignchar);
	    print_index w1 0 pat;
	  end;
	  for v=0 to vmax do
	    if names() then fprintf (out()) "%s%s" (fill_int w0 v) !sep;
	    print_pat v pat;
	  done;
	    
    (* =================================================================== *)
    | PATVAR sign ->
	let model = read_model (open_in_alt !modelfile !defaultfile)
	in let pat =
	    if !sort then
	      List.sort Pattern.compare_index (model#all_patterns())
	    else (model#all_patterns())
	in let var = model#get_variables()
	in let vmax = List.fold_left max (-1) var
	in let w0,w1 = if (!align)
	  then (width_int (List.length pat -1),
		width_int vmax)
	  else (0,0)
	in
	let rec print_pat k ps = 
	  match ps with
	  | []    -> ()
	  | p::ps ->
	      if (names()) then fprintf (out()) "%s%s" (fill_int w0 k) !sep;
	      for v=0 to vmax do
		fprintf (out()) "%s%s"
		  (fill_int w1 (if p#contains v then 
				  if sign then
				    if (p#lb v != neg_infinity) then 
				      if p#kind = 1 then 1 else -1
				    else
				      if p#kind = 1 then -1 else 1
				  else 1
				else 0))
		  (!sep)
	      done;
	      fprintf (out()) "%s" !eol;
	      print_pat (k+1) ps
	in
	  if (names()) then begin
	    fprintf (out()) "%s" (String.make w0 !alignchar);
	    for i=0 to vmax do
	      fprintf (out()) "%s%s" !sep (fill_int w1 i);
	    done;
	    fprintf (out()) "%s" !eol
	  end;
	  print_pat 0 pat;
	    
    (* =================================================================== *)
    | OBSPAT sign ->
	let model = read_model (open_in !modelfile)
	and inst  = read_instance (open_in_alt !instfile !defaultfile)
	in let pat =
	    if !sort then
	      List.sort Pattern.compare_index (model#all_patterns())
	    else (model#all_patterns())
	   and m = inst#m - 1
	in let w0,w1 = if (!align)
	  then (width_int m,
		max (if sign then 2 else 1) (width_int (List.length pat -1)))
	  else (0,0)
	in
	let rec print_pat o ps = match ps with
	  | []    -> fprintf (out()) "%s" !eol
	  | p::[] -> fprintf (out()) "%s%s"
	      (fill_int w1
		 (if (p#covers o) then
		    if sign && p#kind != o#kind then -1
		    else 1
		  else 0))
		!eol
	  | p::ps -> fprintf (out()) "%s%s"
	      (fill_int w1
		 (if (p#covers o) then
		    if sign && p#kind != o#kind then -1
		    else 1
		  else 0))
	      (!sep);
	      (print_pat o ps)
	in
	  if !sort then inst#sort();
	  if (names()) then begin
	    fprintf (out()) "%s" (String.make w0 !alignchar);
	    print_index w1 0 pat;
	  end;
	  for o=0 to m do
	    if (!forcenames) then
	      fprintf (out()) "%s%s" (inst#observation o)#name !sep;
	    if names() then fprintf (out()) "%s%s" (fill_int w0 o) !sep;
	    print_pat (inst#observation o) pat;
	  done;

    (* =================================================================== *)
    | PATOBS sign ->
	let model = read_model (open_in !modelfile)
	and inst  = read_instance (open_in_alt !instfile !defaultfile)
	in let pat =
	    if !sort then
	      List.sort Pattern.compare_index (model#all_patterns())
	    else (model#all_patterns())
	   and m = inst#m - 1
	in let w0,w1 = if (!align)
	  then (width_int (List.length pat -1),
		max (if sign then 2 else 1) (width_int m))
	  else (0,0)
	in
	let rec print_pat k ps = 
	  match ps with
	  | []    -> ()
	  | p::ps ->
	      if (names()) then fprintf (out()) "%s%s" (fill_int w0 k) !sep;
	      for o=0 to m do let o = inst#observation o in
		fprintf (out()) "%s%s"
		  (fill_int w1
		     (if p#covers o then
			if sign && p#kind != o#kind then -1
			else 1
		      else 0))
		  (!sep)
	      done;
	      fprintf (out()) "%s" !eol;
	      print_pat (k+1) ps
	in
	  if !sort then inst#sort();
	  if (names()) then begin
	    fprintf (out()) "%s" (String.make w0 !alignchar);
	    for i=0 to m do fprintf (out()) "%s%s" !sep (fill_int w1 i); done;
	    fprintf (out()) "%s" !eol
	  end;
	  print_pat 0 pat;
	    
    (* =================================================================== *)
    end;
    (* terminates program *)
    0
with
  | _ as ex -> terminal_catch ex
;;

