(* $Id: messages.ml 1 2005-07-15 19:39:24Z lemaire $ *)
(* ============================================================================
author: Pierre Lemaire
2004-10-26             : creation
============================================================================ *)

open Printf;;
open Lexing;;

type trace_level_type = TRACE_NONE | TRACE_SOME | TRACE_MOST | TRACE_ALL
                      | TRACE_DEBUG;;
                      
(* hidden types and variables for this module *)
type messages_type =
{ mutable fwarnings: out_channel;  (* channel where warnings are written *)
  mutable ferrors: out_channel;    (* channel where errors are written *)
  mutable ftrace: out_channel;    (* channel where trace is written *)
  mutable trace_level: trace_level_type; (* current trace level *)
  mutable foutput: out_channel;    (* channel where normal output is written *)
  mutable names: bool;             (* if names are printed *)
}

let messages =
{ fwarnings = stderr;
  ferrors   = stderr;
  ftrace    = stderr;
  trace_level = TRACE_NONE;
  foutput   = stdout;
  names     = true;
}

(* ========================================================================= *)
let names () = messages.names;;

let set_names b = messages.names <- b;;

(* ========================================================================= *)
let close_messages () =
  close_out_noerr messages.foutput;
  close_out_noerr messages.ftrace;
  close_out_noerr messages.fwarnings;
  close_out_noerr messages.ferrors
;;

let get_current_outputs () =
  (messages.foutput, messages.ftrace, messages.fwarnings, messages.ferrors,
   messages.names)
;;

let restore_outputs x =
  let (o,t,w,e,n) = x 
  in
    messages.foutput <- o;
    messages.ftrace <- t;
    messages.fwarnings <- w;
    messages.ferrors <- e;
    messages.names <- n;
;;

(* ========================================================================= *)
let warning msg =
  output_string messages.fwarnings "warning: ";
  output_string messages.fwarnings msg;
  output_string messages.fwarnings "\n";
  flush messages.fwarnings

(* ========================================================================= *)
let open_file_out ?(append=false) f alt =
  try
    if append
    then open_out_gen [Open_append; Open_creat; Open_text] 0o644 f
    else open_out f
  with
    Sys_error _ ->
      warning (sprintf "cannot open file %s (%s used instead.)" f
               (if (alt=stdout) then "stdout"
                else if (alt=stderr) then "stderr"
                else "-?-"));
      alt
;;

(* ========================================================================= *)
let open_file s = 
  try if (String.length s > 0) then open_in s else stdin
  with
    | Sys_error msg -> 
	warning (sprintf "can't open file (%s); stdin used instead" msg);
	stdin
;;

(* ========================================================================= *)
let open_in s = 
  let lex,s = 
    try if (String.length s > 0) then (open_in s,s) else (stdin, "<stdin>")
    with
      | Sys_error msg -> 
	  warning (sprintf "can't open file (%s); stdin used instead" msg);
	  (stdin, "<stdin>")
  in
  let lex = from_channel lex
  in
    lex.lex_curr_p <- {pos_fname = s; pos_lnum=1; pos_bol=0; pos_cnum=0;};
    lex;
;;

let open_in_alt s alt =
  if (String.compare s "" == 0) then open_in alt else open_in s
;;

(* ========================================================================= *)
let set_warning_file ?append s =
  messages.fwarnings <- open_file_out ?append s stderr;;

(* ========================================================================= *)
let error n msg =
  output_string messages.ferrors "ERROR: ";
  output_string messages.ferrors msg;
  output_string messages.ferrors "\n";
  flush messages.ferrors;
  if (n > 0) then exit n;
;;

let set_error_file ?append s =
  messages.ferrors <- open_file_out ?append s stderr;;


type io_error_type =
    Eof of Lexing.lexbuf
  | Parse of Lexing.lexbuf
  | Misc of string;;

exception IO_error of io_error_type;;

(* ========================================================================= *)
let print_io_error err = 
  warning 
    (match err with
     | Eof buf   -> let fname = (lexeme_start_p buf).pos_fname
                    in (sprintf "unexpected end of file %s"
                          (if String.length fname = 0 then ""
                           else "in "^fname))
     | Parse buf ->
	 let p = buf.lex_start_p 
	 and q = buf.lex_curr_p
         in (sprintf "parse error %s at %d:%d-%d"
               (if String.length p.pos_fname = 0 then ""
                else "in "^p.pos_fname)
               p.pos_lnum (p.pos_cnum-p.pos_bol+1) (q.pos_cnum-q.pos_bol+1))

     | Misc s   -> s 
    )
;;

(* ========================================================================= *)
let terminal_catch e = match e with
 (* | Failure msg     -> error  1 msg*)
  | Arg.Bad _ as ex -> raise ex
  | IO_error t      ->
      print_io_error t;
      1
  | Stack_overflow  ->
      error (-1) ("Sorry: out of memory :o( [stack overflow].");
      2
  | _ as ex         ->
      at_exit (fun () ->
                 error (-1)
                   ("(unexpected): this is probably a bug.\n"
                    ^"Please, report it to lemaire@rutcor.rutgers.edu, "
                    ^"with as many details as\npossible, so that it can"
                    ^" be fixed. The problem was: "));
      raise ex
;;

(* ========================================================================= *)
let set_trace_level t = messages.trace_level <- t;;

let set_trace_level_from_string s =
 messages.trace_level <-  match (s) with
  | "none" -> TRACE_NONE
  | "some" -> TRACE_SOME
  | "most" -> TRACE_MOST
  | "all"  -> TRACE_ALL
  | _      -> warning
                (sprintf "undefined trace level %s (none used instead)." s);
              TRACE_NONE
;;

let trace n msg =
  if (n <= messages.trace_level) then begin
    output_string messages.ftrace msg;
    flush messages.ftrace;
  end
;;

let set_trace_file ?append s =
  messages.ftrace <- open_file_out ?append s stderr;;

let hide_trace x =
  let old = messages.trace_level
  in
    messages.trace_level <- TRACE_NONE;
    try let y = x() in messages.trace_level <- old; y
    with Sys.Break -> begin  messages.trace_level <- old; raise Sys.Break end
;;

(* ========================================================================= *)
let output msg =
  output_string messages.foutput msg;
  flush messages.foutput

let set_output_file ?append s =
  messages.foutput <-  open_file_out ?append s stdout
;;

let out() = messages.foutput;;

