(* $Id: messages.mli 1 2005-07-15 19:39:24Z lemaire $ *)
(** 
 Functions to display messages (errors, warnings, trace, normal output).

 @version October 26, 2004
 @author Pierre Lemaire
*)

(* ============================================================================
author: Pierre Lemaire
2004-10-26             : creation
============================================================================ *)

(** turns on/off printing of names for observations. *)
val set_names: bool -> unit

(** returns whether names should be printed or not. *)
val names: unit -> bool

(** forces all outputs to be printed, and close all message channels. *)
val close_messages: unit -> unit

(** returns the current channels use for outputs. *)
val get_current_outputs: unit
  -> (out_channel * out_channel * out_channel * out_channel * bool)

(** restores the channels for outputs. *)
val restore_outputs:
  (out_channel * out_channel * out_channel * out_channel * bool)
  -> unit

(* ========================================================================= *)
(** {6 warnings} *)

(** [warning msg] prints the warning message [msg]. *)
val warning: string -> unit

(** [set_warning_file s] specifies that all warnings will be printed
    in the file [s]. If this file cannot be openned in write mode, then
    the standard error channel is used. *)
val set_warning_file: ?append:bool -> string -> unit

(* ========================================================================= *)
(** {6 errors and error handling} *)

(** [error n msg] prints the error message [msg] and then exits with value [n],
    if this one is greater then 0. *)
val error: int -> string -> unit

(** [set_error_file s] specifies that all error messages will be printed
    in the file [s]. If this file cannot be openned in write mode, then
    the standard error channel is used. *)
val set_error_file: ?append:bool -> string -> unit

(** error type for i/o error. *)
type io_error_type = Eof of Lexing.lexbuf | Parse of Lexing.lexbuf
		     | Misc of string;;

exception IO_error of io_error_type

(** prints the i/o error as a warning. *)
val print_io_error: io_error_type -> unit

(** generic catch to end a program. *)
val terminal_catch: exn -> int

(* ========================================================================= *)
(** {6 trace} *)

(** Type for the different level of trace. *)
type trace_level_type = TRACE_NONE | TRACE_SOME | TRACE_MOST | TRACE_ALL
                      | TRACE_DEBUG;;

(** [set_trace_level t] sets the current level of trace to [t]. *)
val set_trace_level: trace_level_type -> unit

(** [set_trace_level_from_string s] sets the current level of trace to the
    level correspond to [s], if it is in [none, some, most, all]. If the
    string is not recognized, then TRACE_NONE is used. *)
val set_trace_level_from_string: string -> unit

(** [trace n msg] prints the trace message [msg] if the level [n] is 
    less than the current level of trace. *)
val trace: trace_level_type -> string -> unit

(** [set_trace_file s] specifies that all trace messages will be printed
    in the file [s]. If this file cannot be openned in write mode, then
    the standard error channel is used. *)
val set_trace_file: ?append:bool -> string -> unit

(** [hide_trace f] executes [f] hidding its potential trace. *)
val hide_trace: (unit -> 'a) -> 'a

(* ========================================================================= *)
(** {6 normal output} *)

(** [output msg] prints the output message [msg]. *)
val output: string -> unit

(** [set_output_file b s] specifies that all normal messages will be
    printed in the file [s]. If this file cannot be opened in write
    mode, then the standard output channel is used. The option
    parameter [b] specify whether the file is opened in append mode or
    not. *)
val set_output_file: ?append:bool -> string -> unit

(** the channel for normal output. *)
val out: unit -> out_channel

(* ======================================================================= *)
(** {6 i/o file opening} *)

(** [open_file_out f alt] opens file of name [f] in writing mode; uses
    alt if [f] cannot be opened. *)
val open_file_out: ?append:bool -> string -> out_channel -> out_channel

(** [open_in s] opens file of name [s]; uses stdin if opening of [s]
    fails. *)
val open_file: string -> in_channel

(** [open_in s] opens file of name [s] as a Lexing.lexbuf; uses stdin if
    opening of [s] fails. *)
val open_in: string -> Lexing.lexbuf

(** [open_in_alt a b] opens file of name [a] if [a] is not empty, of
    name [b] otherwise ; uses stdin if opening fails. *)
val open_in_alt: string -> string -> Lexing.lexbuf

