(* $Id: model.ml 11 2007-02-28 09:38:48Z pierre $ *)
(* ============================================================================
author: Pierre Lemaire
2004-11-03             : creation
2004-11-03..2004-11-08 : basics
============================================================================ *)

open Messages;;
open Observation;;
open Instance;;
open Cutpointset;;
open Pattern;;
open Pandect;;

(* =========================================================================
   class model
   ========================================================================= *)
class model
   _kmax
   _pats_init = 
object(self)
   inherit pandect _kmax _pats_init
  (* =======================================================================
     fields
     ======================================================================= *)
   
  val mutable right   = Array.make 0 0
  val mutable wrong   = Array.make 0 0
  val mutable unknown = Array.make 0 0
  val mutable sensibility = Array.make 0 0.0
   
  (* =======================================================================
     counters
     ======================================================================= *)
  
  (* ======================================================================= *)
  method count_right_of_kind k =
    try right.(k)
    with | Invalid_argument "index out of bounds" -> 0
  
  (* ======================================================================= *)
  method count_right = Array.fold_left (+) 0 right
  
  (* ======================================================================= *)
  method count_undef_of_kind k =
    try unknown.(k)
    with | Invalid_argument "index out of bounds" -> 0
    
  (* ======================================================================= *)
  method count_undef = Array.fold_left (+) 0 unknown

  (* ======================================================================= *)
  method count_wrong_of_kind k =
    try wrong.(k)
    with | Invalid_argument "index out of bounds" -> 0
    
  (* ======================================================================= *)
  method count_wrong = Array.fold_left (+) 0 wrong
    
  (* ======================================================================= *)
  method accuracy =
    let r = float_of_int self#count_right
    and w = float_of_int self#count_wrong
    and u = float_of_int self#count_undef
    in (r +. u/.2.0)/.(r+.w+.u)

  (* ======================================================================= *)
  method datascope_accuracy =
    let rpos = float_of_int (self#count_right_of_kind 1)
    and rneg = float_of_int (self#count_right_of_kind 0)
    and wpos = float_of_int (self#count_wrong_of_kind 1)
    and wneg = float_of_int (self#count_wrong_of_kind 0)
    and upos = float_of_int (self#count_undef_of_kind 1)
    and uneg = float_of_int (self#count_undef_of_kind 0)
    in 
    let pos = rpos +. wpos +. upos
    and neg = rneg +. wneg +. uneg
    in (rpos/.pos +. rneg/.neg +. 1.0-.wpos/.pos +. 1.0-.wneg/.neg) /. 4.0

  (* ======================================================================= *)
  method sensibility k =
    try sensibility.(k)
    with | Invalid_argument "index out of bounds" -> nan
  
  (* ======================================================================= *)
  method sensitivity =
    try sensibility.(1)
    with | Invalid_argument "index out of bounds" -> nan

  (* ======================================================================= *)
  method specificity =
    try sensibility.(0)
    with | Invalid_argument "index out of bounds" -> nan

  (* =======================================================================
     methods
     ======================================================================= *)
  
  (* ======================================================================= *)
  method compute_for (inst: instance) =
    (* re-init *)
    right   <- Array.make inst#k 0;
    wrong   <- Array.make inst#k 0;
    unknown <- Array.make inst#k 0;
    (* right/wrong *)
    for o=0 to inst#m-1 do
      let c = self#classification (inst#observation o)
      and k = (inst#observation o)#kind 
      in if (k >= 0) then begin
        if c < 0 then unknown.(k) <- unknown.(k) + 1
        else if c = k then right.(k) <- right.(k) + 1
        else               wrong.(k) <- wrong.(k) + 1
      end
    done;
    (* sensibilities *)
    sensibility <- Array.make inst#k 0.0;
    for k=0 to inst#k-1 do
      sensibility.(k) <- (float_of_int right.(k))
                           /. (float_of_int (inst#count_of_kind k))
    done

  (* ======================================================================= *)
  method total_coverage o = Array.fold_left (+) 0 (self#coverage o)
  
  (* ======================================================================= *)
  method coverage o =
    let c = Array.make (Array.length pats) 0
    in let rec covrec o ps = 
        match ps with
        | []    -> 0
        | p::ps -> (if p#covers o then 1 else 0) + (covrec o ps)
    in for k=0 to (Array.length pats)-1 do
         c.(k) <- (covrec o pats.(k))
      done;
      c

  (* ======================================================================= *)
  method coverage_with_epsilon e o =
    let c = Array.make (Array.length pats) 0
    in let rec covrec o ps = 
        match ps with
        | []    -> 0
        | p::ps -> (if p#covers_with_epsilon e o then 1 else 0) + (covrec o ps)
    in for k=0 to (Array.length pats)-1 do
         c.(k) <- (covrec o pats.(k))
      done;
      c

  (* ======================================================================= *)
  method classification o =
    let c = self#coverage o
    and kbest = ref (-1)
    and rbest = ref 0.0
    and tie   = ref false
    in for k=0 to (Array.length pats) -1 do
      let r = (float_of_int c.(k)) /. (float_of_int (self#count_of_kind k))
      in
        if (abs_float(r -. !rbest)) < epsilon_float then tie := true
        else if (r > !rbest) then begin
          kbest := k; rbest := r; tie := false
        end
      done;
      if !tie then -1 else !kbest
      
  (* ======================================================================= *)
  method classification_with_epsilon e o =
    let c = self#coverage_with_epsilon e o
    and kbest = ref (-1)
    and rbest = ref 0.0
    and tie   = ref false
    in for k=0 to (Array.length pats) -1 do
      let r = (float_of_int c.(k)) /. (float_of_int (self#count_of_kind k))
      in
        if (abs_float(r -. !rbest)) < epsilon_float then tie := true
        else if (r > !rbest) then begin
          kbest := k; rbest := r; tie := false
        end
      done;
      if !tie then -1 else !kbest
      
end  

(* ======================================================================= *)
let greedy_old c (inst:instance) (pand:pandect) =
  let cover = Array.make (inst#m) 0
  in let rec best ps = 
    match ps with 
    | []    -> (empty_pattern, -1, [])
    | p::ps ->
      let (bp,bv,wps) = best ps
      and vp = ref 0
      in for i=0 to inst#m-1 do
        if   cover.(i) < c
          && p#kind = (inst#observation i)#kind
	  && p#covers (inst#observation i)
        then vp := !vp + 1;
        done;
      if !vp > bv then (p,!vp, bp::wps) else (bp,bv, p::wps)
  in let selection = ref []
  and pats  = ref (pand#all_patterns())
  and ok = ref false
  in while not !ok do
      let bp,bv,wps = best !pats
      in if bv <= 0 then ok := true
         else begin
           selection := bp::(!selection);
           for i=0 to inst#m-1 do
             if bp#kind = (inst#observation i)#kind
	       && bp#covers (inst#observation i)
             then cover.(i) <- cover.(i) + 1;
           done;
           pats := wps
         end
    done;
    new model (pand#k) !selection
;;

(* ======================================================================= *)
let greedy comp cov (inst:instance) (pand:pandect) =
  let cover = Array.make (inst#m) 0
  in 
  let rec best ps = 
    match ps with 
      | []    -> (empty_pattern, -1, [])
      | p::ps ->
	  let (bp,bv,wps) = best ps
	  and vp = ref 0
	  in for i=0 to inst#m-1 do
	      let k = (inst#observation i)#kind
	      in
		try
		  if   cover.(i) < cov.(k)
		    && p#kind = k
		    && p#covers (inst#observation i)
		  then vp := !vp + 1;
		with _ -> ()
            done;
	    if !vp > bv then (p,!vp, bp::wps) else (bp,bv, p::wps)
  in let selection = ref []
     and pats  = ref (List.sort comp (pand#all_patterns()))
     and ok = ref false
  in while not !ok do
      let bp,bv,wps = best !pats
      in if bv <= 0 then ok := true
        else begin
          selection := bp::(!selection);
          for i=0 to inst#m-1 do
            if bp#kind = (inst#observation i)#kind
	      && bp#covers (inst#observation i)
            then cover.(i) <- cover.(i) + 1;
          done;
          pats := wps
        end
    done;
    new model (pand#k) !selection
;;

(* ======================================================================= *)
let select comp cov (inst:instance) (pand:pandect) =
  let cover = Array.make (inst#m) 0
  and selection = ref []
  and pats = List.sort comp (pand#all_patterns())
  in 
    List.iter 
      (fun p -> 
	 let keep = ref false
	 in for i=0 to inst#m-1 do
	     let k = (inst#observation i)#kind
	     in
		try
		  if   cover.(i) < cov.(k)
		    && p#kind = k
		    && p#covers (inst#observation i)
		  then begin
		    keep := true;
		    cover.(i) <- cover.(i) + 1;
		  end
		with _ -> ()
            done;
	   if !keep then selection := p::!selection
      )
      pats;
    new model (pand#k) !selection
;;

(* ======================================================================= *)
let boxes (_: cutpointset list) (_:instance) =
  Printf.fprintf stderr "boxes; NOT IMPLEMENTED\n";
  new model 1 []
;;
