(* $Id: model.mli 1 2005-07-15 19:39:24Z lemaire $ *)
(** 
 The class for a model.

 @version November 16, 2004
 @author Pierre Lemaire
*)

(* ============================================================================
author: Pierre Lemaire
2004-11-03             : creation
2004-11-03..2004-11-08 : basics
============================================================================ *)

open Observation;;
open Instance;;
open Cutpointset;;
open Pattern;;
open Pandect;;

(* ======================================================================= *)
(** {6 the class [model]}*)

(** The class [model] defines a model for LAD.
    
    A model is a collection of [patterns], each of them belonging to a
    given kind/class. Hence, it is a particular [pandect], from which
    it inherits. However, the [patterns] with undefined kind/class are
    ignored in a [model].
  *)
class model:
  int ->
  pattern list ->
object
  inherit Pandect.pandect
  (* ===================================================================== *)
  (** {6 fields} *)
  
  (** number of well-classified [observations] of each kind/class. *) 
  val mutable right: int array
  
  (** number of misclassified [observations] of each kind/class. *) 
  val mutable wrong: int array
  
  (** number of unclassified [observations] of each kind/class. *) 
  val mutable unknown: int array
  
  (** sensibility for eachh kind/class. *)
  val mutable sensibility: float array

  (* ===================================================================== *)
  (** {6 properties} *)
  
  (** returns the number of well-classified [observations] of the given
      kind/class, for this [model]. *)
  method count_right_of_kind: int -> int
  
  (** returns the number of well-classified [observations] for this [model]. *)
  method count_right: int

  (** returns the number of unclassified [observations] of the given
      kind/class, for this [model]. *)
  method count_undef_of_kind: int -> int
  
  (** returns the number of unclassified [observations] for this [model]. *)
  method count_undef: int

  (** returns the number of misclassified [observations] of the given
      kind/class, for this [model]. *)
  method count_wrong_of_kind: int -> int
  
  (** returns the number of misclassified [observations] for this [model]. *)
  method count_wrong: int
    
  (** returns the accuracy of this [model]. *)
  method accuracy: float

  method datascope_accuracy: float
  
  (** returns the sensitivity of this [model].
   
      The sensitivity is defined as A/B, where A is the number of positive
      observations (kind = 1) classified as positive; and B is the number of
      positive observations. This notion, as well as the specificity below,
      makes sense in binary cases; otherwise one should consider the
      sensibilities. *)
  method sensitivity: float
   
  (** returns the specificity of this [model].
   
      The specificity is defined as A/B, where A is the number of negative
      observations (kind = 0) classified as posinegative; and B is the number of
      negative observations. *)
  method specificity: float
    
  (** returns the sensibility of this [model] for the given kind/class. This notion
      generalises the notions of sensitivity and specificity.
  
      The sensibility for a kind k is defined as A/B, where A is the number
      of observations of kind k classified as kind k; and B is the number of
      observations of kind k.*)
   method sensibility: int -> float
  
  (* ===================================================================== *)
  (** {6 methods} *)
 
  (** computes the properties of [model] for the given [instance]. Unclassified
      observations are ignored. This method deletes any former values. *)
  method compute_for : instance -> unit
  
  (** returns the total coverage of an [observation] in this [model], that is
      the number of [patterns] that covers it, regardless of the kind/class.*)
  method total_coverage: observation -> int
  
  (** returns the coverage, for each kind/class of [patterns], of an [observation]
      by this [model] *)
  method coverage: observation -> int array
  method coverage_with_epsilon: float -> observation -> int array

  (** returns the classification of the [observation] by this [model]. A negative
      value is returned if the [observation] is not classified.
      
      In this [model] class, the classification is given by highest {e ratio}
      of covering patterns for each kind. *)
  method classification: observation -> int

  method classification_with_epsilon: float -> observation -> int

end  

(** deprecated... use greedy instead. *)
val greedy_old: int -> instance -> pandect -> model

(** Greedily selects patterns. The best pattern, at any iteration, is
    the one that covers the biggest number of unsuffisantly covered
    observations of its kind. In case of ties, the first found pattern is 
    kept (i.e smallest according to comp). *)
val greedy: (pattern -> pattern -> int) -> int array
  -> instance -> pandect -> model

(** Selects patterns. The patterns are sorted according to the
    comparison function. Then, each one is kept if it covers at least
    one unsuffisantly covered observations of its kind. *)
val select: (pattern -> pattern -> int) -> int array
  -> instance -> pandect -> model

val boxes: cutpointset list -> instance -> model
