(* $Id: observation.mli 1 2005-07-15 19:39:24Z lemaire $ *)
(** 
 The class for an observation.

 @version November 08, 2004
 @author Pierre Lemaire
*)

(* ============================================================================
author: Pierre Lemaire
2004-10-26             : creation
2004-10-26..2004-11-08 : basics 
============================================================================ *)

open Indexed;;

(* ======================================================================= *)
(** {6 the class [observation]}*)

(** The class [observation] defines an observation for LAD.

    [observation i k vs] creates a new [observation] of kind/class [k] with
    the values [vs] for the variables; [i] is used as index.
    If the kind/class is not known, it should be set to a negative value (-1);
    each unknown value, for the variables, should be set to [nan].
  *)
class observation:
  ?_name:string ->   (* name *)
  int ->             (* the index *)
  int ->             (* the kind/class of the observation *)
  float array ->     (* the variables *)
object ('a)
  inherit Indexed.indexed

  val mutable name: string
  
  (* ===================================================================== *)
  (** {6 properties} *)

  (** returns the name of this [observation]. *)
  method name: string

  (** sets the name of this [observation]. *)
  method set_name: string -> unit
    
  (** returns the class/kind of this [observation]. *)
  method kind: int

  (** returns the number of variables of this [observation]. *)
  method n: int

  (** [value i] returns the value of the [i]th variable of this [observation].
      
      Raises [Invalid_argument "Observation.value: no such variable"] if the
      [observation] does not have a variable [i].
      *)
  method value: int -> float
  
  method set_value: int -> float -> unit
  
  method clone: unit -> observation
  
  method distance: observation -> float

end

(** an empty observation. *)
val empty_observation: observation

(* ======================================================================= *)
(** {6 comparisons}*)

(** compares 2 [observations] according to their index. *)
val compare_index: observation -> observation -> int

(** compares 2 [observations] according to their index. *)
val compare_name: observation -> observation -> int

(** compares 2 [observations] according first to their kind/class, and
    then, in case of equality, to their first variable, and so on.
    
    Transmits [Invalid_argument "Observation.value: no such variable"] if
    the first [observation] has more variables than the second, and the
    equality holds for all common ones. *)
val compare_all: observation -> observation -> int

(** compares 2 [observations] according to their kind/class. *)
val compare_kind: observation -> observation -> int

(** compares 2 [observations] according to the values of the given
    variable (a defined value is greater than an undefined one).
    
    Transmits [Invalid_argument "Observation.value: no such variable"] if
    one of the [observation] does not have a variable [i]. *)
val compare_variable: int -> observation -> observation -> int

(** sets the current default comparison function. *)
val set_compare_function: (observation -> observation -> int) -> unit

(** compares 2 [observations] according to the current default
    comparison function. *)
val compare: observation -> observation -> int
