(* $Id: pandect.ml 13 2007-05-30 14:34:34Z pierre $ *)
(* ============================================================================
author: Pierre Lemaire
2004-11-02             : creation
============================================================================ *)

open Messages;;
open Instance;;
open Cutpointset;;
open Monome;;
open Pattern;;

(* =========================================================================
   class pandect
   ========================================================================= *)
class pandect
   _kmax
   _pats_init = 
object(self)
  (* =======================================================================
     fields
     ======================================================================= *)
  val mutable pats = (Array.make (_kmax+1) []: pattern list array)
  val mutable undef= ([]: pattern list)

  initializer
    let rec parse ps =
      match ps with
      | [] -> ()
      | p::ps -> let k = p#kind
                 in if (k<0) || (k>_kmax) then undef <- p::undef
                    else pats.(k) <- p::pats.(k);
                    parse ps
    in
      parse _pats_init
    
  (* =======================================================================
     counters
     ======================================================================= *)
  
  method is_empty () =
    Array.fold_left
      (fun x y -> (match y with [] -> true | _ -> false) && x)
      true 
      pats

  method k = _kmax+1
     
  method count_of_kind k =
    try List.length pats.(k)
    with | Invalid_argument "index out of bounds" -> 0

  method count_of_undef () = List.length undef

  (* =======================================================================
     methods
     ======================================================================= *)
 
  (* ======================================================================= *)
  method all_patterns () = List.concat (Array.to_list pats)
     
  method patterns_of_kind k =
    try pats.(k)
    with | Invalid_argument "index out of bounds" -> []

  method patterns_undef () = undef

  (* ======================================================================= *)
  method get_variables () =
    let rec add_var ys x =
      match ys with
	| [] -> [x]
	| y::ys as yys ->
	    if (y>x) then x::yys
	    else if (y<x) then y::(add_var ys x)
	    else yys
    in
    let tmp = ref []
    in let pat = Array.of_list (self#all_patterns())
    in for k=0 to Array.length pat - 1 do
        let ms = Array.of_list (pat.(k)#monomes())
        in for m=0 to Array.length ms - 1 do
	    tmp := add_var !tmp ms.(m)#variable
	  done;
      done;
      !tmp
    

  (* ======================================================================= *)
  method get_monomes () =
    let tmp = ref []
    in let pat = Array.of_list (self#all_patterns())
    in for k=0 to Array.length pat - 1 do
        let ms = Array.of_list (pat.(k)#monomes())
        in for m=0 to Array.length ms - 1 do
	    match ms.(m)#inequality with
	      | GT _  -> tmp := add_gt_monome !tmp ms.(m)
	      | LT x  -> tmp := add_gt_monome !tmp
		  (new monome ms.(m)#variable (GT x))
	      | BOX (x,y) ->
		  tmp := add_gt_monome
		    (add_gt_monome !tmp (new monome ms.(m)#variable (GT x)))
		    (new monome ms.(m)#variable (GT y))
	  done;
      done;
      !tmp
  
  (* ======================================================================= *)
  method filter f =
    let rec listfilter pl =
      match pl with
      | [] -> []
      | p::pl -> if (f p) then p::(listfilter pl) else listfilter pl
    in
    let arrayfilter a =
      let tmp = Array.make _kmax []
      in for k=0 to _kmax do tmp.(k) <- listfilter a.(k) done;
       tmp  
    in
    {< pats = arrayfilter pats;
       undef = listfilter undef
    >}

  (* ======================================================================= *)
  method clean_coverage (inst: instance) = 
    let rec cleanrec2 q ps =
      match ps with
      | [] -> true, []
      | p::ps ->
        if (p#kind != q#kind) then true, p::ps
        else  
          let pq_pos,pq_neg = ref false, ref false
          and qp_pos,qp_neg = ref false, ref false
          in for o=0 to inst#m - 1 do
            let o = inst#observation o
            in if p#covers o && not (q#covers o) then
                 if o#kind = 0 then pq_neg := true else pq_pos := true;
               if q#covers o && not (p#covers o) then
                 if o#kind = 0 then qp_neg := true else qp_pos := true;
        done;
        if p#kind = 0 then begin
          if !pq_neg && !qp_neg then (* keep both and goes on *)
            let (a,b) = cleanrec2 q ps in a, p::b
          else if !pq_neg && not (!pq_pos) then (* keep p, not q, and hence stops *)
            false, p::ps
          else if !qp_neg && not (!qp_pos) then  (* keep q, not p, and goes on *)
            cleanrec2 q ps
          else (* keep ARBITRARY the most homogenous one *)
            if (p#homogeneity >= q#homogeneity) then false, p::ps
            else cleanrec2 q ps
        end else begin
          if !pq_pos && !qp_pos then (* keep both and goes on *)
            let (a,b) = cleanrec2 q ps in a, p::b
          else if !pq_pos && not (!pq_neg) then (* keep p, not q, and hence stops *)
            false, p::ps
          else if !qp_pos && not (!qp_neg) then  (* keep q, not p, and goes on *)
            cleanrec2 q ps
          else (* keep ARBITRARY the most homogenous one *)
            if (p#homogeneity >= q#homogeneity) then false, p::ps
            else cleanrec2 q ps
        end
    in
    let rec cleanrec ps =
      match ps with
      | []    -> []
      | p::ps -> let ok,ps = cleanrec2 p ps
                 in if ok then p::(cleanrec ps) else cleanrec ps;
    in
      for i=0 to _kmax do pats.(i) <- cleanrec pats.(i) done;

  (* ======================================================================= *)
  method clean_inclusion () = 
    let rec cleanrec2 q ps =
      match ps with
      | [] -> true, []
      | p::ps as all ->
        if      p#includes q then false, all
        else if q#includes p then cleanrec2 q ps
        else let ok,ps = cleanrec2 q ps in ok, p::ps
    in
    let rec cleanrec ps =
      match ps with
      | []    -> []
      | p::ps -> let ok,ps = cleanrec2 p ps
                 in if ok then p::(cleanrec ps) else cleanrec ps;
    in
      for i=0 to _kmax do pats.(i) <- cleanrec pats.(i) done;
        
  (* ======================================================================= *)
  method cutpointsets () =
    let cpsl = ref []
    in
    let rec addc cs =
      match cs with
      | [] -> ()
      | c::cs -> cpsl := Cutpointset.add c !cpsl; addc cs
    in
    let rec addp ps =
      match ps with
      | []    -> ()
      | p::ps -> addc (Cutpointset.of_monomes (p#monomes()));
                 addp ps
    in
     for k=0 to _kmax do addp pats.(k) done;
     !cpsl
    
  (* =======================================================================
     sorting
     ======================================================================= *)
 
  method sort () =
    for k=0 to _kmax do
      pats.(k) <- List.stable_sort (Pattern.compare) pats.(k)
    done;
    undef <- List.stable_sort (Pattern.compare) undef

end;;

(* =========================================================================
   construction procedures
   ========================================================================= *)

(* ========================================================================= *)
let all_with_action
    ?(pseudoprimes=false)
    (action: pattern -> unit)
    (deg:int) (hom:float array) (pre:float array) (hr: float array)
    (cpsl:cutpointset list) (inst:instance) =
  (* each sub-function returns the list of patterns to extend and
     adds the good ones to 'all'. *)
  (* add a monome to each pattern of a list *)       
  let rec add_monome pl m =
    trace TRACE_DEBUG (Printf.sprintf "...add_monome\n");
    match pl with
      | [] -> []
      | p::pl ->
	  let p = p#clone [m]
	  in
            p#compute_for inst;
	    if p#degree = deg then begin
              (* cannot be extended: keep or forget *)
              if p#kind >= 0 then begin
		let k=p#kind
		in
		  if  p#homogeneity    >= hom.(k)
                    && p#max_prevalence >= pre.(k)
                    && p#hazardratio    >= hr.(k)
		  then action p;
              end;
              add_monome pl m
	    end else begin (* can be extended *)
              if (p#prevalence 0) < pre.(0) && (p#prevalence 1) < pre.(1)
              then add_monome pl m (* don't extend it *)
              else begin (* should be extended *)
		let k=p#kind
		in
		  if (k >= 0) then begin 
		    if  p#homogeneity     >= hom.(k)
                      && p#max_prevalence >= pre.(k)
                      && p#hazardratio    >= hr.(k)
		    then begin 
		      action p;
		      if not pseudoprimes then p::(add_monome pl m)
		      else (add_monome pl m)
		    end
		    else p::(add_monome pl m)
		  end else p::(add_monome pl m)
              end
	    end
  in
    (* add each monome of a list to each pattern of a list *)
  let rec add_monome_list pl ml = 
    match ml with
      | []    -> []
      | m::ml ->
	  trace TRACE_DEBUG (Printf.sprintf "...add_monome_list [%d]\n" (List.length ml));
	  (* each m must be independant: do not use b1 for add__list... *)
	  let b1 = add_monome pl m
	  and b2 = add_monome_list pl ml
	  in List.rev_append b1 b2
  in
    (* *)
  let rec add_cutpointset_list pl cpsl =
    match cpsl with
      | []        -> []
      | cps::cpsl ->
	  trace TRACE_MOST (Printf.sprintf "...add_cutpointset_list [%d]\n" (List.length cpsl));
	  (* the cpsl are not independant: USE cps for add_list... *)
	  let b1 = add_monome_list pl (cps#to_simple_monomes())
	  in 
            add_cutpointset_list (List.rev_append pl b1) cpsl
  in
  let p = empty_pattern
  in
    p#compute_for inst;
    if p#kind >= 0 then
      if p#max_prevalence > pre.(p#kind)
        && p#homogeneity > hom.(p#kind) && p#hazardratio > hr.(p#kind)
      then action p;
    ignore (add_cutpointset_list [p] cpsl);
;;

(* ========================================================================= *)
let all ?(pseudoprimes=false) deg hom pre hr cpsm inst =
  let pat  = ref []
  and kmax = ref 0
  in let f p = pat := p::(!pat); if (p#kind > !kmax) then kmax := p#kind
  in all_with_action ~pseudoprimes:pseudoprimes f deg hom pre hr cpsm inst;
    new pandect !kmax !pat
;;

(* ========================================================================= *)
let rec intersection ps qs =
  let rec inter ps q = 
    match ps with
    | [] -> []
    | p::ps -> let x = p#intersection q
               in
                 if    x#degree > 0
                      && (compare_equal x p) <> 0
                      && (compare_equal x q) <> 0
                   then x::(inter ps q) else (inter ps q)
  in match qs with 
  | [] -> []
  | q::qs -> (inter ps q)@(intersection ps qs)

