(* $Id: pandect.mli 1 2005-07-15 19:39:24Z lemaire $ *)
(** 
 The class for a pandect (a collection of patterns).

 @version November 16, 2004
 @author Pierre Lemaire
*)

(* ============================================================================
author: Pierre Lemaire
2004-11-02             : creation
2004-11-02..2004-11-08 : basics
============================================================================ *)

open Instance;;
open Cutpointset;;
open Monome;;
open Pattern;;

(* ======================================================================= *)
(** {6 the class [pandect]}*)

(** The class [pandect] defines a collection of patterns for LAD.
    
    Formally, a pandect should be the entire collection of patterns (that
    match some criteria), but it can be use just as a set of patterns.
    
    [pandect k pats] creates a new pandect using [pats] as patterns. [k] is 
    required for efficiency reasons and patterns of [pats] should have a
    kind/class in \[0..[k]\] (other kinds/classes are considered 'undefined').
    The function [Pattern.count_kinds] may be used
    to find a suitable value, if unknown.
  *)
class pandect:
  int ->
  pattern list ->
object('a)
  (* ===================================================================== *)
  (** {6 fields} *)
  
  (** the collection of [patterns] of each kind/class. *)
  val mutable pats: (pattern list) array
  
  (** the collection of [patterns] of undefined kind/class. *)
  val mutable undef: pattern list

  (* ===================================================================== *)
  (** {6 counters} *)
  
  (** returns whether this [pandect] is empty or not. *)
  method is_empty: unit -> bool

  (** returns the number of kinds/classes in this [pandect]. *)
  method k: int
  
  (** returns the number of [patterns] of the given kind/class. *)
  method count_of_kind: int -> int

  (** returns the number of [patterns] of undefined kind/class. *)
  method count_of_undef: unit -> int

  (* ===================================================================== *)
  (** {6 methods} *)
 
  method all_patterns: unit -> pattern list
  
  (** returns the list of [patterns] of the given kind/class. *)
  method patterns_of_kind: int -> pattern list

  (** returns the list of undefined [patterns]. *)
  method patterns_undef: unit -> pattern list


  (** returns the list of indexes of the variables used. *)
  method get_variables: unit -> int list

  (** returns the list of binary variables (form x>...) used. *)
  method get_monomes: unit -> monome list

  (** returns a fresh pandect by filtering the [patterns] according to the
      given filter. *)
  method filter: (pattern -> bool) -> 'a
  
  method clean_coverage: instance -> unit
  method clean_inclusion: unit -> unit
  
  method cutpointsets: unit -> cutpointset list
  
  (* ===================================================================== *)
  (** {6 sorting} *)
 
  (** sorts the lists of [patterns]. *)
  method sort: unit -> unit

end  

(* ======================================================================= *)
(** {6 construction procedures}*)
  
(** [all d h p hr cp inst] builds the complete sets of [patterns] based on the
    cut-points of [cp], with a degree less or equal to [d], an homogeneity
    at least [h], prevalence at least [p] and an hazard ratio at least [hr] on
    [inst] (patterns with an undefined class/kind are not returned). *)
val all: ?pseudoprimes:bool -> int -> float array -> float array -> float array -> cutpointset list -> instance -> pandect

(** same as [all], but perform the given action on the produced [patterns],
    instead of returning them in a [pandect]. *)
val all_with_action: ?pseudoprimes:bool -> (pattern -> unit) -> int -> float array -> float array -> float array -> cutpointset list -> instance -> unit

val intersection: pattern list -> pattern list -> pattern list
