(* $Id: pattern.ml 14 2007-06-29 08:14:54Z pierre $ *)
(* ============================================================================
author: Pierre Lemaire
2004-11-01             : creation
============================================================================ *)

open Messages;;
open Indexed;;
open Observation;;
open Instance;;
open Monome;;

(* max acceptable ratios of covered 'undefined' (missing) pos/neg obser.
  among the observations covered by a pattern. *)
let missing_threshold = [|0.0; 0.0|];;

(* whether missing values are considered uncovered by every pattern or only
   the ones of the same kind. *)
let missing_for_all = ref false;; (* ??? !!! ??? *)

(* =========================================================================
   class pattern
   ========================================================================= *)
class pattern
  (_index: int)
  (_init_cuts: monome list) = 
object (self)
  inherit indexed _index
    
  (* =======================================================================
     fields
     ======================================================================= *)
    
  val mutable cuts         = _init_cuts;
  val mutable coverage     = (-1);
  val mutable kind         = (-1)
  val mutable homogeneity  = nan
  val mutable prevalence   = Array.make 0 nan
  val mutable hazardratio  = nan
    
  (* =======================================================================
     properties
     ======================================================================= *)
    
  (* ======================================================================= *)
  method degree = 
    let rec add x cs = 
      match cs with
	| [] -> [x]
	| c::cs -> if x=c then c::cs else c::(add x cs)
    in let rec parse cs =
	match cs with
	  | []    -> []
	  | c::cs -> add (c#variable) (parse cs)
    in List.length (parse cuts)
	 
  (* ======================================================================= *)
  method length = List.length cuts

  (* ======================================================================= *)
  method monomes () = cuts

  (* ======================================================================= *)
  method ub i = 
    let rec parse ms =
      match ms with
	| [] -> infinity
	| m::ms -> if m#variable = i then min m#ub (parse ms)
          else (parse ms)
    in parse cuts

  (* ======================================================================= *)
  method lb i = 
    let rec parse ms =
      match ms with
	| [] -> neg_infinity
	| m::ms -> if m#variable = i then max m#lb (parse ms)
          else (parse ms)
    in parse cuts
	  
  (* ======================================================================= *)
  method contains i = 
    let rec parse ms =
      match ms with
	| [] -> false
	| m::ms -> if m#variable = i then true else parse ms 
    in parse cuts
	  
  (* =======================================================================
     instance-dependent properties
     ======================================================================= *)
	  
  (* ======================================================================= *)
  method kind = kind

  method set_kind k = kind <- k
    
  (* ======================================================================= *)
  method coverage = coverage

  (* ======================================================================= *)
  method homogeneity = homogeneity
    
  method set_homogeneity h = homogeneity <- h
    
  (* ======================================================================= *)
  method prevalence i =
    try prevalence.(i)    
    with | Invalid_argument "index out of bounds" -> nan
      
  (* ======================================================================= *)
  method max_prevalence =
    if Array.length prevalence = 0 then nan
    else if (kind>=0) then prevalence.(kind)
    else let pmax = ref neg_infinity
    in for i=0 to Array.length prevalence-1 do
      if prevalence.(i) > !pmax then pmax := prevalence.(i)
    done;
      !pmax   (* use Array.fold !!! *)
	
  method set_max_prevalence p = 
    try prevalence.(kind) <- p
    with | Invalid_argument "index out of bounds"
      -> prevalence <- Array.make (kind+1) nan;
	prevalence.(kind) <- p

  (* ======================================================================= *)
  method hazardratio = hazardratio
    
  method set_hazardratio x = hazardratio <- x
    
  (* =======================================================================
     methods
     ======================================================================= *)
    
  (* ======================================================================= *)
  method compute_for (inst: instance) =
    (* re-init *)
    let inst_k = inst#k
    in let prev = Array.make inst_k 0
       and dk = [|0;0|]
    in
      coverage <- 0;
      kind <- (-1);
      prevalence <- Array.make inst_k 0.0;
      homogeneity <- nan;
      hazardratio <- nan;
      (* computes coverages *)
      for o=0 to inst#m - 1 do
        try
          if (self#covers (inst#observation o)) then begin
            coverage <- coverage + 1;
            let k = (inst#observation o)#kind
            in if (k>=0) then prev.(k) <- prev.(k) + 1;
          end;
        with
          | Failure "undefined pos" -> dk.(1) <- dk.(1) + 1;
          | Failure "undefined neg" -> dk.(0) <- dk.(0) + 1;
      done;     
      (* prevalencies and kind *)
      let rbest = ref 0.0
      and tie = ref false
      and tmpkind =
        if (float_of_int prev.(0)) /. (float_of_int (inst#count_of_kind 0))
          >(float_of_int prev.(1)) /. (float_of_int (inst#count_of_kind 1))
	then 0 else 1;
      in 
      let undef =
	(not !missing_for_all)
	&& (dk.(0)+dk.(1) != 0)
        && ((float_of_int dk.(tmpkind)) /. (float_of_int (prev.(tmpkind)))
	    >= missing_threshold.(tmpkind))
      in
        if undef then begin
          trace TRACE_DEBUG "undefined coverage : ";
          let fdk = [|float_of_int dk.(0);   float_of_int dk.(1)|]
          and fpr = [|float_of_int prev.(0); float_of_int prev.(1)|]
          in
          let a = [|fpr.(0) /. (float_of_int (inst#count_of_kind 0));
                    fpr.(1) /. (float_of_int (inst#count_of_kind 1))|]
          and b = [|(fpr.(0)+.fdk.(0))/.(float_of_int (inst#count_of_kind 0));
                    (fpr.(1)+.fdk.(1))/.(float_of_int (inst#count_of_kind 1))|]
          in
            kind <-
	      if      a.(0) > b.(1) && (not (a.(1) > b.(0))) then 0
              else if a.(1) > b.(0) && (not (a.(0) > b.(1))) then 1
              else -1;
            trace TRACE_DEBUG
	      (Printf.sprintf "%d : %.2f > %.2f vs %.2f > %.2f\n"
		 kind a.(0) b.(1) a.(1) b.(0));
            if (kind<0)
	    then begin
              prevalence.(0) <- a.(0);
              prevalence.(1) <- a.(1);
              tie := true;
            end
	    else begin (* a kind/class has been found, use it! *)
	      (* dk values are assumed to be covered for the OTHER class *)
              prevalence.(kind) <- a.(kind);
              prevalence.(1-kind) <- b.(1-kind);
	      prev.(1-kind) <- prev.(1-kind) + dk.(1-kind);
              coverage <- coverage + dk.(1-kind);
            end
        end
	else
          for k=0 to inst_k-1 do
            let r = (float_of_int prev.(k))
	      /. (float_of_int (inst#count_of_kind k))
            in prevalence.(k) <- r;
              if (abs_float(r -. !rbest)) < epsilon_float then tie := true
              else if !rbest < r then begin
                kind <- k; rbest := r; tie := false
              end;
          done;
        if !tie then kind <- (-1)
        else begin
          (* homogeneity *)
          homogeneity <- (float_of_int prev.(kind)) /. (float_of_int coverage);
          (* hazard ratio *)
          let a,b = prev.(1), prev.(0)
          in let c,d = (inst#count_of_kind 1) - a, (inst#count_of_kind 0) - b
          in let fa,fb,fc,fd =
	      float_of_int a, float_of_int b, float_of_int c, float_of_int d 
          in 
            hazardratio <-
              if kind = 1 then begin
                if (c=0) then
                  if (d=0) then nan else infinity
                else (fa /. (fa +. fb)) /. (fc /. (fc +. fd))
              end else begin
                if (d=0) then
                  if (c=0) then nan else infinity
                else (fb /. (fa +. fb)) /. (fd /. (fc +. fd))
              end
        end
	  
  (* ======================================================================= *)
  method covers (o: observation) =
    let undef = ref false
    in
    let rec covrec cp = 
      match cp with
	| []    -> true
	| c::xp ->
            try if (c#covers o) then covrec xp else false
            with
              | Failure "undefined" ->
		  if (kind<0) then begin undef := true; covrec xp end
		  else
		    if !missing_for_all then false
		    else if (kind = o#kind) then false else (covrec xp)
    in
      if (covrec cuts) then begin
        if (!undef) then begin
          if o#kind=0 then failwith "undefined neg"
	  else failwith "undefined pos"
        end else true
      end else false

  (* ======================================================================= *)
  method covers_with_epsilon e (o: observation) =
    let rec covrec cp = 
      match cp with
	| []    -> true
	| c::xp -> if (c#covers_with_epsilon e o) then covrec xp else false
    in covrec cuts
	 
  (* ======================================================================= *)
  method clone ml =
    {< prevalence = Array.make  (Array.length prevalence) (-1.0);
       cuts = List.rev_append ml cuts
    >}
      
  (* ======================================================================= *)
  method distance_of_observation (obs:observation) coutout =
    let d b =
      if b then (* obs is covered *)
	let rec parse cs =
	  match cs with
	    | [] -> infinity
	    | c::cs ->
		let v = obs#value (c#variable)
		in min (min (c#ub -. v) (v -. c#lb)) (parse cs)
	in parse cuts
      else      (* obs is not covered *)
	let tmp = new observation (-1) (-1) (Array.make obs#n 0.0)
	in for i=0 to obs#n-1 do
	    let v = obs#value i
	    and l = self#lb i
	    and u = self#ub i
	    in if v < l then tmp#set_value i l
	      else if v > u then tmp#set_value i u
	      else tmp#set_value i v
	  done;
	  obs#distance tmp
    in
      if self#covers obs then d true
      else if coutout then -. (d false)
      else 0.0

  (* ======================================================================= *)
  method projections (o: observation) =
    let rec parse ms = 
      match ms with
	| []    -> []
	| m::ms ->
            match m#inequality with
              | LT _ -> let o = o#clone()
                in o#set_value (m#variable) (m#ub);
                  o::(parse ms)
              | GT _ -> let o = o#clone()
                in o#set_value (m#variable) (m#lb);
                  o::(parse ms)
              | BOX (_,_) ->
                  let o1 = o#clone()
                  and o2 = o#clone()
                  in o1#set_value (m#variable) (m#ub);
                    o2#set_value (m#variable) (m#lb);
                    o1::o2::(parse ms)
    in if self#covers o then parse cuts
      else let o = o#clone()
      in for i=0 to o#n - 1 do
        let (x,u,l) = o#value i, self#ub i, self#lb i
        in if (x<l) then o#set_value i l else if (x>u) then o#set_value i u;
      done;
        [o]
	  
  (* ======================================================================= *)
  method intersection (p: pattern) =
    let rec add (i,l,u) vs =
      match vs with
	| [] -> [i,l,u]
	| (i0,l0,u0)::vs
	  -> if i0 = i then (i, max l0 l, min u0 u)::vs
          else (i0,l0,u0)::(add (i,l,u) vs)
    in
    let rec parse ms =
      match ms with
	| [] -> []
	| m::ms -> add (m#variable,m#lb,m#ub) (parse ms) 
    in
    let rec to_monomes vs =
      match vs with
	| [] -> []
	| (i,l,u)::vs ->
            ((*if (l>u) then failwith "empty"
               else*) if (l = neg_infinity) then new monome i (LT u)
             else if (u = infinity) then new monome i (GT l)
             else new monome i (BOX (l,u)))
            ::(to_monomes vs)
    in
      new pattern (-1) (to_monomes (parse (cuts @ (p#monomes()))))
	
  (* ======================================================================= *)
  method includes (p: pattern) = 
    let rec parse ms = 
      match ms with
	| []    -> true
	| m::ms
	  -> let i = m#variable
          in let (l,u)   = self#lb i, self#ub i
             and (lp,up) = p#lb i,    p#ub i
          in l-.epsilon_float <= lp && u+.epsilon_float >= up && parse ms
    in parse cuts
          
end;;

let empty_pattern = new pattern (-1) [];;

(* =========================================================================
   comparison
   ========================================================================= *)

(* ========================================================================= *)
let compare_index p1 p2 = Indexed.compare (p1:>indexed) (p2:>indexed);;

(* ========================================================================= *)
let compare_degree p1 p2 = Pervasives.compare p1#degree p2#degree;;

(* ========================================================================= *)
let compare_coverage p1 p2 = Pervasives.compare p1#coverage p2#coverage;;

(* ========================================================================= *)
let compare_homogeneity p1 p2 =
  Pervasives.compare p1#homogeneity p2#homogeneity;;

(* ========================================================================= *)
let compare_prevalence p1 p2 =
  Pervasives.compare (p1#prevalence (p1#kind)) (p2#prevalence (p2#kind));;

(* ========================================================================= *)
let compare_equal p1 p2 =
  if p1#includes p2 && p2#includes p1 then 0 else 1

(* ========================================================================= *)
let compare_function = ref compare_degree;; (* private and hidden *)
let compare p1 p2 = !compare_function p1 p2;;

(* ========================================================================= *)
let set_compare_function f = compare_function := f;;

(* =========================================================================
   auxilary functions
   ========================================================================= *)

let rec count_kinds ps =
  match ps with
  | []    -> 0
  | p::ps -> max (count_kinds ps) p#kind
;;
