(* $Id: pattern.mli 1 2005-07-15 19:39:24Z lemaire $ *)
(** 
 The class for a pattern.
 
 @version November 16, 2004
 @author Pierre Lemaire
*)

(* ============================================================================
author: Pierre Lemaire
2004-11-01             : creation
2004-11-01..2004-11-08 : basics
============================================================================ *)

open Indexed;;
open Observation;;
open Instance;;
open Monome;;


val missing_threshold: float array
val missing_for_all: bool ref
            
(* ======================================================================= *)
(** {6 the class [pattern]}*)

(** The class [pattern] defines a pattern for LAD, that is a set of
    [monomes] with a kind/class.
    
    [pattern i mons] creates a new pattern with the monomes [mons];
    [i] is used as index.
    
    A [pattern] has a degree and a length; it has a kind/class, an homogeneity
    and prevalences with regards to a given [instance]. 
*)
class pattern:
  int -> (* index *)
  monome list -> (* init values *)
object('a)
   inherit Indexed.indexed
  (* ===================================================================== *)
  (** {6 fields} *)

  (** the list of the cut points of this [pattern]. *)
  val mutable cuts: monome list
      
  (** the number of observations of any kind/class covered by this [pattern]. *)
  val mutable coverage: int
        
  (** the class/kind of the [pattern]. *)
  val mutable kind: int

  (** the homogeneity (ratio) of this [pattern]. *)
  val mutable homogeneity: float
  
  (** the prevalences (ratio) for each kind/class of this [pattern]. *)
  val mutable prevalence: float array

  val mutable hazardratio: float
  
  (* ===================================================================== *)
  (** {6 properties} *)
  
  (** returns the degree of this [pattern] (the number of variables that
      appear in it). *)
  method degree: int

  (** returns the number of monomes of this [pattern]. *)
  method length: int

  (** list of monomes *)
  method monomes: unit -> monome list

  (** lower bound for the given variable. *)
  method lb: int -> float

 (** upper bound for the given variable. *)
  method ub: int -> float

 (** returns whether at least one monome uses this variable. *)  
  method contains: int -> bool

  (* ===================================================================== *)
  (** {6 instance-dependent properties} *)

  (** {b note:} For the following properties that returns a value, a call to
      [compute_for], or to the corresponding [set_...] method, is required
      so that that value makes sense. *)
  
  (** returns the class/kind of this [pattern] (or a negative value if the
      kind/class is unknown).
       
      The kind/class of a pattern is the class/kind of the observations the
      pattern covers the largest {e ratio} of. *)
  method kind: int
  
  method set_kind: int -> unit
  
  (** returns the homogeneity of this [pattern] (or [nan] if unknown). 
      
      The homogeneity is: A/B where A and B are the number of observations
      respectively (A) of the pattern kind/class and covered, (B) of other
      kinds/classes and covered. *)
  method homogeneity: float

  method set_homogeneity: float -> unit
    
  (** returns the number of [observations] this [pattern] covers (or a
      negative value if the coverage is unknwon). *)
  method coverage: int
    
  (** returns the prevalence  of this [pattern] for the given kind/class
      (or [nan] if unknown).
  
      The prevalence is: A/C where A and C are the number of observations
      respectively (A) of the considered kind/class and covered, (C) the
      considered kind/class. *)
  method prevalence: int -> float
  
  (** returns the maximal prevalence of this [pattern] on any kind/class.
      If the pattern kind is defined, this is equivalent to [prevalence 
      kind]. *)
  method max_prevalence: float

  (** kind must be defined! *)
  method set_max_prevalence: float -> unit
    
  method hazardratio: float
  method set_hazardratio: float -> unit
  
  (* ===================================================================== *)
  (** {6 methods} *)
    
  (** computes kind, homogeneity and prevalences of this [pattern] for the
      given [instance]. This method deletes any former values. *)
  method compute_for: instance -> unit
  
  (** returns whether this [pattern] covers the given [observation] or not. *)
  method covers: observation -> bool

  method covers_with_epsilon: float -> observation -> bool
   
  (** returns a fresh [pattern] with the additional [monomes]. *)
  method clone: monome list -> 'a

  (** returns the distance of the [observation] from this [pattern].*)
  method distance_of_observation: observation -> bool -> float

  (** returns the projections of the [observation] on this [pattern]. *)
  method projections: observation -> observation list
  
  method intersection: pattern -> pattern
  
  method includes: pattern -> bool

end

(** an empty pattern for kinds \[0..[k]-1\]. *)
val empty_pattern: pattern

(* ======================================================================= *)
(** {6 comparison}*)

(** compares 2 [pattern] according to their index. *)
val compare_index: pattern -> pattern -> int

(** compares 2 [pattern] according to their degree. *)
val compare_degree: pattern -> pattern -> int

(** compares the 2 [pattern] according to their coverage. *)
val compare_coverage: pattern -> pattern -> int

(** compares the 2 [pattern] according to their homogeneity. *)
val compare_homogeneity: pattern -> pattern -> int

(** compares the 2 [pattern] according to their prevalence. *)
val compare_prevalence: pattern -> pattern -> int

(** 0 for equality... *)
val compare_equal: pattern -> pattern -> int

(** sets the current default comparison function. *)
val set_compare_function: (pattern -> pattern -> int) -> unit

(** compares the 2 [pattern] according to the current default
    comparison function. *)
val compare: pattern -> pattern -> int

(* ======================================================================= *)
(** {6 Auxilary functions}*)

(** returns the smallest value [k] such that all [patterns] of the list have
    a kind/class, if defined, within \[0..[k]\]. *)
val count_kinds: pattern list -> int
