(* $Id: sampler.ml 14 2007-06-29 08:14:54Z pierre $ *)
(** ========================================================================
    sampler: a companion for ladoscope
    Copyright (C) Pierre Lemaire, 2005.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ======================================================================== *)

open Messages;;
open Misc;;
open Instance;;
open Io;;

open Printf;;

type sampling_type = NONE
  | SAMPLING_METHOD of float
  | FOLDING_METHOD of int
  | LEAVE_ONE_OUT of int
  ;;

(* ========================================================================= *)
let sampler arg =
try
  let name = "sampler"
  and author = "Pierre Lemaire"
  and version = "2007-06-29"
  and copyright = "2005"
  and description = "split an instance in TRA and TES sets"
  and usage_msg =
    "usage is: "^Sys.argv.(0)^" [options] [instance]"
  and finst = ref ""
  and seed = ref (-1)
  and sampling = ref NONE
  and tra = ref true
  and fold = ref (-1)
  and sort = ref false
  in let arglist =
    (common_options (name,version,description,author,copyright))
    @[
(* --- options ------------------------------------------------------------- *)
     ("-seed",
        Arg.Set_int seed,
        "(int) sets the random number seed (default: auto-init)");

     ("-sort",
        Arg.Set sort,
        " sorts the output by reading order");
     ("-nosort",
        Arg.Clear sort,
        " does not sort the output.");

     ("-tra",
	Arg.Set tra,
        " the TRA dataset.");
     ("-tes",
	Arg.Clear tra,
        " the TES dataset.");

     ("-fold",
	Arg.Set_int fold,
        "(int) the fold number.");

(* --- methods ------------------------------------------------------------- *)
     ("-sampling",
        Arg.Float (fun s -> sampling := SAMPLING_METHOD s),
        "(float) TRA is made of the given ratio of pos and neg observations.");

     ("-folding",
        Arg.Int (fun s -> sampling := FOLDING_METHOD s),
        "(int) TRA is made of (k-1) of the given number of folds.");

     ("-leaveoneout",
	Arg.Int  (fun s -> sampling := LEAVE_ONE_OUT s),
        "(int) leave  the given one out as TES.");

    ]
  in
    (* initiate program *)
    if (Array.length arg = 0) then
      Arg.parse (Arg.align arglist) (fun s -> finst := s) usage_msg
    else begin
      let x = ref 0
      in
	Arg.parse_argv
	  ~current:x
	  arg
	  (Arg.align arglist)
	  (fun s -> finst := s) usage_msg
    end;
    flush_all();
    if (!seed < 0) then Random.self_init() else Random.init(!seed);
    Observation.set_compare_function Observation.compare_index;

   (* -------------------------------------------------------------------- *)
    let read_data () = 
      let inst = read_instance (open_in !finst)
      and pos,neg = ref [], ref []
      and comp (a,_) (c,_) = Pervasives.compare a c 
      in 
	if !sort then inst#sort();
	for o=0 to inst#m - 1 do
	  let o = inst#observation o
	  in
	    if o#kind = 0 then neg := (Random.float 1.0, o)::!neg
	    else if o#kind > 0 then pos := (Random.float 1.0, o)::!pos
	done;
	List.sort comp !pos, List.sort comp !neg
    in

   (* -------------------------------------------------------------------- *)
    let rec print xs = match xs with
	[] -> ()
      | x::xs -> let _,o = x in
	  print_observation (out()) o; 
	  output_string (out()) "\n";
	  print xs
    in

    begin match !sampling with
   (* ==================================================================== *)
    | NONE -> ()

   (* ==================================================================== *)
    | SAMPLING_METHOD r ->
      let rec split k xs = match xs with
	  []    -> [], []
	| x::xs ->
	    if k<=0
	    then [], xs
	    else let a,b = split (k-1) xs in x::a, b
      in
      let pos,neg = read_data ()
      in let npos,nneg = List.length pos, List.length neg
      in let kpos,kneg =
	  int_of_float (r *. (float_of_int npos)),
	  int_of_float (r *. (float_of_int nneg))
      in let trapos,tespos = split kpos pos
	 and traneg,tesneg = split kneg neg
      in let ans =
	  (if !tra then trapos @ traneg else tespos @ tesneg)
      in 
	print ans

   (* ==================================================================== *)
    | FOLDING_METHOD n ->
      let rec split fs k xs = match xs with
	  []    -> ()
	| x::xs -> 
	    fs.(k) <- x::(fs.(k));
	    split fs ((k + 1) mod (Array.length fs)) xs
      in
      let pos,neg = read_data ()
      and kpos,kneg = Array.make n [], Array.make n []
      in
	if (!fold < 0 || !fold >= n) then error 1 "invalid fold number";
	split kpos 0 pos;
	split kneg 0 neg;
	if !tra then begin
	  for k=0 to n-1 do
	    if k <> !fold then begin print kpos.(k); print kneg.(k); end
	  done
	end else begin print kpos.(!fold); print kneg.(!fold); end

   (* ==================================================================== *)
    | LEAVE_ONE_OUT n ->
      let inst = read_instance (open_in !finst)
      in 
	if !sort then inst#sort();
	for o=0 to inst#m - 1 do
	  if (!tra) && (o!=n) then begin
	    print_observation (out()) (inst#observation o);
	    output_string (out()) "\n"
	  end
	  else if not !tra && o = n then begin
	    print_observation (out()) (inst#observation o);
	    output_string (out()) "\n"
	  end
	done;

   (* ==================================================================== *)
    end;
    0;
with
  | _ as ex -> terminal_catch ex
;;

