(* $Id: sometests.ml 1 2005-07-15 19:39:24Z lemaire $ *)
(* This program tests several features of the LAD-O'Scope
  implementation. *)
  
(* ============================================================================
author: Pierre Lemaire
2004-10-29             : creation
============================================================================ *)

open Messages;;
open Misc;;
open Observation;;
open Instance;;
open Cutpointset;;
open Monome;;
open Pattern;;
open Pandect;;
open Model;;
open Io;;

(* ========================================================================= *)
(* the instance to work with
   x1
   4 - + ? + +
   3 + - ? + +
   2 - - - - ? -
   1 + +   + 
   0 + +   +
     0 1 2 3 4 5 x0
*)

let inst = new instance 3
             [| new observation  1    1 [|0.0;0.0|];
                new observation  2    1 [|1.0;0.0|];
                new observation  3    1 [|3.0;0.0|];
                new observation  4    1 [|0.0;1.0|];
                new observation  5    1 [|1.0;1.0|];
                new observation  6    1 [|3.0;1.0|];
                new observation  7    0 [|0.0;2.0|];
                new observation  8    0 [|1.0;2.0|];
                new observation  9    0 [|2.0;2.0|];
                new observation 10    0 [|3.0;2.0|];
                new observation 11 (-1) [|4.0;2.0|];
                new observation 12    0 [|5.0;2.0|];
                new observation 12    1 [|0.0;3.0|];
                new observation 12    0 [|1.0;3.0|];
                new observation 12 (-1) [|2.0;3.0|];
                new observation 12    1 [|3.0;3.0|];
                new observation 12    1 [|4.0;3.0|];
                new observation 12    0 [|0.0;4.0|];
                new observation 12    1 [|1.0;4.0|];
                new observation 12 (-1) [|2.0;4.0|];
                new observation 12    1 [|3.0;4.0|];
                new observation 12    1 [|4.0;4.0|]|];

;;

(* ========================================================================= *)
let test b = output (if b then "ok\n" else "FAILED!\n");;

(* ========================================================================= *)
let sometests()=
try
  let usage_msg = "usage is: "^Sys.argv.(0)^" [options]"
  and seed = ref (-1)
  and all = ref false
  and obs = ref false
  and ins = ref false
  and cut = ref false
  and mon = ref false
  and pat = ref false
  and pan = ref false
  and mol = ref false
  in let arglist =
    [("-trace",
        Arg.String (set_trace_level_from_string),
        "(val) sets the verbosity [none|some|most|all] of the trace (default: none)");
     ("-ftrace",
        Arg.String (set_trace_file),
        "(file) sets where the trace is printed (default: stderr)");
     ("-fwarning",
        Arg.String (set_warning_file),
        "(file) sets where the warnings are printed (default: stderr)");
     ("-ferror",
        Arg.String (set_error_file),
        "(file) sets where the errors are printed (default: stderr)");
     ("-o",
        Arg.String (set_output_file),
        "(file) sets where the output is printed (default: stdout)");
     ("-seed",
        Arg.Int (set_ref seed),
        "(int) sets the random number seed (default: auto-init)");
     ("-debug",
        Arg.Unit (fun () -> set_trace_level TRACE_DEBUG),
        "debug mode (a lot of trace)");
     ("-all", Arg.Set (all), "performs all tests");
     ("-obs", Arg.Set (obs), "performs tests for observations");
     ("-ins", Arg.Set (ins), "performs tests for instances");
     ("-cut", Arg.Set (cut), "performs tests for cut-point sets");
     ("-mon", Arg.Set (mon), "performs tests for monomes");
     ("-pat", Arg.Set (pat), "performs tests for patterns");
     ("-pan", Arg.Set (pan), "performs tests for pandects");
     ("-mod", Arg.Set (mol), "performs tests for models");
    ]
  in
    (* ==================================================================== *)
    (* initiate program *)
    Arg.parse (Arg.align arglist) (fun x -> ()) usage_msg;
    if (!seed < 0) then Random.self_init() else Random.init(!seed);

    Printf.printf "nan=%f\n" nan;

    (* ==================================================================== *)
    (* Tests for observation *)
    if (!all || !obs) then begin
      trace TRACE_SOME "Tests for [observation]...\n";
      let o0 = empty_observation
      and o1 = new observation 7 3 [|1.1;2.2;3.3;4.4|]
      and o2 = new observation 7 3 [|nan|]
      in
        trace TRACE_MOST "...properties\n";
        test (o0#n = 0);
        test (o1#index = 7 && o1#kind = 3 && o1#n = 4
              && o1#value 0 = 1.1 && o1#value 1 = 2.2
              && o1#value 2 = 3.3 && o1#value 3 = 4.4
              );
        test (Pervasives.compare (o2#value 0) nan = 0);
      let o1 = new observation 1 0 [|1.1;2.2;3.3;4.4;5.5|]
      and o2 = new observation 2 0 [|5.5|]
      and o3 = new observation 1 1 [|5.5;2.2;3.3;4.4;5.5|]
      and o4 = new observation 1 0 [|1.1;2.2;3.3;4.4;5.7|]
      in 
        trace TRACE_MOST "...comparisons\n";
        Observation.set_compare_function Observation.compare_index;
        test (  Observation.compare o1 o3 = 0 
             && Observation.compare o1 o2 < 0
             && Observation.compare o2 o1 > 0);
        Observation.set_compare_function compare_kind;
        test (  Observation.compare o1 o2 = 0
             && Observation.compare o1 o3 < 0
             && Observation.compare o3 o1 > 0);
        Observation.set_compare_function (compare_variable 0);
        test (  Observation.compare o2 o3 = 0
             && Observation.compare o1 o2 < 0
             && Observation.compare o2 o1 > 0);
        Observation.set_compare_function compare_all;
        test (  Observation.compare o1 o1 = 0
             && Observation.compare o1 o2 < 0
             && Observation.compare o2 o1 > 0
             && Observation.compare o1 o3 < 0
             && Observation.compare o3 o1 > 0
             && Observation.compare o1 o4 < 0
             && Observation.compare o4 o1 > 0);
      let o1 = new observation 0 0 [|1.1|]
      and o2 = new observation 0 0 [|1.1;2.2;3.3;4.4|]
      in 
        trace TRACE_MOST "...exceptions\n";
        try ignore (o1#value 1); test false;
        with Invalid_argument "Observation.value: no such variable"
        ->  test true;
        Observation.set_compare_function (compare_variable 3);
        try ignore (Observation.compare o1 o2); test false;
        with Invalid_argument "Observation.value: no such variable"
        ->  test true;
     end;

    (* ==================================================================== *)
    (* Tests for instance *)
    if (!all || !ins) then begin
      trace TRACE_SOME "Tests for [instance]...\n";
      let i1 = new instance 4
               [| new observation 1 0 [|1.1;2.1;3.1|];
                  new observation 2 1 [|1.2;2.2;3.2|];
                  new observation 3 0 [|1.3;2.3;3.3|];
                  new observation 4 1 [|1.4;2.4;3.4|];
                  new observation 5 0 [|1.5;2.5;3.5|];
                  new observation 6 1 [|1.6;2.6;3.6|];
                  new observation 7 2 [|1.7;2.7;3.7|]|]
      in
        trace TRACE_MOST "...properties\n";        
        test (i1#k = 4 && i1#m = 7 && i1#n = 3 && i1#count_kinds = 4
              && i1#count_observations = 7 && i1#count_variables = 3);
        test (i1#count_of_kind 0 = 3 && i1#count_of_kind 1 = 3
             && i1#count_of_kind 2 = 1 && i1#count_of_kind 3 = 0
             && i1#count_undefined = 0);
        
        i1#set_observation 2 (new observation 1 (-1) [|0.7;0.7;0.7|]);
        test (i1#k = 4 && i1#m = 7 && i1#n = 3 && i1#count_kinds = 4
              && i1#count_observations = 7 && i1#count_variables = 3);
        test (i1#count_of_kind 0 = 2 && i1#count_of_kind 1 = 3
             && i1#count_of_kind 2 = 1 && i1#count_undefined = 1);
      
      let i1 = new instance 3
               [| new observation 1 0 [|1.1;2.1;3.1|];
                  new observation 2 1 [|1.2;2.2;3.2|];
                  new observation 3 0 [|1.3;2.3;3.3|];
                  new observation 4 2 [|1.4;2.4;3.4|];
                  new observation 5 0 [|1.5;2.5;3.5|];
                  new observation 6 1 [|1.6;2.6;3.6|];
                  new observation 7 1 [|1.7;2.7;3.7|]|]
      in
        trace TRACE_MOST "...sorting\n";
        Observation.set_compare_function compare_kind;
        i1#sort();
        test (  (i1#observation 0)#index = 1
             && (i1#observation 1)#index = 3
             && (i1#observation 2)#index = 5
             && (i1#observation 3)#index = 2
             && (i1#observation 4)#index = 6
             && (i1#observation 5)#index = 7
             && (i1#observation 6)#index = 4);
        Observation.set_compare_function Observation.compare_index;
        i1#sort();
        test (  (i1#observation 0)#index = 1
             && (i1#observation 1)#index = 2
             && (i1#observation 2)#index = 3
             && (i1#observation 3)#index = 4
             && (i1#observation 4)#index = 5
             && (i1#observation 5)#index = 6
             && (i1#observation 6)#index = 7);
        i1#mixup();
        test (  (i1#observation 0)#index <> 1
             || (i1#observation 1)#index <> 2
             || (i1#observation 2)#index <> 3
             || (i1#observation 3)#index <> 4
             || (i1#observation 4)#index <> 5
             || (i1#observation 5)#index <> 6
             || (i1#observation 6)#index <> 7);
        i1#sort();
        test (  (i1#observation 0)#index = 1
             && (i1#observation 1)#index = 2
             && (i1#observation 2)#index = 3
             && (i1#observation 3)#index = 4
             && (i1#observation 4)#index = 5
             && (i1#observation 5)#index = 6
             && (i1#observation 6)#index = 7);
             
      let i1 = new instance 3
               [| new observation 1 0 [|1.1;2.1;3.1|];
                  new observation 2 1 [|1.2;2.2;3.2|]|]
      in
        trace TRACE_MOST "...exceptions\n";
        try ignore (i1#count_of_kind 3); test false;
        with Invalid_argument "Instance.count_of_kind: no such kind"
        ->  test true;
        try ignore (i1#observation 2); test false;
        with Invalid_argument "Instance.observation: no such observation"
        ->  test true;
        try ignore (i1#set_observation 3 empty_observation); test false;
        with Invalid_argument "Instance.set_observation: no such observation"
        ->  test true;
        try ignore (i1#set_observation 1 (new observation 0 5 [||])); test false;
        with Invalid_argument "Instance.set_observation: no such kind"
        ->  test true;
    end;
    
    (* ==================================================================== *)
    (* Tests for cut-point sets *)
    if (!all || !cut) then begin
      trace TRACE_SOME "Tests for [cutpointset]...\n";
      let c1 = new cutpointset 0 [7.1;3.2;5.3;2.4]
      in
        trace TRACE_MOST "...properties\n"; 
        test (c1#variable = 0 && c1#count = 4
             && (abs_float (c1#first -. 7.1) < epsilon_float)
             && (abs_float (c1#next  -. 3.2) < epsilon_float)
             && (abs_float (c1#next  -. 5.3) < epsilon_float)
             && (abs_float (c1#next  -. 2.4) < epsilon_float)
             && (Pervasives.compare c1#next nan = 0));
      let c1 = new cutpointset 0 [7.1;3.2;5.3;2.4]
      in
        trace TRACE_MOST "...manipulation\n";
        test (not (c1#exists 3.4) && (c1#add 3.4; c1#exists 3.4));
        test ((c1#exists 3.2) && (c1#del 3.2; not (c1#exists 3.2)));
        test (c1#exists 5.3 && not (c1#exists 3.2) && c1#exists 3.4);
      let c1 = new cutpointset 0 [7.1;3.2;5.3;2.4]
      in
        trace TRACE_MOST "...sorting\n"; 
        c1#sort();
        test (c1#variable = 0 && c1#count = 4
             && (abs_float (c1#first -. 2.4) < epsilon_float)
             && (abs_float (c1#next  -. 3.2) < epsilon_float)
             && (abs_float (c1#next  -. 5.3) < epsilon_float)
             && (abs_float (c1#next  -. 7.1) < epsilon_float)
             && (Pervasives.compare c1#next nan = 0));
        c1#sort_decreasing();
        test (c1#variable = 0 && c1#count = 4
             && (abs_float (c1#first -. 7.1) < epsilon_float)
             && (abs_float (c1#next  -. 5.3) < epsilon_float)
             && (abs_float (c1#next  -. 3.2) < epsilon_float)
             && (abs_float (c1#next  -. 2.4) < epsilon_float)
             && (Pervasives.compare c1#next nan = 0));       
      let c0 = Cutpointset.all inst 0
      and c1 = Cutpointset.all inst 1
      in
        trace TRACE_MOST "...building\n";
        test (  c0#count = 5 && c0#exists 0.5 && c0#exists 1.5
             && c0#exists 2.5 && c0#exists 3.5 && c0#exists 4.5);
        test (c1#count = 3 && c1#exists 1.5 && c1#exists 2.5 && c1#exists 3.5);
    end;

    (* ==================================================================== *)
    (* Tests for monomes *)
    if (!all || !mon) then begin
      trace TRACE_SOME "Tests for [monome]...\n";
      let m1 = new monome 0 (LT 2.15)
      and m2 = new monome 1 (GT 1.15)
      and m3 = new monome 0 (BOX (2.5,3.5))
      in
        trace TRACE_MOST "...properties\n";
        test (  m1#variable = 0 && m2#variable = 1 && m3#variable = 0
             && m1#lb = neg_infinity
             && abs_float(m1#ub -. 2.15) < epsilon_float
             && abs_float(m2#lb -. 1.15) < epsilon_float
             && m2#ub = infinity
             && abs_float(m3#lb -. 2.5 ) < epsilon_float
             && abs_float(m3#ub -. 3.5 ) < epsilon_float);
        trace TRACE_MOST "...methods\n";
        test (  m1#covers (inst#observation 3) && not (m1#covers (inst#observation 15))
             && m2#covers (inst#observation 9) && not (m2#covers (inst#observation 1))
             && m3#covers (inst#observation 2) && not (m3#covers (inst#observation 12)) && not (m3#covers (inst#observation 16)))
    end; 
    
    (* ==================================================================== *)
    (* Tests for patterns *)
    if (!all || !pat) then begin
      trace TRACE_SOME "Tests for [pattern]...\n";
      let p1 = new pattern 1 [new monome 1 (GT 1.25); new monome 0 (GT 2.75)]
      and p2 = new pattern 2 [new monome 1 (LT 1.5)]
      and p3 = new pattern 3 [new monome 1 (BOX(1.5,3.5)); new monome 0 (LT 2.5)]
      and p4 = new pattern 4 [new monome 0 (LT 0.0); new monome 0 (GT 2.0)]
      in
        trace TRACE_MOST "...properties\n";
        test (  p1#degree = 2 && p1#length = 2
             && p2#degree = 1 && p2#length = 1
             && p3#degree = 2 && p3#length = 2
             && p4#degree = 1 && p4#length = 2);
        p1#compute_for inst; p2#compute_for inst; p3#compute_for inst;
        test (  p1#kind = 1 && p2#kind = 1 && p3#kind = 0);
        test (  p1#coverage = 7 && p2#coverage = 6 && p3#coverage = 6);
        test (  abs_float(p1#homogeneity -. 4.0/.7.0) < epsilon_float
             && abs_float(p2#homogeneity -. 1.0) < epsilon_float
             && abs_float(p3#homogeneity -. 4.0/.6.0) < epsilon_float);
        test (  abs_float((p1#prevalence 0) -. 2.0/. 7.0) < epsilon_float
             && abs_float((p1#prevalence 1) -. 4.0/.12.0) < epsilon_float
             && abs_float((p2#prevalence 0) -. 0.0/. 7.0) < epsilon_float
             && abs_float((p2#prevalence 1) -. 6.0/.12.0) < epsilon_float
             && abs_float((p3#prevalence 0) -. 4.0/. 7.0) < epsilon_float
             && abs_float((p3#prevalence 1) -. 1.0/.12.0) < epsilon_float);
             
        trace TRACE_MOST "...methods\n";
        test (  p1#covers (inst#observation 9) && not (p1#covers (inst#observation 1))
             && p2#covers (inst#observation 3) && not (p2#covers (inst#observation 6))
             && p3#covers (inst#observation 7) && not (p3#covers (inst#observation 16)));
      
      let p1 = new pattern 1 [new monome 1 (GT 1.25); new monome 0 (GT 2.75)]
      in let p2 = p1#clone []
         and p3 = p1#clone [new monome 3 (GT 2.3)]
         and p4 = p1#clone [new monome 0 (GT 2.3); new monome 1 (LT 2.0)]
      in 
        trace TRACE_MOST "...clone\n";
        test (  p1#degree = 2 && p1#length = 2
             && p2#degree = 2 && p2#length = 2
             && p3#degree = 3 && p3#length = 3
             && p4#degree = 2 && p4#length = 4);
             
      let p1 = new pattern 0 [new monome 0 (LT 99.9)]
      and p2 = new pattern 0 [new monome 0 (GT 99.9)]
      in
        trace TRACE_MOST "...limit cases\n";
        p1#compute_for inst; p2#compute_for inst;
        test (p1#kind = -1 && p2#kind = -1);
    end;

    (* ==================================================================== *)
    (* Tests for pandects *)
    if (!all || !pan) then begin
      trace TRACE_SOME "Tests for [pandect]...\n";
      let p1 = new pattern 1 [new monome 1 (GT 1.25); new monome 0 (GT 2.75)]
      and p2 = new pattern 2 [new monome 1 (LT 1.5)]
      and p3 = new pattern 3 [new monome 1 (BOX(1.5,3.5)); new monome 0 (LT 2.5)]
      and p4 = new pattern 4 [new monome 0 (LT 0.0);]
      in p1#compute_for inst; p2#compute_for inst;
         p3#compute_for inst; p4#compute_for inst;
      let q1 = new pandect 1 [p1;p2;p3;p4]
      in
        trace TRACE_MOST "...properties\n";
        test (  q1#count_of_kind 0 = 1 && q1#count_of_kind 1 = 2
             && q1#count_of_undef() = 1);
    end;

    (* ==================================================================== *)
    (* Tests for models *)
    if (!all || !mol) then begin
      trace TRACE_SOME "Tests for [model]...\n";
      let p1 = new pattern 1 [new monome 1 (GT 1.25); new monome 0 (GT 2.75)]
      and p2 = new pattern 2 [new monome 1 (LT 1.5)]
      and p3 = new pattern 3 [new monome 1 (BOX(1.5,3.5)); new monome 0 (LT 2.5)]
      in p1#compute_for inst; p2#compute_for inst; p3#compute_for inst;
      let m1 = new model 1 [p1;p2;p3]
      in
        m1#compute_for inst;
        trace TRACE_MOST "...properties\n";
        test (  m1#count_right = 14 && m1#count_wrong = 3 && m1#count_undef = 2);
        test (  m1#count_right_of_kind 0 = 4
             && m1#count_right_of_kind 1 = 10
             && m1#count_wrong_of_kind 0 = 2
             && m1#count_wrong_of_kind 1 = 1
             && m1#count_undef_of_kind 0 = 1
             && m1#count_undef_of_kind 0 = 1);
        test (abs_float(m1#accuracy -. 15.0/.19.0) < epsilon_float);
        test (  abs_float((m1#sensibility 0) -. 4.0/.7.0) < epsilon_float
             && abs_float((m1#sensibility 1) -. 10.0/.12.0) < epsilon_float);
    end;

    (* ==================================================================== *)
    (* terminates program *)
    close_messages();
    exit 0;
with
  | IO_error t -> print_io_error t
  | _ as ex    -> at_exit (fun () ->
                    error (-1) "(unexpected): this is probably a bug.");
                  raise ex
;;

(* ========================================================================= *)
if !Sys.interactive then () else sometests();;
