(* $Id: stat_instance.ml 14 2007-06-29 08:14:54Z pierre $ *)
(** ========================================================================
    stat-instance: a companion for ladoscope
    Copyright (C) Pierre Lemaire, 2005.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ======================================================================== *)

open Messages;;
open Misc;;
open Instance;;
open Io;;

open Printf;;


(* ========================================================================= *)
let index_of_string s =
  try
    List.map
      int_of_string
      (Str.split (Str.regexp "[ \t\n\r]+") s)
  with Failure _ -> raise (IO_error (Misc "parse error reading indices"))
;;

let read_transformations filename =
  let rec parse f =
    try let s = input_line f in (index_of_string s)::(parse f)
    with End_of_file -> []
  in
    parse (open_file filename)
;;

(* ========================================================================= *)

(* what is (list of transformations as returned by read_transformation) *)
let rec what_is i ls =
  let rec find_one i n xs = match xs with
    | [] -> -1
    | x::xs -> if i==x then n else find_one i (n+1) xs
  in match ls with
    | [] -> -1
    | l::[] -> find_one i 0 l 
    | l::ls ->
	let i = find_one i 0 l
	in if (i>=0) then what_is i ls else -1
;;

(* what was (list of transformations reversed) *)

let rec what_was i ls =
  let find_one n xs = 
    try List.nth xs n
    with
      | Failure "nth"
      | Invalid_argument "List.nth" -> -1
  in match ls with
    | [] -> -1
    | l::[] -> find_one i l 
    | l::ls ->
	let i = find_one i l
	in if (i>=0) then what_was i ls else -1
;;

(* ========================================================================= *)
type instance_action =
    STATS
  | COLUMNS of string | LINES of string
  | WHATIS of int | WHATWAS of int | SIMPLIFY
  | BOUNDS of (float*float)
  | FILTER of (float*float)
  | LOG of float
  | NORM_LINE
  | NORM_COL;;

let stat_instance arg =
  try
    let name = "stat_instance"
    and author = "Pierre Lemaire"
    and version = "2007-06-29"
    and copyright = "2005"
    and description = "computes some characteristic of an instance"
    and usage_msg = "usage is: "^Sys.argv.(0)^" [options] instance"
    and infile = ref ""
    and nbvar = ref true
    and nbobs = ref true
    and nbpos = ref true
    and nbneg = ref true
    and misval = ref true
    and misobs = ref true
    and fdumb = ref 0.
    and action = ref STATS
    and sort = ref true
    in let arglist =
	(common_options (name,version,description,author,copyright))
	@(eol_options)
	@[
	  (* --- common options ----------------------------------- *)
	  ("-sort",
           Arg.Set sort,
           " sorts the output by reading order");
	  ("-nosort",
           Arg.Clear sort,
           " does not sort the output.");
	  ("-all",
           Arg.Unit (fun () ->
		       set_names true;
		       nbvar := true;
		       nbobs := true;
		       nbpos := true;
		       nbneg := true;
		       misval := true;
		       misobs := true;
		    ),
           " turns everything on.");
	  ("-none",
           Arg.Unit (fun () ->
		       set_names false;
		       nbvar := false;
		       nbobs := false;
		       nbpos := false;
		       nbneg := false;
		       misval := false;
		       misobs := false;
		    ),
           " turns everything off.");
	  
	  (* --- stats options ------------------------------------- *)

	  ("-nbvar",
           Arg.Set nbvar,
           " print the number of variables.");
	  ("-nonbvar",
           Arg.Clear nbvar,
           " do not print the number of variables.");
	  ("-nbobs",
           Arg.Set nbobs,
           " print the number of observations.");
	  ("-nonbobs",
           Arg.Clear nbobs,
           " do not print the number of observations.");
	  ("-nbpos",
           Arg.Set nbpos,
           " print the number of positive observations.");
	  ("-nonbpos",
           Arg.Clear nbpos,
           " do not print the number of positive observations.");
	  ("-nbneg",
           Arg.Set nbneg,
           " print the number of negative observations.");
	  ("-nonbneg",
           Arg.Clear nbneg,
           " do not print the number of negative observations.");
	  ("-allobs",
           Arg.Unit (fun () -> nbobs := true; nbpos := true; nbneg := true),
           " turns on all observation related features.");
	  ("-noneobs",
           Arg.Unit (fun () -> nbobs := false; nbpos := false; nbneg := false),
           " turns off all observation related features.");

	  ("-misval",
           Arg.Set misval,
           " print ratio of missing values.");
	  ("-noamisval",
           Arg.Clear misval,
           " do not print ratio of missing values.");
	  ("-misobs",
           Arg.Set misobs,
           " print ratio of observations with missing values.");
	  ("-noamisobs",
           Arg.Clear misobs,
           " do not print ratio of observations with missing values.");

	  (* --- sub options --------------------------------------- *)


	  (* --- action -------------------------------------------- *)
	  ("-stats",
	   Arg.Unit (fun () -> action := STATS),
	   " output stats (default behavior).");
	  ("-columns",
	   Arg.String (fun s -> action := COLUMNS s),
	   " extract the given columns from the instance.");
	  ("-lines",
	   Arg.String (fun s -> action := LINES s),
	   " extract the given lines from the instance.");
 	  ("-whatis",
	   Arg.Int (fun i -> action := WHATIS i),
	   " give the new index of the given original line/column.");
 	  ("-whatwas",
	   Arg.Int (fun i -> action := WHATWAS i),
	   " give the original index of the given new line/column.");
 	  ("-simplify",
	   Arg.Unit (fun () -> action := SIMPLIFY),
	   " simplify an index file to a single transformation.");
	  ("-bounds",
	   Arg.Tuple
	     [Arg.Float (fun i -> fdumb := i);
	      Arg.Float (fun i -> action := BOUNDS (!fdumb,i))],
	   "(float*float) force values between the 2 bounds.");
	  ("-filter",
	   Arg.Tuple
	     [Arg.Float (fun i -> fdumb := i);
	      Arg.Float (fun i -> action := FILTER (!fdumb,i))],
	   "(float*float) list features with max/min >= $1, max-min >= $2");
 	  ("-log",
	   Arg.Float (fun i -> action := LOG i),
	   "(float) transform x into log(x), in the given basis.");
 	  ("-normline",
	   Arg.Unit (fun () -> action := NORM_LINE),
	   " normalize, by line.");
 	  ("-normcol",
	   Arg.Unit (fun () -> action := NORM_COL),
	   " normalize, by column.");
	]
    in
      (* initiate program *)

      if (Array.length arg = 0) then
	Arg.parse (Arg.align arglist) (fun s -> infile := s) usage_msg
      else begin
	let x = ref 0
	in
	  Arg.parse_argv
	    ~current:x
	    arg
	    (Arg.align arglist)
	    (fun s -> infile := s) usage_msg
      end;
      flush_all();

      Observation.set_compare_function Observation.compare_index;
      begin match (!action) with
	  
	(* ======================================================== *)
	| STATS -> 
	    let inst = read_instance (open_in !infile)
	    in
	      if !sort then inst#sort();
	      let missingval,missingobs =
		if (!misval || !misobs) then
		  let r = ref 0
		  and s = ref 0
		  in for o=0 to inst#m-1 do
		      let o=inst#observation o
		      and b = ref false
		      in for i=0 to inst#n-1 do
			  if Pervasives.compare (o#value i) nan = 0 then begin
			    r := !r + 1;
			    b := true;
			  end
			done;
			if !b then s := !s + 1
		    done;
		    ((float_of_int !r) /. (float_of_int (inst#m * inst#n)),
		     (float_of_int !s) /. (float_of_int (inst#m))
		    )
		else 0.0, 0.0
	      in
		if (!nbvar) then
		  fprintf (out()) "%s%d%s"
		    (if (names()) then "#var = " else "")
		    inst#n
		    !eol;
		if (!nbobs) then
		  fprintf (out()) "%s%d%s"
		    (if (names()) then "#obs = " else "")
		    inst#m
		    !eol;
		if (!nbpos) then
		  fprintf (out()) "%s%d%s"
		    (if (names()) then "#pos = " else "")
		    (inst#count_of_kind 1)
		    !eol;
		if (!nbneg) then
		  fprintf (out()) "%s%d%s"
		    (if (names()) then "#neg = " else "")
		    (inst#count_of_kind 0)
		    !eol;
		if (!misval) then
		  fprintf (out()) "%s%f%s"
		    (if (names()) then "missing values = " else "")
		    missingval
		    !eol;
		if (!misobs) then
		  fprintf (out()) "%s%f%s"
		    (if (names()) then "obs w/ missing = " else "")
		    missingobs
		    !eol;

	(* ======================================================== *)
	| COLUMNS s ->
	    let inst = read_instance (open_in !infile)
	    in
	      if !sort then inst#sort();
	      let rec parse o is =
		match is with
		  | [] -> fprintf (out()) "%s" !eol
		  | i::is -> 
		      fprintf (out()) ";%f" (o#value i);
		      parse o is
	      in
		for o=0 to inst#m - 1 do
		  if (names()) then
		    fprintf (out()) "%s;" (inst#observation o)#name;
		  fprintf (out()) "%d" ((inst#observation o)#kind);
		  parse (inst#observation o) (index_of_string s)
		done;

	(* ======================================================== *)
	| LINES s ->
	    let inst = read_instance (open_in !infile)
	    in
	      if !sort then inst#sort();
	      let rec parse is = 
		match is with
		  | [] -> ()
		  | i::is ->
		      fprintf (out()) "%s%s"
			(string_of_observation (inst#observation i))
			!eol;
		      parse is		      
	      in
		parse (index_of_string s)

	(* ======================================================== *)
	| WHATIS i ->
	    fprintf (out()) "%d%s"
	      (what_is i (read_transformations !infile))
	      !eol

	(* ======================================================== *)
	| WHATWAS i ->
	    fprintf (out()) "%d%s"
	      (what_was i (List.rev (read_transformations !infile)))
	      !eol

	(* ======================================================== *)
	| SIMPLIFY ->
	    begin try
	      let ts = List.rev (read_transformations !infile)
	      in let n = List.length (List.hd ts)
	      in
		for i=0 to n-2 do fprintf (out()) "%d " (what_was i ts) done;
		fprintf (out()) "%d%s" (what_was (n-1) ts) !eol
	    with
		Failure "hd" -> fprintf (out()) "%s" !eol
	    end

	(* ======================================================== *)
	| BOUNDS x ->
	    let lb,ub = x
	    and inst = read_instance (open_in !infile)
	    in
	      if !sort then inst#sort();
	      for o=0 to inst#m-1 do
		let o=inst#observation o
		in for i=0 to inst#n-1 do
		    let x = min ub (max lb (o#value i))
		    in o#set_value i x
		  done;
		  fprintf (out()) "%s%s" (string_of_observation o) !eol;
	      done;

	(* ======================================================== *)
	| FILTER x ->
	    let qx,dx = x
	    and inst = read_instance (open_in !infile)
	    in
	      if !sort then inst#sort();
	      for i=0 to inst#n-1 do
		let smin = ref infinity
		and smax = ref neg_infinity
		in for o=0 to inst#m-1 do
		    let x = (inst#observation o)#value i
		    in
		      smin := min !smin x;
		      smax := max !smax x;
		  done;
		  if (!smax /. !smin >= qx) && (!smax -. !smin >= dx) then
		    fprintf (out()) "%d " i;
	      done;
	      fprintf (out()) "%s" !eol;

	(* ======================================================== *)
	| LOG x ->
	    let inst = read_instance (open_in !infile)
	    in
	      if !sort then inst#sort();
	      for o=0 to inst#m-1 do
		let o=inst#observation o
		in for i=0 to inst#n-1 do
		    o#set_value i ((log (o#value i)) /. (log x));
		  done;
		  fprintf (out()) "%s%s" (string_of_observation o) !eol;
	      done;

	(* ======================================================== *)
	| NORM_LINE ->
	    let inst = read_instance (open_in !infile)
	    in
	      if !sort then inst#sort();
	      for o = 0 to inst#m - 1 do
		let o = inst#observation o
		and sum = ref 0.0
		and sumsq = ref 0.0
		in
		  for v=0 to inst#n - 1 do
		    sum := !sum +. (o#value v);
		    sumsq := !sumsq +. (o#value v) *. (o#value v);
		  done;
		  let n = float_of_int inst#n
		  in let m = !sum /. n
		     and s = sqrt(((!sumsq)-.(!sum *. !sum)/.n)/. n)
		  in
		    for v=0 to inst#n - 1 do
		      let x = o#value v in o#set_value v ((x-.m)/.s)
		    done;
		    fprintf (out()) "%s%s" (string_of_observation o) !eol;
	      done;

	(* ======================================================== *)
	| NORM_COL ->
	    let inst = read_instance (open_in !infile)
	    in
	      if !sort then inst#sort();

	      for v = 0 to inst#n - 1 do
		let sum = ref 0.0
		and sumsq = ref 0.0
		in
		  for o=0 to inst#m - 1 do
		    let x = (inst#observation o)#value v;
		    in
		      sum := !sum +. x;
		      sumsq := !sumsq +. x *. x;
		  done;
		  let n = float_of_int inst#m
		  in let m = !sum /. n
		     and s = sqrt(((!sumsq)-.(!sum *. !sum)/.n)/. n)
		  in
		    for o=0 to inst#m - 1 do
		      let x = (inst#observation o)#value v
		      in (inst#observation o)#set_value v ((x-.m)/.s)
		    done;
	      done;
	      for o=0 to inst#m - 1 do
		fprintf (out()) "%s%s"
		  (string_of_observation (inst#observation o))
		  !eol;
	      done;

      end;
      (* terminates program *)
      0;
  with
    | _ as ex -> terminal_catch ex
;;

