(* $Id: stat_model.ml 14 2007-06-29 08:14:54Z pierre $ *)
(** ========================================================================
    stat-model: a companion for ladoscope
    Copyright (C) Pierre Lemaire, 2005.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ======================================================================== *)

open Messages;;
open Misc;;
open Instance;;
open Pattern;;
open Pandect;;
open Model;;
open Io;;

open Printf;;

(* ========================================================================= *)
type filter_type =
    FILTER_SUM_PREV_HOM | FILTER_PREV | FILTER_HOM | FILTER_HR;;

type model_action_type =
    STATS
  | ERRORS
  | FILTER of filter_type
  | PRETTY_PRINT of (string*int*int)
;;

let stat_model arg =
  try
    let name = "stat_model"
    and author = "Pierre Lemaire"
    and version = "2007-06-29"
    and copyright = "2005"
    and description = "computes some characteristic of a model"
    and usage_msg = "usage is: "^Sys.argv.(0)^" [options] model"
    and instfiles = ref []
    and modelfile = ref ""

    and nbvar = ref true
    and listvar = ref true
    and nbbinvar = ref true
    and listbinvar = ref true
    and nbpat = ref true
    and nbpos = ref true
    and nbneg = ref true
    and accuracy = ref true
    and sensitivity = ref true
    and specificity = ref true
    and coverage = ref true
    and poscover = ref true
    and posbyneg = ref true
    and posbypos = ref true
    and negcover = ref true
    and negbypos = ref true
    and negbyneg = ref true
    and uncovered = ref true

    and nerror = ref true
    and nposer = ref true
    and nneger = ref true
    and rerror = ref true
    and rposer = ref true
    and rneger = ref true
    and lposer = ref true
    and lneger = ref true

    and totalpat = ref 100

    and action = ref STATS

    and sdumb = ref ""
    and idumb = ref 0
    in let arglist =
	(common_options (name,version,description,author,copyright))
	@(eol_options)
	@[

	  (* --- i/o ---------------------------------------------- *)
	  ("-inst",
           Arg.String (fun s -> instfiles := s::(!instfiles)),
           "(file) add this instance for evaluation.");

	  (* --- common options ----------------------------------- *)
	  ("-all",
           Arg.Unit (fun () ->
		       set_names true;
		       nbvar := true;
		       listvar := true;
		       nbbinvar := true;
		       listbinvar := true;
		       nbpat := true;
		       nbpos := true;
		       nbneg := true;
		       accuracy := true;
		       sensitivity := true;
		       specificity := true;
		       coverage := true;
		       poscover := true;
		       posbypos := true;
		       posbyneg := true;
		       negcover := true;
		       negbypos := true;
		       negbyneg := true;
		       uncovered := true;
		       nerror := true;
		       nposer := true;
		       nneger := true;
		       rerror := true;
		       rposer := true;
		       rneger := true;
		       lposer := true;
		       lneger := true;
		    ),
           " turns everything on.");
	  ("-none",
           Arg.Unit (fun () ->
		       set_names false;
		       nbvar := false;
		       listvar := false;
		       nbbinvar := false;
		       listbinvar := false;
		       nbpat := false;
		       nbpos := false;
		       nbneg := false;
		       accuracy := false;
		       sensitivity := false;
		       specificity := false;
		       coverage := false;
		       poscover := false;
		       posbypos := false;
		       posbyneg := false;
		       negcover := false;
		       negbypos := false;
		       negbyneg := false;
		       uncovered := false;
		       nerror := false;
		       nposer := false;
		       nneger := false;
		       rerror := false;
		       rposer := false;
		       rneger := false;
		       lposer := false;
		       lneger := false;
		    ),
           " turns everything off.");

	  (* --- stats options ------------------------------------ *)

	  ("-nbvar",
           Arg.Set nbvar,
           " print the number of variables.");
	  ("-nonbvar",
           Arg.Clear nbvar,
           " do not print the number of variables.");
	  ("-varlist",
           Arg.Set listvar,
           " print the list of the variables.");
	  ("-novarlist",
           Arg.Clear listvar,
           " do not print the list of the variables.");
	  ("-nbbinvar",
           Arg.Set nbbinvar,
           " print the number of binary variables.");
	  ("-nonbbinvar",
           Arg.Clear nbbinvar,
           " do not print the number of binary variables.");
	  ("-binvarlist",
           Arg.Set listbinvar,
           " print the list of the binary variables.");
	  ("-nobinvarlist",
           Arg.Clear listbinvar,
           " do not print the list of the binary variables.");

	  ("-nbpat",
           Arg.Set nbpat,
           " print the number of patterns.");
	  ("-nonbpat",
           Arg.Clear nbpat,
           " do not print the number of patterns.");
	  ("-nbpos",
           Arg.Set nbpos,
           " print the number of positive patterns.");
	  ("-nonbpos",
           Arg.Clear nbpos,
           " do not print the number of positive patterns.");
	  ("-nbneg",
           Arg.Set nbneg,
           " print the number of negative patterns.");
	  ("-nonbneg",
           Arg.Clear nbneg,
           " do not print the number of negative patterns.");
	  ("-allpat",
           Arg.Unit (fun () -> nbpat := true; nbpos := true; nbneg := true),
           " turns on all pattern related features.");
	  ("-nonepat",
           Arg.Unit (fun () -> nbpat := false; nbpos := false; nbneg := false),
           " turns off all pattern related features.");

	  ("-accuracy",
           Arg.Set accuracy,
           " print accuracy.");
	  ("-noaccuracy",
           Arg.Clear accuracy,
           " do not print accuracy.");
	  ("-sensitivity",
           Arg.Set sensitivity,
           " print sensitivity.");
	  ("-nosensitivity",
           Arg.Clear sensitivity,
           " do not print sensitivity.");
	  ("-specificity",
           Arg.Set specificity,
           " print specificity.");
	  ("-nospecificity",
           Arg.Clear specificity,
           " do not print specificity.");
	  ("-coverage",
           Arg.Set coverage,
           " print coverage.");
	  ("-nocoverage",
           Arg.Clear coverage,
           " do not print coverage.");
	  ("-poscover",
           Arg.Set poscover,
           " print positive coverage.");
	  ("-noposcover",
           Arg.Clear poscover,
           " do not print positive coverage.");
	  ("-posbypos",
           Arg.Set posbypos,
           " print positive coverage by positive patterns.");
	  ("-noposbypos",
           Arg.Clear posbypos,
           " do not print positive coverage by positive patterns.");
	  ("-posbyneg",
           Arg.Set posbyneg,
           " print positive coverage by negative patterns.");
	  ("-noposbypos",
           Arg.Clear posbyneg,
           " do not print positive coverage by nagative patterns.");
	  ("-negcover",
           Arg.Set negcover,
           " print negative coverage.");
	  ("-nonegcover",
           Arg.Clear negcover,
           " do not print negative coverage.");
	  ("-negbypos",
           Arg.Set negbypos,
           " print negative coverage by positive patterns.");
	  ("-nonegbypos",
           Arg.Clear negbypos,
           " do not print negative coverage by positive patterns.");
	  ("-negbyneg",
           Arg.Set negbyneg,
           " print negative coverage by negative patterns.");
	  ("-nonegbypos",
           Arg.Clear negbyneg,
           " do not print negative coverage by nagative patterns.");
	  ("-uncovered",
           Arg.Set uncovered,
           " print uncovered observations (instance order, first index is 0).");
	  ("-nouncovered",
           Arg.Clear uncovered,
           " do not print uncovered observations.");
	  ("-allcov", Arg.Unit (fun () -> 
				  coverage := true;
				  poscover := true;
				  posbypos := true;
				  posbyneg := true;
				  negcover := true;
				  negbypos := true;
				  negbyneg := true;
				  uncovered := true;
			       ),
           " turns on all coverage-related features.");
	  ("-nonecov", Arg.Unit (fun () -> 
				   coverage := false;
				   poscover := false;
				   posbypos := false;
				   posbyneg := false;
				   negcover := false;
				   negbypos := false;
				   negbyneg := false;
				   uncovered := false;
				),
           " turns off all coverage-related features.");

	  (* --- error options ------------------------------------ *)

	  ("-nerror", Arg.Set nerror,
	   " print number of errors");
	  ("-nonerror", Arg.Clear nerror,
	   " do not print number of errors");
	  ("-nposer", Arg.Set nposer,
	   " print number of positive errors");
	  ("-nonposer", Arg.Clear nposer,
	   " do not print number of positive errors");
	  ("-nneger", Arg.Set nneger,
	   " print number of negative errors");
	  ("-nonnerer", Arg.Clear nneger,
	   " do not print number of negative errors");
	  ("-rerror", Arg.Set rerror,
	   " print ratio of errors");
	  ("-norerror", Arg.Clear rerror,
	   " do not print ratio of errors");
	  ("-rposer", Arg.Set rposer,
	   " print ratio of positive errors");
	  ("-norposer", Arg.Clear rposer,
	   " do not print ratio of positive errors");
	  ("-rneger", Arg.Set rneger,
	   " print ratio of regative errors");
	  ("-nornerer", Arg.Clear rneger,
	   " do not print ratio of regative errors");
	  ("-lposer", Arg.Set lposer,
	   " print list of positive errors");
	  ("-nolposer", Arg.Clear lposer,
	   " do not print list of positive errors");
	  ("-lneger", Arg.Set lneger,
	   " print list of negative errors");
	  ("-nolneger", Arg.Clear lneger,
	   " do not print list of negative errors");

	  (* --- filter options ----------------------------------- *)

	  ("-n", Arg.Set_int totalpat,
	   " total number of patterns to keep (filter)");

	  (* --- actions ------------------------------------------ *)

	  ("-errors",
	   Arg.Unit (fun () -> action := ERRORS),
	   " prints the errors of the model, instead of the stats.");

	  ("-filter",
	   Arg.String (fun s -> try match s.[0] with
			 | 's' -> action := FILTER FILTER_SUM_PREV_HOM
			 | 'p' -> action := FILTER FILTER_PREV
			 | 'h' -> action := FILTER FILTER_HOM
			 | 'x' -> action := FILTER FILTER_HR
			 | _ -> warning "unknwon filter method (use s)";
			     action := FILTER FILTER_SUM_PREV_HOM
		       with
			 | _ -> warning "unknwon filter method (use s)";
			     action := FILTER FILTER_SUM_PREV_HOM
		      ),
	   "(string) filter patterns according to s-um(pre+hom), h-om, p-rev, x(hr).");

	  ("-prettyprint",
	   Arg.Tuple 
	     [Arg.Set_string sdumb;
	      Arg.Set_int idumb;
	      Arg.Int (fun i -> action := PRETTY_PRINT (!sdumb, !idumb, i))],
	   "(string*int*int) pretty print a model for mode*#pos*#neg (modes are [txt|tex|xtxt|xtex])");

	]
    in
      (* initiate program *)
      if (Array.length arg = 0) then
	Arg.parse (Arg.align arglist) (fun s -> modelfile := s) usage_msg
      else begin
	let x = ref 0
	in
	  Arg.parse_argv
	    ~current:x
	    arg
	    (Arg.align arglist)
	    (fun s -> modelfile := s) usage_msg
      end;
      flush_all();

      Observation.set_compare_function Observation.compare_index;
      let model = read_model (open_in !modelfile)
      in begin match (!action) with

	(*======================================================= *)
	| STATS -> 
	    (* instance independent statistics *)
	    let var =
	      if !nbvar || !listvar 
	      then model#get_variables() else []
	    and binvar =
	      if !nbbinvar || !listbinvar then model#get_monomes() else []
	    in
	      if (!nbvar) then
		fprintf (out()) "%s%d%s"
		  (if (names()) then "#var = " else "")
		  (List.length var)
		  !eol;
	      if (!listvar) then
		fprintf (out()) "%s%s%s"
		  (if (names()) then "variables = " else "")
		  (String.concat
		     " "
		     (List.map (fun i -> "x"^(string_of_int i)) var))
		  !eol;

	      if (!nbbinvar) then
		fprintf (out()) "%s%d%s"
		  (if (names()) then "#binvar = " else "")
		  (List.length binvar)
		  !eol;
	      if (!listbinvar) then
		fprintf (out()) "%s%s%s"
		  (if (names()) then "binvariables = " else "")
		  (String.concat
		     "; "
		     (List.map (string_of_monome) binvar))
		  !eol;

	      if (!nbpat) then
		fprintf (out()) "%s%d%s"
		  (if (names()) then "#pat = " else "")
		  ((model#count_of_kind 1)+(model#count_of_kind 0))
		  !eol;
	      if (!nbpos) then
		fprintf (out()) "%s%d%s"
		  (if (names()) then "#pos = " else "")
		  (model#count_of_kind 1)
		  !eol;
	      if (!nbneg) then
		fprintf (out()) "%s%d%s"
		  (if (names()) then "#neg = " else "")
		  (model#count_of_kind 0)
		  !eol;
	      
	      (* instance dependent statistics *)
	      let inst_stat finst = try
		let inst = read_instance (open_in finst)
		in
		  inst#sort();
		  model#compute_for inst; 
		  let (cov,
		       poscov,pospos,posneg,
		       negcov,negpos,negneg,
		       listcov) = 
		    if (!coverage
			|| !poscover || !posbypos || !posbyneg
			|| !negcover || !negbypos || !negbyneg
			|| !uncovered) then
		      let c,pc,pp,pn,nc,np,nn,cl =
			ref 0,ref 0,ref 0,ref 0,ref 0,ref 0,ref 0,ref []
		      in
			for o=0 to inst#m - 1 do
			  let x = model#coverage (inst#observation o)
			  and k = (inst#observation o)#kind
			  in
			    begin
			      let s = Array.fold_left (fun x y -> x+y) 0 x
			      in try if s > 0 then begin
				  c := !c + 1;
				  if (k=0) then nc := !nc + 1;
				  if (k=1) then pc := !pc + 1;
				end
				else cl := (inst#m-o-1)::!cl;
				with Invalid_argument "index out of bounds"
				    -> ()
			    end;
			    begin try if x.(0) > 0 then begin
			      if (k=0) then nn := !nn + 1;
			      if (k=1) then pn := !pn + 1;
			    end with Invalid_argument "index out of bounds"
				-> ()
			    end;
			    begin try if x.(1) > 0 then begin
			      if (k=0) then np := !np + 1;
			      if (k=1) then pp := !pp + 1;
			    end with Invalid_argument "index out of bounds"
				-> ()
			    end;
			done;
			let fm = float_of_int inst#m
			and fp = float_of_int (inst#count_of_kind 1)
			and fn = float_of_int (inst#count_of_kind 0)
			in
			  ((float_of_int !c) /. fm,
 			   (float_of_int !pc)/. fp,
			   (float_of_int !pp)/. fp,
			   (float_of_int !pn)/. fp,
 			   (float_of_int !nc)/. fn,
			   (float_of_int !np)/. fn,
			   (float_of_int !nn)/. fn,
			   !cl)
		    else (0.0,0.0,0.0,0.0,0.0,0.0,0.0,[])
		  in
		    fprintf (out()) "%s%s%s%s%s%s%s%s%s%s%s%s%s"
		      (if (names()) then finst else "")
		      (if (!accuracy) then
			 sprintf "\t%f" model#accuracy else "")
		      (if (!sensitivity) then
			 sprintf "\t%f" model#sensitivity else "")
		      (if (!specificity) then
			 sprintf "\t%f" model#specificity else "")
		      (if (!coverage) then sprintf "\t%f" cov else "")
		      (if (!poscover) then sprintf "\t%f" poscov else "")
		      (if (!posbypos) then sprintf "\t%f" pospos else "")
		      (if (!posbyneg) then sprintf "\t%f" posneg else "")
		      (if (!negcover) then sprintf "\t%f" negcov else "")
		      (if (!negbypos) then sprintf "\t%f" negpos else "")
		      (if (!negbyneg) then sprintf "\t%f" negneg else "")
		      (if (!uncovered) then sprintf "\t%s"
			 (String.concat " " (List.map string_of_int listcov))
		       else "")
		      !eol
	      with 
		| Sys_error msg -> 
		    warning (sprintf "can't open file (%s) [discard file]"
			       msg);
		| IO_error t ->
		    print_io_error t;
		    warning (sprintf "troubles with %s [discard file]"
			       finst);
	      in
		if names() then begin
		  fprintf (out()) "%s%s%s%s%s%s%s%s%s%s%s%s%s"
		    (if (names()) then "FILE" else "")
		    (if (!accuracy) then "\tACCURACY" else "")
		    (if (!sensitivity) then "\tSENSITIVITY" else "")
		    (if (!specificity) then "\tSPECIFICITY" else "")
		    (if (!coverage) then "\tCOVERAGE" else "")
		    (if (!poscover) then "\tPOSCOVER" else "")
		    (if (!posbypos) then "\tPOSBYPOS" else "")
		    (if (!posbyneg) then "\tPOSBYNEG" else "")
		    (if (!negcover) then "\tNEGCOVER" else "")
		    (if (!negbypos) then "\tNEGBYPOS" else "")
		    (if (!negbyneg) then "\tNEGBYNEG" else "")
		    (if (!uncovered) then "\tUNCOVERED" else "")
		    !eol
		end;
		List.iter inst_stat !instfiles;

	(*======================================================= *)
	| ERRORS ->
	    let inst_errors finst =
	      let inst = read_instance (open_in finst)
	      in
		inst#sort();
		let ep,en,lp,ln = ref 0, ref 0, ref [], ref []
		in for o=0 to inst#m - 1 do
		    let k = (inst#observation o)#kind
		    and p = model#classification (inst#observation o)
		    in
		      if p != k then
			if k=0 then begin en := !en + 1; ln := o::!ln; end
			else begin  ep := !ep + 1; lp := o::!lp; end;
		  done;
		  let fep = float_of_int !ep
		  and fen = float_of_int !en
		  in
		    fprintf (out())
		      "%s%s%s%s%s%s%s%s%s%s"
		      (if names() then finst else "")
		      (if !nerror then sprintf "\t%d" (!ep + !en) else "")
		      (if !nposer then sprintf "\t%d" !ep else "")
		      (if !nneger then sprintf "\t%d" !en else "")
		      (if !rerror then
			 sprintf "\t%.4f" ((fep+.fen)/.(float_of_int inst#m))
		       else "")
		      (if !rposer then
			 sprintf "\t%.4f"
			   (fep/.(float_of_int (inst#count_of_kind 1)))
		       else "")
		      (if !rneger then
			 sprintf "\t%.4f"
			   (fen /. (float_of_int (inst#count_of_kind 0)))
		       else "")
		      (if !lposer then sprintf "\t%s"
			 (String.concat " " (List.map string_of_int !lp))
		       else "")
		      (if !lneger then sprintf "\t%s"
			 (String.concat " " (List.map string_of_int !ln))
		       else "")
		      !eol
	    in
	      if names() then begin
		fprintf (out()) "%s%s%s%s%s%s%s%s%s%s"
		  (if (names()) then "FILE" else "")
		  (if (!nerror) then "\t#ERROR" else "")
		  (if (!nposer) then "\t#POSER" else "")
		  (if (!nneger) then "\t#NEGER" else "")
		  (if (!rerror) then "\t%ERROR" else "")
		  (if (!rposer) then "\t%POSER" else "")
		  (if (!rneger) then "\t%NEGER" else "")
		  (if (!lposer) then "\tLPOSER" else "")
		  (if (!lneger) then "\tLNEGER" else "")
		  !eol
	      end;
	      List.iter inst_errors !instfiles;
	      

	(*======================================================= *)
	| FILTER filter ->
	    let rec keep n ls =
	      match ls with 
		| []    -> []
		| l::ls -> if n>0 then l::(keep (n-1) ls) else [] 
	    in
	    let totalpos = model#count_of_kind 1
	    and totalneg = model#count_of_kind 0
	    in let totalpos = totalpos * !totalpat / (totalpos+totalneg)
	       and totalneg = totalneg * !totalpat / (totalpos+totalneg)
	    in
	      Pattern.set_compare_function
		(match filter with
		   | FILTER_SUM_PREV_HOM ->
		       (fun (p:pattern) (q:pattern) ->
			  - Pervasives.compare
			    (p#homogeneity +. p#max_prevalence) 
			    (q#homogeneity +. q#max_prevalence)
		       )
		   | FILTER_PREV ->
		       (fun (p:pattern) (q:pattern) ->
			  - Pervasives.compare
			    p#max_prevalence
			    q#max_prevalence
		       )
		   | FILTER_HOM ->
		       (fun (p:pattern) (q:pattern) ->
			  - Pervasives.compare
			    p#homogeneity
			    q#homogeneity		
		       )
		   | FILTER_HR ->
		       (fun (p:pattern) (q:pattern) ->
			  - Pervasives.compare
			    p#hazardratio
			    q#hazardratio
		       )
		);
	      model#sort();
	      let pos = keep totalpos (model#patterns_of_kind 1)
	      and neg = keep totalneg (model#patterns_of_kind 0)
	      in let m = new pandect 1 (List.rev_append pos neg)
	      in
		print_pandect (out()) m

	(*======================================================= *)
	| PRETTY_PRINT x -> let mode,pnum,nnum = x in
	  let sep,nl = match mode with
	    | "txt" | "xtxt" -> ";", ""
	    | "tex" | "xtex" -> "&", "\\\\"
	    | _ -> error 1 (sprintf "unknown pretty print mode %s" mode);
		"", ""
	  and exact = match mode with
	    | "xtxt" | "xtex" -> true
	    | _ -> false
	  in
	  let pnum,nnum = float_of_int pnum, float_of_int nnum
	  and var = model#get_variables()
	  in
	  let headers =
	    "name"^sep^"pos.prev"^sep^"neg.prev"^sep^"homogeneity"^sep
	    ^"hazardratio"^sep
	    ^(String.concat
		sep 
		(List.map (fun v -> "x"^(string_of_int v)) var)
	     )
	  in
	  let inst =
	    if exact then
	      read_instance (open_in (List.hd !instfiles))
	    else (new instance 0 [||])
	  in
	  let print p =
	    let round x = (truncate x) + (if snd (modf x) >= 0.5 then 1 else 0)
	    in
	    let s = ref ""
	    and pre, hom, hr = p#max_prevalence, p#homogeneity, p#hazardratio
	    in 
	    let pcov, ncov, ppre, npre =
	      if exact then begin
		p#compute_for inst;
		((int_of_float ((p#prevalence 1) *. (float_of_int (inst#count_of_kind 1)))),
		 (int_of_float ((p#prevalence 0) *. (float_of_int (inst#count_of_kind 0)))),
		 p#prevalence 1,
		 p#prevalence 0)
	      end
	      else
		let pcov, ncov = 
		  if p#kind == 0 then
		    round (nnum *. pre *. (1./.hom -. 1.)), round (nnum *. pre)
		  else
		    round (pnum *. pre), round (pnum *. pre *. (1./.hom -. 1.))
		in
		  (pcov,
		   ncov,
		   (float_of_int pcov) /. pnum,
		   (float_of_int ncov) /. nnum)
	    in
	      (* name, ppre, npre, hom, hr *)
	      s := 
		( (if p#kind = 0 then "N" else "P")
		  ^(string_of_int p#index)
		  ^sep
		  ^(sprintf "%d (%.2f%%)" pcov (ppre*.100.))
		  ^sep
		  ^(sprintf "%d (%.2f%%)" ncov (npre*.100.))
		  ^sep
		  ^(sprintf "%.2f%%" (hom*.100.))
		  ^sep
		  ^(sprintf "%.2f" hr)
		);
	      (* variables *)
	      List.iter
		(fun v ->
		   let u,l = p#ub v, p#lb v
		   in
		     s := !s^sep;
		     if u < infinity && l > neg_infinity then
		       s := !s^(sprintf "> %6f, < %6f" l u)
		     else if u < infinity then
		       s := !s^(sprintf "< %6f" u)
		     else if l > neg_infinity then
		       s := !s^(sprintf "> %6f" l);
		)
		var;
	      s := !s^nl;
	      !s
	  in
	    (* headers *)
	    output_string (out()) (headers^(!eol));
	    (* patterns *)
	    List.iter
	      (fun p -> output_string (out()) ((print p)^(!eol)))
	      (model#all_patterns())
	end;

	(* terminates program *)
	0;
  with
    | _ as ex -> terminal_catch ex
;;

